/******************************************************************************

    Program:        pdf_inc.p

    Written By:     Gordon Campbell - PRO-SYS Consultants Ltd.
    Written On:     November 10, 2003

    Description:    Contains function and variable definitions for
                    generating a PDF document from within Progress

    Note:           This was copied from pdf_inc.i

    --------------------- Revision History ------------------

    Date:     Author        Change Description

    11/10/03  G Campbell    Initial Copy from pdf_inc.i

    11/17/03  G Campbell    Removed hard-coded 'sPDF' strings

                            Added pdf_encrypt procedure to allow for
                            silent or interactive PDF encryption.  Integrates
                            with the PSP (Pretty Safe PDF) product from PDFlib
                            GmbH.

                            Contact gcampbell@epro-sys.com to purchase a copy
                            of the PSP product.  Discounts for PEG Members.

    11/19/03  G Campbell    Changed how the skip was working.  Now skip based
                            on the current PointSize.

    11/20/03  G Campbell    Changed how pdf_text_xy was being processed.  It
                            now processes this text in the Graphic Grid.  It no
                            longer updates the TextX and TextY values.

    11/21/03  G Campbell    Added pdf_text_char procedure

    11/24/03  G Campbell    Fixed replace text bug RE:  Bug #847880

    11/25/03  G Campbell    Fixed European Numberic Format problem
                            RE: Bug #848975

                            This issue was previously identified by Herbert Bayer
                            (Herbert.Bayer@bundrinno.de) and I've used his
                            methodology for accomodating this issue.

    11/27/03  G Campbell    Added ability to add bookmarks to PDF documents. See
                            the custlist.p samples for sample implementation.

                            Included adding the following procedures:

                            pdf_bookmark (public procedure)
                            pdf_load_bookmarks (private procedure)
                            pdf_process_bookmarks (private procedure)

    12/04/03  G Campbell    Added following procedures:

                            pdf_set_linejoin
                            pdf_circle  ( Note: Graphic X and Y points become the
                                                X and Y Parameters - center point)
                            pdf_set_FillRed
                            pdf_set_FillGreen
                            pdf_set_FillBlue

                            Added following functions:

                            pdf_FillRed
                            pdf_FillGreen
                            pdf_FillBlue

                            string2dec (private - as per Peter Kiss
                                                  (peter.kiss@paradyme.be)

                            Plus: a whole whack of procedures used to
                                  merge disparate streams.  See sample
                                  procedures layer1.p and layer2.p.

                                  Thanks to Peter Kiss for this merging
                                  capability.  Hopefully I'm decribing the
                                  procedure uses effectively.

                            pdf_merge_stream
                            pdf_merge_stream_content (private)
                            pdf_merge_stream_link (private)
                            pdf_merge_stream_param (private)
                            pdf_merge_stream_image (private)
                            pdf_merge_stream_diff (private)
                            pdf_merge_stream_book (private)
                            pdf_merge_stream_book_detail (private)
                            pdf_ReplaceText

  12/17/03  G Campbell      Added procedure pdf_set_PageRotation
                            Added function pdf_PageRotation
                            Added temp-table TT_Pages to track Pages.

                            Updated pdf_reset_stream and pdf_reset_all to
                            ensure that all Temp Tables are being cleared
                            correctly.

  12/18/03  G Campbell      Added ability to Compress Page Streams using the
                            zlib (www.zlib.net) compression library.

                            Thanks to Maurits van Rijnen (peg@vanrijnen.nl)
                            for providing the command syntax in zlib.p and
                            zlibbind.p.

                            Added function pdf_get_parameter

                            Added procedures:
                            pdf_set_parameter
                            OutputMemPtr (private) - used to output Compressed
                                                     data

                            Added appropriate indices to Temp Tables.  Helped
                            with performance on large PDF files (thanks again
                            Herbert Bayer).

  12/29/03  G Campbell      Added ability to Encrypt using PDFencrypt.p (which
                            uses a couple of C functions for binary data
                            manipulation)

                            pdfencryopt.p currently only handles 40-bit
                            Standard Adobe Encryption.

  01/08/04  G Campbell      Started implementing Tools functionality.  This
                            included adding:

                            h_PDF-tool  Procedure Handle

                            pdf_tool_add Internal Procedure
                            pdf_tool_create Internal Procedure
                            pdf_set_tool_parameter Internal Procedure
                            pdf_get_tool_parameter Function

  01/09/04  G Campbell      Fixed issue with hardcoding of c:\temp\content.txt.
                            Now uses TT_pdf_stream.obj_UniqueID + "-C.txt" to
                            create Content file.

                            Thanks to Mikael Harjpe [mikael.hjerpe@stamford.se]
                            for pointing out this error.

  01/15/04  G Campbell      As per Gerben Wieringa (g.wieringa@vcd.nl)
  
                            In order to accommodate double-byte codepage 
                            (eg: 1252), with the startup parameter -checkdbe 
                            (check double byte enabled)
                            
                              - Added "character" option to all length() and 
                                substr() functions.
                              - Had to add "-1" in some of the substr() 
                                function-calls as third argument
                              
                            Added SESSION:TEMP-DIR when generating Content
                            File.
  
  01/16/04  G Campbell      Modified how the pdf_wrap_text procedure was 
                            working.  It wasn't correctly accounting for CHR(10)
                            within the text string plus it wasn't determining
                            the maximum allowable characters properly.  
                            
                            NOTE:  If you are currently using pdf_wrap_text 
                                   then these changes may result in changes to 
                                   your output

  01/20/04  G Campbell      Included updated merge methods - supplied by
                            Peter Kiss (peter.kiss@paradyme.be) of Paradyme 
                            Belgium.
                            
  01/20/04  G Campbell      Added procedure pdf_curve - this procedure allows
                            you to create a Bezier curve based on three x,y
                            locations.  If you call pdf_curve you need to also 
                            call pdf_close_path otherwise the curve won't 
                            appear.
                            
                            Also modified pdf_circle to use pdf_curve. 
                            pdf_circle automatically closes the 'path' (via
                            pdf_close_path or the curve.p in the samples/super
                            subdirectory).
                            
  01/21/04  G Campbell      Added function pdf_GetNumFittingChars.  This
                            replaces pdf_get_NumFittingChars (deprecated).
                            
                            Requires that you pass in the From X and and to X
                            point locations.  The difference is then used to
                            determine the maximum allowable characters.

                            Plus, Gerben missed adding one 'character' option on
                            a LENGTH statement.  Resolved.

  01/21/04  G Campbell    Added pdf_text_charxy procedure
                          This allows us to add a special character at a given
                          X/Y position

  01/21/04  G Campbell    Added function pdf_text_widthdec.  Similar to 
                          pdf_text_width except that it returns a decimal value 
                          for the text width.  This is useful for more specific
                          placement and verification of text size.

  01/22/04  G Campbell    Fixed issue with pdf_get_NumFittingChars and 
                          pdf_GetNumFittingChars.  Need to increment the Extent
                          by plus one for External fonts because the AFM file 
                          starts at Character zero.  Also had to increment the
                          FILL in the Base 14 font definitions to accomodate 
                          this.
  
                          Added Symbol and ZapfDingbats Standard Fonts.  The 
                          Base14 Fonts are now complete.

                          Added procedure pdf_text_align.  This allow us to 
                          position text based on an explicit X location. 
                          
                          Possible alignment possiblities are LEFT, RIGHT

                          Removed pdf_get_numFittingChars ... 

  01/28/04  G Campbell    BUG: During the addition of the ZapfDingbats and 
                          Symbol basefonts I had set the default Encoding to
                          StandardEncoding from WinAnsiEncoding.  This caused
                          issues with some installations.  I have restored the
                          default Encoding to WinAnsiEncoding and specifically
                          set the Encoding on the Symbol and ZapfDingbats to
                          StandardEncoding.
                          
  01/28/04  G Campbell    Changed Based Font Names from /Fn to /BFn.  Hopefully
                          this will be unique enough for future processing. At
                          this point I'm currently looking at being able to
                          include Pages from existing PDFs.  Including Fonts
                          may be an issue ... 

  01/29/04  G Campbell    Use dec2string in pdf_text_align - as per Peter Kiss.
  
  02/20/04  G Campbell    Added NO-UNDO to all TEMP-TABLE declarations

  02/24/04  G Campbell    Added new parameter called "LineSpacer".  This
                          parameter allows you to 'add' a number of points 
                          between each line when applying a pdf_skip.  This will
                          hopefully allow people to mimic the older method of
                          handling line feeds.

  02/25/04  G Campbell    Added procedure pdf_exec_footer as per Peter Kiss
  
                          This procedure allows you to call the Footer creation
                          procedure from anywhere.  This was the result of a 
                          small problem using the footer when you are sharing 
                          the stream over diff. programs.

                          Eg: When using a main program (without a footer set) 
                          then calling another program to do some printing 
                          using the same stream (file) and in the called 
                          program you specify a header and footer, 
                          then the last footer will not be printed bacause 
                          this is done in pdf_close
                          
******************************************************************************/


&IF OPSYS = "UNIX" &THEN
  &GLOBAL-DEFINE pdfSKIP     CHR(13) CHR(10)
&ELSE
  &GLOBAL-DEFINE pdfSKIP     SKIP
&ENDIF

{ zlib.i }

/* The following defines are used to determine the JPEG Image Height and Width */
&GLOBAL-DEFINE M_SOF0  "0xC0"		/* Start Of Frame N */
&GLOBAL-DEFINE M_SOF1  "0xC1"		/* N indicates which compression process */
&GLOBAL-DEFINE M_SOF2  "0xC2"       /* Only SOF0-SOF2 are now in common use */
&GLOBAL-DEFINE M_SOF3  "0xC3"
&GLOBAL-DEFINE M_SOF5  "0xC5"		/* NB: codes C4 and CC are NOT SOF markers */
&GLOBAL-DEFINE M_SOF6  "0xC6"
&GLOBAL-DEFINE M_SOF7  "0xC7"
&GLOBAL-DEFINE M_SOF9  "0xC9"
&GLOBAL-DEFINE M_SOF10 "0xCA"
&GLOBAL-DEFINE M_SOF11 "0xCB"
&GLOBAL-DEFINE M_SOF13 "0xCD"
&GLOBAL-DEFINE M_SOF14 "0xCE"
&GLOBAL-DEFINE M_SOF15 "0xCF"
&GLOBAL-DEFINE M_SOI   "0xD8"		/* Start Of Image (beginning of datastream) */
&GLOBAL-DEFINE M_EOI   "0xD9"		/* End Of Image (end of datastream) */
&GLOBAL-DEFINE M_SOS   "0xDA"		/* Start Of Scan (begins compressed data) */
&GLOBAL-DEFINE M_APP0  "0xE0"		/* Application-specific marker, type N */
&GLOBAL-DEFINE M_APP12 "0xEC"		/* (we don't bother to list all 16 APPn's) */
&GLOBAL-DEFINE M_COM   "0xFE"		/* COMment */
&GLOBAL-DEFINE M_MARK  "0xFF"   /* Marker */

/* ---------------------------- Define TEMP-TABLES -------------------------
   The temp-tables are used to store the PDF streams and resources used when
   generating a PDF document */
DEFINE TEMP-TABLE TT_pdf_stream NO-UNDO
    FIELD obj_stream      AS CHARACTER
    FIELD obj_file        AS CHARACTER
    FIELD obj_encrypt     AS LOGICAL INIT FALSE
    FIELD obj_master      AS CHARACTER
    FIELD obj_user        AS CHARACTER
    FIELD obj_access      AS CHARACTER
    FIELD obj_key         AS INTEGER
    FIELD obj_silent      AS LOGICAL INIT FALSE
    FIELD obj_mode        AS CHARACTER
    FIELD obj_CallProc    AS HANDLE
    FIELD obj_footer      AS CHARACTER
    FIELD obj_header      AS CHARACTER
    FIELD obj_last        AS CHARACTER
    FIELD obj_id          AS CHARACTER
    FIELD obj_EncryptDict AS INTEGER
    FIELD obj_UniqueID    AS CHARACTER
    FIELD obj_P           AS INTEGER
INDEX obj_stream  AS PRIMARY
      obj_stream.

/* The following temp-table is used to store/track parameters per stream */
DEFINE TEMP-TABLE TT_pdf_param NO-UNDO
    FIELD obj_stream    AS CHARACTER
    FIELD obj_parameter AS CHARACTER
    FIELD obj_valid     AS CHARACTER
    FIELD obj_value     AS CHARACTER
INDEX obj_stream AS PRIMARY
      obj_stream
      obj_parameter.

DEFINE TEMP-TABLE TT_pdf_error NO-UNDO
  FIELD obj_stream  AS CHARACTER
  FIELD obj_func    AS CHARACTER FORMAT "x(20)"
  FIELD obj_error   AS CHARACTER FORMAT "x(40)"
INDEX obj_stream AS PRIMARY
      obj_stream.

DEFINE TEMP-TABLE TT_pdf_page NO-UNDO
  FIELD obj_stream  AS CHARACTER
  FIELD page_nbr    AS INTEGER
  FIELD page_rotate AS INTEGER
INDEX obj_stream AS PRIMARY
      obj_stream
      page_nbr.

DEFINE TEMP-TABLE TT_pdf_tool NO-UNDO
  FIELD obj_stream   AS CHARACTER
  FIELD tool_name    AS CHARACTER
  FIELD tool_type    AS CHARACTER
  FIELD tool_handle  AS HANDLE
INDEX obj_stream AS PRIMARY UNIQUE
      obj_stream
      tool_name.

DEFINE TEMP-TABLE TT_pdf_tool_param NO-UNDO
  FIELD obj_stream   AS CHARACTER
  FIELD tool_name    AS CHARACTER
  FIELD tool_param   AS CHARACTER
  FIELD tool_col     AS INTEGER
  FIELD tool_value   AS CHARACTER
INDEX obj_stream AS PRIMARY UNIQUE
      obj_stream
      tool_name
      tool_param
      tool_col.

/* The following temp-table is used to build a list of objects that will appear
   in the PDF document */
DEFINE TEMP-TABLE TT_pdf_object NO-UNDO
  FIELD obj_stream  AS CHARACTER
  FIELD obj_nbr     AS INTEGER
  FIELD obj_desc    AS CHARACTER
  FIELD obj_offset  AS DECIMAL DECIMALS 0 FORMAT "9999999999"
  FIELD gen_nbr     AS INTEGER FORMAT "99999"
  FIELD obj_type    AS CHARACTER FORMAT "X"
  FIELD obj_page    AS INTEGER
INDEX obj_stream AS PRIMARY
      obj_stream
      obj_nbr.

/* The following temp-table is used to build a list of Bookmarks that will appear
   in the PDF document */
DEFINE TEMP-TABLE TT_pdf_bookmark NO-UNDO
  FIELD obj_stream   AS CHARACTER
  FIELD book_obj     AS INTEGER
  FIELD book_nbr     AS INTEGER
  FIELD book_title   AS CHARACTER
  FIELD book_parent  AS INTEGER
  FIELD book_expand  AS LOGICAL
  FIELD book_child   AS INTEGER
  FIELD book_first   AS INTEGER
  FIELD book_last    AS INTEGER
  FIELD book_page    AS INTEGER
  FIELD book_Y       AS INTEGER
INDEX book_nbr AS PRIMARY UNIQUE
      obj_stream
      book_nbr.

/* The following temp-table is used to store the actual content that will appear
   in the PDF document */
DEFINE TEMP-TABLE TT_pdf_content NO-UNDO
    FIELD obj_stream    AS CHARACTER
    FIELD obj_seq       AS INTEGER
    FIELD obj_type      AS CHARACTER
    FIELD obj_content   AS CHARACTER FORMAT "X(60)"
    FIELD obj_page      AS INTEGER
    FIELD obj_line      AS INTEGER
    FIELD obj_length    AS INTEGER
    FIELD obj_function  AS LOGICAL
INDEX obj_seq AS PRIMARY
      obj_stream
      obj_seq
INDEX obj_page
      obj_page
INDEX obj_stream
      obj_stream
INDEX obj_type
      obj_type
INDEX obj_pagetype
      obj_page
      obj_type.

/* The following temp-table is used to track Document Information */
DEFINE TEMP-TABLE TT_pdf_info NO-UNDO
    FIELD obj_stream    AS CHARACTER
    FIELD info_attr     AS CHARACTER
    FIELD info_value    AS CHARACTER
INDEX obj_stream AS PRIMARY
      obj_stream.

/* The following temp-table is used to track Links (Annotations)
   loaded into a PDF stream */
DEFINE TEMP-TABLE TT_pdf_link NO-UNDO
    FIELD obj_stream    AS CHARACTER
    FIELD link_color    AS CHARACTER
    FIELD link_name     AS CHARACTER
    FIELD link_page     AS INTEGER
    FIELD link_x        AS INTEGER
    FIELD link_y        AS INTEGER
    FIELD link_h        AS INTEGER
    FIELD link_w        AS INTEGER
    FIELD link_border   AS INTEGER
    FIELD link_style    AS CHARACTER
    FIELD link_obj      AS INTEGER
INDEX obj_stream  AS PRIMARY
      obj_stream.

/* The following temp-table is used to track Images loaded into a PDF stream */
DEFINE TEMP-TABLE TT_pdf_image NO-UNDO
    FIELD obj_stream    AS CHARACTER
    FIELD image_name    AS CHARACTER
    FIELD image_file    AS CHARACTER
    FIELD image_tag     AS CHARACTER
    FIELD image_obj     AS INTEGER
    FIELD image_len     AS INTEGER
    FIELD image_h       AS INTEGER
    FIELD image_w       AS INTEGER
INDEX obj_image AS PRIMARY
      obj_stream
      image_name.

/* The following temp-table is used to track Fonts loaded into a PDF stream */
DEFINE TEMP-TABLE TT_pdf_font NO-UNDO
    FIELD obj_stream        AS CHARACTER
    FIELD font_name         AS CHARACTER
    FIELD font_file         AS CHARACTER
    FIELD font_afm          AS CHARACTER
    FIELD font_dif          AS CHARACTER
    FIELD font_type         AS CHARACTER
    FIELD font_width        AS CHARACTER
    FIELD font_obj          AS INTEGER
    FIELD font_encoding     AS INTEGER
    FIELD font_descr        AS INTEGER
    FIELD font_stream       AS INTEGER
    FIELD font_len          AS INTEGER
    FIELD font_tag          AS CHARACTER
    FIELD afm_ItalicAngle   AS INTEGER
    FIELD afm_Ascender      AS CHARACTER
    FIELD afm_Descender     AS CHARACTER
    FIELD afm_FontBBox      AS CHARACTER
    FIELD afm_FirstChar     AS CHARACTER
    FIELD afm_LastChar      AS CHARACTER
    FIELD afm_Widths        AS CHARACTER
    FIELD afm_IsFixedPitch  AS CHARACTER INIT "0"
    FIELD afm_flags         AS CHARACTER
INDEX font_name AS PRIMARY
      font_name
INDEX obj_stream
      obj_stream.

DEFINE TEMP-TABLE TT_pdf_diff NO-UNDO
  FIELD obj_stream  AS CHARACTER
  FIELD font_name   AS CHARACTER
  FIELD char_num    AS INTEGER
  FIELD PS_name     AS CHARACTER
INDEX obj_stream  AS PRIMARY
      obj_stream.

DEFINE TEMP-TABLE TT_pdf_ReplaceTxt /* peki */ NO-UNDO
  FIELD obj_stream  AS CHARACTER
  FIELD mergenr     AS INTEGER
  FIELD txt_from    AS CHARACTER
  FIELD txt_to      AS CHARACTER
INDEX obj_stream    AS PRIMARY
      obj_stream.

DEFINE TEMP-TABLE TT-Merge-Pages NO-UNDO
  FIELD PageFrom    AS INTEGER
  FIELD PageTo      AS INTEGER
  FIELD MergeNr     AS INTEGER.

/* ---------------------- Define LOCAL VARIABLES -------------------------- */
DEFINE VARIABLE pdf_inc_ContentSequence AS INTEGER NO-UNDO.
DEFINE VARIABLE pdf_inc_ObjectSequence  AS INTEGER NO-UNDO.
DEFINE VARIABLE pdf_OutlinesDict        AS INTEGER NO-UNDO.
DEFINE VARIABLE pdf_OutlinesLast        AS INTEGER NO-UNDO.

/* The following variables are used to store the Image Height Width */
DEFINE VARIABLE pdf_width     AS INTEGER NO-UNDO.
DEFINE VARIABLE pdf_height    AS INTEGER NO-UNDO.

/* Variables used in conjunction with Header/Footer requirements */
DEFINE VARIABLE pdf_ForFooter       AS LOGICAL   NO-UNDO.

/* Miscellaneous Variables */
DEFINE VARIABLE pdf-Res-Object        AS INTEGER NO-UNDO.
DEFINE VARIABLE Vuse-font             AS CHARACTER NO-UNDO.
DEFINE VARIABLE pdf-Stream-Start      AS INTEGER NO-UNDO.
DEFINE VARIABLE pdf-Stream-End        AS INTEGER NO-UNDO.
DEFINE VARIABLE pdf-EncryptKeyMemPtr  AS MEMPTR NO-UNDO.
DEFINE VARIABLE pdf-EncryptMemPtr     AS MEMPTR NO-UNDO.
DEFINE VARIABLE h_PDF-Encrypt         AS HANDLE NO-UNDO.
DEFINE VARIABLE h_PDF-Tool            AS HANDLE NO-UNDO.

DEFINE VARIABLE mContent              AS MEMPTR    NO-UNDO.
DEFINE VARIABLE mHolder               AS MEMPTR    NO-UNDO.

DEFINE STREAM S_pdf_inc.
DEFINE STREAM S_pdf_inp.

DEFINE BUFFER B_TT_pdf_content FOR TT_pdf_content.
DEFINE BUFFER B_TT_pdf_param   FOR TT_pdf_param.

/* ---------------------------- Define FUNCTIONS -------------------------- */

  /* The following functions are used to determine the Images Width/Height */

FUNCTION hex RETURNS CHARACTER (INPUT asc-value AS INTEGER).
  DEF VAR j AS INT  NO-UNDO.
  DEF VAR h AS CHAR NO-UNDO.

  DO WHILE TRUE:
    j = asc-value MODULO 16.
    h = (IF j < 10 THEN STRING(j) ELSE CHR(ASC("A") + j - 10)) + h.
    IF asc-value < 16 THEN LEAVE.
      asc-value = (asc-value - j) / 16.
  END.

  IF LENGTH(h, "character":u) = 1 THEN
    h = "0" + h.

  RETURN ("0x" + h).
END FUNCTION. /* hex */

FUNCTION nextbyte RETURNS INTEGER ():
  DEFINE VARIABLE L_data    AS RAW NO-UNDO.

  LENGTH(L_data) = 1.

  IMPORT STREAM S_pdf_inp UNFORMATTED L_data.
  RETURN GET-BYTE(L_data,1).

END FUNCTION. /* nextbyte */

FUNCTION next2bytes RETURNS INTEGER ():
  DEFINE VARIABLE L_c1      AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_c2      AS INTEGER NO-UNDO.

  L_c1 = nextbyte().
  L_c2 = nextbyte().

  RETURN INT(L_c1 * EXP(2, 8) + L_c2).

END FUNCTION. /* next2bytes */

FUNCTION first_marker RETURN LOGICAL ():
  DEFINE VARIABLE L_c1        AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_c2        AS INTEGER NO-UNDO.

  L_c1 = nextbyte().
  L_c2 = nextbyte().

  IF hex(L_c1) <> {&M_Mark} AND hex(L_c2) <> {&M_SOI} THEN
    RETURN FALSE.
  ELSE RETURN TRUE.
END FUNCTION. /* first_marker */

FUNCTION next_marker RETURN INTEGER():
  DEFINE VARIABLE L_data    AS RAW NO-UNDO.

  LENGTH(L_data) = 1.
  DEFINE VARIABLE L_c       AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_discard AS INTEGER NO-UNDO.

  L_c = nextbyte().
  DO WHILE hex(L_c) <> {&M_MARK}:
    L_discard = L_discard + 1.
    L_c = nextbyte().
  END. /* <> 0xFF */

  DO WHILE hex(L_c) = {&M_MARK}:
    L_c = nextbyte().
  END.

  RETURN L_c.
END FUNCTION. /* next_marker */

FUNCTION skip_variable RETURN LOGICAL ():
  DEFINE VARIABLE L_Length  AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Loop    AS INTEGER NO-UNDO.

  L_length = next2bytes().

  DO L_Loop = 1 TO (L_Length - 2):
    nextbyte().
  END. /* Loop */
END FUNCTION. /* skip_variable */

FUNCTION process_SOF RETURNS LOGICAL ():
  DEFINE VARIABLE L_Length  AS INTEGER NO-UNDO.

  next2bytes().       /* Skip Length */
  nextbyte().         /* Skip Data Precision */
  pdf_height = next2bytes().
  pdf_width  = next2bytes().

END FUNCTION. /* process_SOF */

/* end of Functions used to determine Image Height/Width */

FUNCTION dec2string RETURNS CHARACTER (INPUT decvalue AS DECIMAL):
  IF SESSION:NUMERIC-FORMAT = "EUROPEAN" THEN
    RETURN REPLACE( STRING(decvalue),",",".").
  ELSE
    RETURN STRING(decvalue).
END.  /* dec2string */

FUNCTION string2dec RETURNS DECIMAL (INPUT stringvalue AS CHARACTER):
  IF SESSION:NUMERIC-FORMAT = "EUROPEAN" THEN
    RETURN DEC(REPLACE(stringvalue,".",",")).
  ELSE
    RETURN DEC(stringvalue).
END.  /* string2dec */

FUNCTION pdf_inc_ContentSequence RETURNS INTEGER():
  pdf_inc_ContentSequence = pdf_inc_ContentSequence + 1.

  RETURN pdf_inc_ContentSequence.
END. /* pdf_inc_ContentSequence */

PROCEDURE pdf_error :
  DEFINE INPUT PARAMETER pdfStream     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFunction   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfError      AS CHARACTER NO-UNDO.

  CREATE TT_pdf_error.
  ASSIGN TT_pdf_error.obj_stream = pdfStream
         TT_pdf_error.obj_func   = pdfFunction
         TT_pdf_error.obj_error  = pdfError.

END. /* RUN pdf_error */

FUNCTION pdf_get_parameter RETURNS CHARACTER
         (INPUT pdfStream     AS CHARACTER,
          INPUT pdfParameter  AS CHARACTER):

  DEFINE BUFFER B_TT_pdf_param FOR TT_pdf_param.

  FIND B_TT_pdf_param WHERE B_TT_pdf_param.obj_stream = pdfStream
                        AND B_TT_pdf_param.obj_parameter = pdfParameter
                        NO-LOCK NO-ERROR.
  IF AVAIL B_TT_pdf_param THEN
    RETURN B_TT_pdf_param.obj_value.
  ELSE
    RETURN "".

END FUNCTION. /* pdf_get_parameter */

FUNCTION pdf_Font RETURN CHARACTER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_Font","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                      AND TT_pdf_param.obj_parameter = "Font"
                      NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN TT_pdf_param.obj_value.
  ELSE
    RETURN "Courier".

END FUNCTION. /* pdf_Font */

FUNCTION pdf_FontType RETURN CHARACTER ( INPUT pdfStream AS CHARACTER):

  DEFINE BUFFER L_B_TT_pdf_font FOR TT_pdf_font.

  DEFINE VARIABLE L_Font    AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_FontType","Cannot find Stream!").
    RETURN ERROR.
  END.

  L_Font = pdf_Font(pdfStream).
  FIND FIRST L_B_TT_pdf_font WHERE L_B_TT_pdf_font.obj_stream = pdfStream
                               AND L_B_TT_pdf_font.font_name = L_Font
                               NO-LOCK NO-ERROR.
  IF AVAIL L_B_TT_pdf_font THEN
    RETURN L_B_TT_pdf_font.Font_Type.
  ELSE
    RETURN "FIXED".

END FUNCTION. /* pdf_FontType */

FUNCTION pdf_ImageDim RETURN INTEGER ( INPUT pdfStream AS CHARACTER,
                                       INPUT pdfImage  AS CHARACTER,
                                       INPUT pdfDim    AS CHARACTER):

  DEFINE BUFFER L_B_TT_pdf_image FOR TT_pdf_image.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_ImageDim","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST L_B_TT_pdf_image WHERE L_B_TT_pdf_image.obj_stream = pdfStream
                               AND L_B_TT_pdf_image.image_name  = pdfImage
                               NO-LOCK NO-ERROR.
  IF AVAIL L_B_TT_pdf_image THEN DO:
    IF pdfDim = "HEIGHT" THEN
      RETURN L_B_TT_pdf_image.image_h.
    ELSE IF pdfDim = "WIDTH" THEN
      RETURN L_B_TT_pdf_image.image_w.
    ELSE
      RETURN 0.
  END.
  ELSE
    RETURN 0.

END FUNCTION. /* pdf_ImageDim */

FUNCTION pdf_TextX RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND (FIRST TT_pdf_stream
                   WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_TextX","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "TextX"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.

END FUNCTION. /* pdf_TextX*/

FUNCTION pdf_TextY RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  FIND FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_TextY","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "TextY"
                            NO-LOCK NO-ERROR.

  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.

END FUNCTION. /* pdf_TextY */

FUNCTION ObjectSequence RETURNS INTEGER ( INPUT pdfStream     AS CHARACTER,
                                          INPUT pdfSequence   AS INTEGER,
                                          INPUT pdfObjectDesc AS CHARACTER,
                                          INPUT pdfPage       AS INTEGER ):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"ObjectSequence","Cannot find Stream!").
    RETURN ERROR.
  END.

  CREATE TT_pdf_object.
  ASSIGN TT_pdf_object.obj_stream = pdfStream
         TT_pdf_object.obj_nbr    = pdfSequence
         TT_pdf_object.obj_desc   = pdfObjectDesc
         TT_pdf_object.obj_offset = IF pdfSequence <> 0 THEN
                                      SEEK(S_pdf_inc) + 1
                                    ELSE 0
         TT_pdf_object.gen_nbr    = IF pdfSequence <> 0 THEN 0
                                    ELSE 65535
         TT_pdf_object.obj_type   = IF pdfSequence = 0 THEN "f"
                                    ELSE "n"
         TT_pdf_object.obj_page   = pdfPage.

  pdf_inc_ObjectSequence = pdfSequence.

  RETURN pdf_inc_ObjectSequence.

END FUNCTION. /* ObjectSequence */

PROCEDURE pdf_new :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFileName AS CHARACTER NO-UNDO.

  IF INDEX(pdfStream, " ") > 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_new","Cannot have a space in the Stream Name!").
    RETURN.
  END.

  CREATE TT_pdf_stream.
  ASSIGN TT_pdf_stream.obj_stream = pdfStream
         TT_pdf_stream.obj_file   = pdfFileName.

  RUN pdf_LoadBase14 (pdfStream).
  RUN pdf_init_param (pdfStream).

  RUN pdf_set_info(pdfStream,"OutputTo",pdfFileName).

END. /* pdf_new */

FUNCTION pdf_VerticalSpace RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_VerticalSpace","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "VerticalSpace"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 10.

END FUNCTION. /* pdf_VerticalSpace */

FUNCTION pdf_PointSize RETURN DECIMAL ( INPUT pdfStream AS CHARACTER ):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_PointSize","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "PointSize"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 10.0.

END FUNCTION. /* pdf_PointSize */

FUNCTION pdf_text_width RETURNS INTEGER ( INPUT pdfStream   AS CHARACTER,
                                          INPUT pdfText     AS CHARACTER):

  DEFINE VARIABLE L_width  AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_font    AS CHARACTER NO-UNDO.

  L_font = pdf_Font(pdfStream).

  DEFINE VARIABLE L_Loop    AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_tot     AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_width","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                           AND TT_pdf_font.font_name  = L_Font
                           NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_font THEN DO:
    IF TT_pdf_Font.font_type = "FIXED" THEN DO:
      L_width = INT((LENGTH(pdfText, "character":u) * INT(TT_pdf_font.font_width) / 1000) * pdf_PointSize(pdfStream)).
    end.
    ELSE DO:
      DO L_loop = 1 TO LENGTH(pdfText, "character":u):
        L_tot = L_tot + INT(ENTRY(INT(ASC(SUBSTR(pdfText,L_Loop,1, "character":u))) + 1 ,TT_pdf_Font.font_width, " ")) NO-ERROR.
      END.

      l_width = INT((L_tot / 1000) * pdf_PointSize(pdfStream)) NO-ERROR.
    END. /* Variable Width Font */
  END. /* Found the current font */

  RETURN l_width.
END FUNCTION. /* pdf_text_width */

FUNCTION pdf_text_widthdec RETURNS DECIMAL ( INPUT pdfStream   AS CHARACTER,
                                             INPUT pdfText     AS CHARACTER):

  DEFINE VARIABLE L_width   AS DECIMAL   NO-UNDO.
  DEFINE VARIABLE L_font    AS CHARACTER NO-UNDO.

  L_font = pdf_Font(pdfStream).

  DEFINE VARIABLE L_Loop    AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_tot     AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_widthdec","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                           AND TT_pdf_font.font_name  = L_Font
                           NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_font THEN DO:
    IF TT_pdf_Font.font_type = "FIXED" THEN DO:
      L_width = DEC((LENGTH(pdfText, "character":u) * INT(TT_pdf_font.font_width) / 1000) * pdf_PointSize(pdfStream)).
    END.
    ELSE DO:
      DO L_loop = 1 TO LENGTH(pdfText, "character":u):
        L_tot = L_tot + INT(ENTRY(INT(ASC(SUBSTR(pdfText,L_Loop,1, "character":u))) + 1 ,TT_pdf_Font.font_width, " ")) NO-ERROR.
      END.

      l_width = DEC((L_tot * pdf_PointSize(pdfStream)) / 1000) NO-ERROR.
    END. /* Variable Width Font */
  END. /* Found the current font */

  RETURN L_width.
END FUNCTION. /* pdf_text_widthDec */

PROCEDURE pdf_set_TextRed :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TextRed","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "TextRed" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "TextRed".
  END.

  TT_pdf_param.obj_value = dec2string(pdfValue).

END. /* pdf_set_TextRed */

PROCEDURE pdf_set_TextGreen :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TextGreen","Cannot find Stream!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "TextGreen" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "TextGreen".
  END.

  TT_pdf_param.obj_value = dec2string(pdfValue).

END. /* pdf_set_TextGreen */

PROCEDURE pdf_set_TextBlue :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TextBlue","Cannot find Stream!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "TextBlue" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "TextBlue".
  END.

  TT_pdf_param.obj_value = dec2string(pdfValue).

END. /* pdf_set_TextBlue */

FUNCTION pdf_TextRed RETURN DECIMAL ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_TextRed","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "TextRed"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.0.

END FUNCTION. /* pdf_TextRed */

FUNCTION pdf_TextGreen RETURN DECIMAL ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_TextGreen","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "TextGreen"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.0.

END FUNCTION. /* pdf_TextGreen */

FUNCTION pdf_TextBlue RETURN DECIMAL ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_TextBlue","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "TextBlue"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.0.

END FUNCTION. /* pdf_TextBlue */

PROCEDURE pdf_set_FillRed :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_FillRed","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "FillRed" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "FillRed".
  END.

  TT_pdf_param.obj_value = dec2string(pdfValue).

END. /* pdf_set_FillRed */

PROCEDURE pdf_set_FillGreen :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_FillGreen","Cannot find Stream!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "FillGreen" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "FillGreen".
  END.

  TT_pdf_param.obj_value = dec2string(pdfValue).

END. /* pdf_set_FillGreen */

PROCEDURE pdf_set_FillBlue :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_FillBlue","Cannot find Stream!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "FillBlue" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "FillBlue".
  END.

  TT_pdf_param.obj_value = dec2string(pdfValue).

END. /* pdf_set_FillBlue */

FUNCTION pdf_FillRed RETURN DECIMAL ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_FillRed","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "FillRed"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.0.

END FUNCTION. /* pdf_FillRed */

FUNCTION pdf_FillGreen RETURN DECIMAL ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_FillGreen","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "FillGreen"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.0.

END FUNCTION. /* pdf_FillGreen */

FUNCTION pdf_FillBlue RETURN DECIMAL ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_FillBlue","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "FillBlue"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN string2dec(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.0.

END FUNCTION. /* pdf_FillBlue */

PROCEDURE pdf_set_PageWidth :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PageWidth","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PageWidth","Page Width cannot be zero!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "PageWidth" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "PageWidth".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_PageWidth */

PROCEDURE pdf_set_PageHeight :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PageHeight","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PageHeight","Page Height cannot be zero!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "PageHeight" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "PageHeight".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_PageHeight */


PROCEDURE pdf_set_Page :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_page","Cannot find Stream!").
    RETURN.
  END.

  IF pdfValue = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_page","Value passed cannot be zero!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "Page" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "Page".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_Page */

PROCEDURE pdf_set_PageRotate :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PageRotate","Cannot find Stream!").
    RETURN .
  END.

  IF  pdfValue <> 0
  AND pdfValue <> 90
  AND pdfValue <> 180
  AND pdfValue <> 270 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PageRotate","Page Rotation value must be 0, 90, 180, or 270!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "PageRotate" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "PageRotate".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_PageRotate */

FUNCTION pdf_Page RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  FIND FIRST TT_pdf_stream
       WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_Page","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "Page"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.

END FUNCTION. /* pdf_Page */

FUNCTION pdf_PageWidth RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_PageWidth","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "PageWidth"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 612.

END FUNCTION. /* pdf_PageWidth */

FUNCTION pdf_Pageheight RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_PageHeight","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "PageHeight"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 792.

END FUNCTION. /* pdf_PageHeight */

FUNCTION pdf_PageRotate RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_PageRotate","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "PageRotate"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.

END FUNCTION. /* pdf_PageRotate */

FUNCTION pdf_Angle RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_Angle","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "Angle"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN RETURN INT(TT_pdf_param.obj_value).
  ELSE RETURN ?.

END. /* pdf_Angle */

PROCEDURE pdf_set_TextX :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  DEFINE BUFFER L_TT_pdf_param FOR TT_pdf_param.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TextX","Cannot find Stream (" + pdfStream + ")!").
    RETURN.
  END.

  IF pdfValue < 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TextX","Value cannot be less than zero!").
    RETURN .
  END.

  FIND FIRST L_TT_pdf_param
       WHERE L_TT_pdf_param.obj_stream    = pdfStream
         AND L_TT_pdf_param.obj_parameter = "TextX" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL L_TT_pdf_param THEN DO:
    CREATE L_TT_pdf_param.
    ASSIGN L_TT_pdf_param.obj_stream    = pdfStream
           L_TT_pdf_param.obj_parameter = "TextX".
  END.

  L_TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_TextX */

FUNCTION pdf_TopMargin RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_TopMargin","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "TopMargin"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 50.

END FUNCTION. /* pdf_TopMargin */

FUNCTION pdf_BottomMargin RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_BottomMargin","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "BottomMargin"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 50.

END FUNCTION. /* pdf_BottomMargin */

PROCEDURE pdf_set_TextY :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  DEFINE BUFFER L_TT_pdf_param FOR TT_pdf_param.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TextY","Cannot find Stream!").
    RETURN.
  END.

  IF pdfValue <= pdf_BottomMargin(pdfStream)
  AND NOT pdf_ForFooter
  THEN DO:
    RUN pdf_new_page(pdfStream).
    pdfValue = pdf_TextY(pdfStream).
  END.

  FIND FIRST L_TT_pdf_param
       WHERE L_TT_pdf_param.obj_stream    = pdfStream
         AND L_TT_pdf_param.obj_parameter = "TextY" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL L_TT_pdf_param THEN DO:
    CREATE L_TT_pdf_param.
    ASSIGN L_TT_pdf_param.obj_stream    = pdfStream
           L_TT_pdf_param.obj_parameter = "TextY".
  END.

  CASE pdf_Angle(pdfStream):
    WHEN 90 THEN
      L_TT_pdf_param.obj_value = STRING(pdf_textY(pdfStream) - pdfValue).
    WHEN 270 THEN
      L_TT_pdf_param.obj_value = STRING(pdf_textY(pdfStream) + pdfValue).
    OTHERWISE
      L_TT_pdf_param.obj_value = STRING(pdfValue).
  END CASE.

END. /* pdf_set_TextY */


PROCEDURE pdf_set_GraphicX :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_GraphicX","Cannot find Stream!").
    RETURN.
  END.

  IF pdfValue < 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_GraphicX","Value cannot be less than 0!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "GraphicX" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "GraphicX".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_GraphicX */

PROCEDURE pdf_set_GraphicY :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_GraphicY","Cannot find Stream!").
    RETURN.
  END.

  IF pdfValue < 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_GraphicY","Value cannot be less than zero!").
    RETURN.
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "GraphicY" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "GraphicY".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_GraphicY */

FUNCTION pdf_GraphicX RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_GraphicX","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "GraphicX"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.

END FUNCTION. /* pdf_GraphicX */

FUNCTION pdf_GraphicY RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_GraphicY","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "GraphicY"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 0.

END. /* pdf_GraphicY */

PROCEDURE pdf_set_info :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfAttribute AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfvalue     AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_info","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_option  AS CHARACTER NO-UNDO.
  L_Option = "Author,Creator,Producer,Keywords,Subject,Title,OutputTo".

  IF LOOKUP(pdfAttribute,L_option) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_info","Invalid Attribute passed!").
    RETURN .
  END.

  IF NOT CAN-FIND( FIRST TT_pdf_info
                   WHERE TT_pdf_info.obj_stream = pdfStream
                     AND TT_pdf_info.info_attr  = pdfAttribute NO-LOCK)
  THEN DO:
    CREATE TT_pdf_info.
    ASSIGN TT_pdf_info.obj_stream = pdfStream
           TT_pdf_info.info_attr  = pdfAttribute.
  END.

  RUN pdf_replace_text(INPUT-OUTPUT pdfValue).

  TT_pdf_info.info_value = pdfValue.

END. /* pdf_set_info */

FUNCTION pdf_get_info RETURNS CHARACTER ( INPUT pdfStream    AS CHARACTER,
                                          INPUT pdfAttribute AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_get_info","Cannot find Stream!").
    RETURN ERROR.
  END.

  DEFINE VARIABLE L_option  AS CHARACTER NO-UNDO.
  L_Option = "Author,Creator,Producer,Keywords,Subject,Title,OutputTo".

  IF LOOKUP(pdfAttribute,L_option) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_Get_Info","Invalid Attribute passed!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_info WHERE TT_pdf_info.obj_stream = pdfStream
                           AND TT_pdf_info.info_attr  = pdfAttribute
                           NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_info THEN DO:
    RETURN TT_pdf_info.info_value.
  END.

  ELSE
    RETURN "".
END FUNCTION. /* pdf_get_info */

FUNCTION pdf_LeftMargin RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_LeftMargin","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "LeftMargin"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN
    RETURN INT(TT_pdf_param.obj_value).
  ELSE
    RETURN 10.

END FUNCTION. /* pdf_LeftMargin */


PROCEDURE pdf_move_to :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToX       AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToY       AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_move_to","Cannot find Stream!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC"
         TT_pdf_content.obj_content = STRING(pdfToX) + " " + STRING(pdfToY) + " m" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_GraphicY (pdfStream,pdfToY).
  RUN pdf_set_GraphicX (pdfStream, pdfToX).

END. /* pdf_moveto */

PROCEDURE pdf_rect :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromX     AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromY     AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfWidth     AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfHeight    AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfWeight    AS DECIMAL NO-UNDO.    /* JES ADDED */

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_rect","Cannot find Stream!").
    RETURN.
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC".
         TT_pdf_content.obj_content = dec2string(pdfWeight).
         TT_pdf_content.obj_content = TT_pdf_content.obj_content
                                    + " w" + CHR(13)  /* JES ADDED */
                                    + STRING(pdfFromX) + " " + STRING(pdfFromY) + " "
                                    + STRING(pdfWidth) + " " + STRING(pdfHeight)
                                    + " re" + CHR(13)
                                    + "B" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_GraphicY (pdfStream,pdfFromY + pdfHeight).
  RUN pdf_set_GraphicX (pdfStream, pdfFromX + pdfWidth).

END. /* pdf_rect */

PROCEDURE pdf_circle :
  /* Note:  pdfX and pdfY represent the center point of the circle.  These
            values become the new Graphic X and Y points after the drawing of
            the circle.  If you want the circle to be filled use pdf_stroke_fill
  */
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfX         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfY         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfRadius    AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfWeight     AS DECIMAL   NO-UNDO.

  DEFINE VARIABLE l_Constant    AS DECIMAL DECIMALS 10 NO-UNDO.
  DEFINE VARIABLE l_Length      AS DECIMAL DECIMALS 5 NO-UNDO.
  DEFINE VARIABLE c_NewY        AS CHARACTER NO-UNDO.
  DEFINE VARIABLE c_NewX        AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_circle","Cannot find Stream!").
    RETURN.
  END.

  l_Constant = 0.5522847498. /* (4/3)*(sqrt(2)-1) - who knows where this calc came from */

  l_Length = pdfRadius * l_Constant.

  /* First Quadrant - Upper Right */
  c_NewX   = dec2string(l_Length + pdfX).
  c_NewY   = dec2string(l_Length + pdfY).
  RUN pdf_move_to(pdfStream, pdfX + pdfRadius, pdfY).

  RUN pdf_curve(pdfStream,
                pdfX + pdfRadius,
                c_newY,
                c_NewX,
                pdfY + pdfRadius,
                pdfX,
                pdfY + pdfRadius,
                pdfWeight).

  /* Second Quadrant - Upper Left */
  c_NewX   = dec2string(pdf_GraphicX(pdfstream) - l_Length).
  c_NewY   = dec2string(pdf_GraphicY(pdfstream) - pdfRadius + l_Length).
  RUN pdf_curve(pdfStream,
                c_NewX,
                pdf_GraphicY(pdfStream),
                pdfX - pdfRadius,
                c_NewY,
                pdfX - pdfRadius,
                pdf_GraphicY(pdfStream) - pdfRadius,
                pdfWeight).

  /* Third Quadrant - Lower Left */
  c_NewX   = dec2string(pdfX - l_Length).
  c_NewY   = dec2string(pdfY - l_Length).
  RUN pdf_curve(pdfStream,
                pdfX - pdfRadius,
                c_NewY,
                c_newX,
                pdfY - pdfRadius,
                pdfX,
                pdfY - pdfRadius,
                pdfWeight).

  /* Fourth Quadrant - Lower Right */
  c_NewX   = dec2string(pdfX + l_Length).
  c_NewY   = dec2string(pdfY - l_Length).
  RUN pdf_curve(pdfStream,
                c_NewX,
                pdfY - pdfRadius,
                pdfX + pdfRadius,
                c_NewY,
                pdfX + pdfRadius,
                pdfY,
                pdfWeight).

  /* Close the Path */
  RUN pdf_close_path(pdfStream).

  /* Set the current point to be the current Graphic X/Y Location */
  RUN pdf_set_GraphicY (pdfStream,pdfY).
  RUN pdf_set_GraphicX (pdfStream, pdfX).
  

END. /* pdf_circle */

PROCEDURE pdf_curve :
  /* A Bzier curve is added from the current Graphic X/Y Location to X3/Y3 
     using X1/Y1 and X2/Y2) as the control points. The X3/Y3 of the curve 
     becomes the new Graphic X/Y Location.  */
     
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfX1         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfY1         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfX2         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfY2         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfX3         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfY3         AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfWeight     AS DECIMAL   NO-UNDO.

  DEFINE VARIABLE l_Constant    AS DECIMAL DECIMALS 10 NO-UNDO.
  DEFINE VARIABLE l_Length      AS DECIMAL DECIMALS 5 NO-UNDO.
  DEFINE VARIABLE c_NewY        AS CHARACTER NO-UNDO.
  DEFINE VARIABLE c_NewX        AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_curve","Cannot find Stream!").
    RETURN.
  END.

  IF pdf_GraphicY(pdfStream) = 0 OR pdf_GraphicX(pdfStream) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_curve","Graphic X/Y location has not been initialized!").
    RETURN.
  END.

  /* First Circle - Upper Right */
  c_NewX   = dec2string(l_Length + pdfX1).
  c_NewY   = dec2string(l_Length + pdfY1).
  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC".
         TT_pdf_content.obj_content = CHR(13) + dec2string(pdfWeight).
         TT_pdf_content.obj_content = TT_pdf_content.obj_content
                                    + " w" + CHR(13)
                                    + STRING(pdf_FillRed(pdfstream)) + " "
                                    + STRING(pdf_FillGreen(pdfstream)) + " "
                                    + STRING(pdf_FillBlue(pdfstream)) + " rg" + CHR(13)

                                    + STRING(pdfX1) + " " + STRING(pdfY1)
                                    + " " + STRING(pdfX2) + " " + STRING(pdfY2) + " "
                                    + STRING(pdfX3) + " " + STRING(pdfY3) + " c" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  /* Set the new Graphic Points */
  RUN pdf_set_GraphicY (pdfStream, pdfY3).
  RUN pdf_set_GraphicX (pdfStream, pdfX3).
  
END. /* pdf_curve */

PROCEDURE pdf_close_path:
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC".
         TT_pdf_content.obj_content = CHR(13) + "B" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).
END.

PROCEDURE pdf_stroke_color :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRed       AS DECIMAL NO-UNDO.
  DEFINE INPUT PARAMETER pdfGreen     AS DECIMAL NO-UNDO.
  DEFINE INPUT PARAMETER pdfBlue      AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_stroke_color","Cannot find Stream!").
    RETURN .
  END.

  pdfRed   = IF pdfRed < 0 THEN 0
             ELSE IF pdfRed > 1 THEN 1
             ELSE pdfRed.
  pdfGreen = IF pdfGreen < 0 THEN 0
             ELSE IF pdfGreen > 1 THEN 1
             ELSE pdfGreen.
  pdfBlue  = IF pdfBlue < 0 THEN 0
             ELSE IF pdfBlue > 1 THEN 1
             ELSE pdfBlue.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC".
         TT_pdf_content.obj_content = " " + dec2string(pdfRed)
                                    + " " + dec2string(pdfGreen)
                                    + " " + dec2string(pdfBlue) + " RG ".
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

END. /* pdf_stroke_color */

PROCEDURE pdf_stroke_fill :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRed       AS DECIMAL NO-UNDO.
  DEFINE INPUT PARAMETER pdfGreen     AS DECIMAL NO-UNDO.
  DEFINE INPUT PARAMETER pdfBlue      AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_stroke_fill","Cannot find Stream!").
    RETURN .
  END.

  pdfRed   = IF pdfRed < 0 THEN 0
             ELSE IF pdfRed > 1 THEN 1
             ELSE pdfRed.
  pdfGreen = IF pdfGreen < 0 THEN 0
             ELSE IF pdfGreen > 1 THEN 1
             ELSE pdfGreen.
  pdfBlue  = IF pdfBlue < 0 THEN 0
             ELSE IF pdfBlue > 1 THEN 1
             ELSE pdfBlue.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC".
         TT_pdf_content.obj_content = " " + dec2string(pdfRed)
                                    + " " + dec2string(pdfGreen)
                                    + " " + dec2string(pdfBlue) + " rg ".
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_FillRed(pdfStream, pdfRed).
  RUN pdf_set_FillGreen(pdfStream, pdfGreen).
  RUN pdf_set_FillBlue(pdfStream, pdfBlue).

END. /* pdf_stroke_fill */

PROCEDURE pdf_set_dash :
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfOn      AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfOff     AS INTEGER NO-UNDO.

  IF pdfOn  < 0 THEN 1.
  IF pdfOff < 0 THEN 1.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_dash","Cannot find Stream!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC"
         TT_pdf_content.obj_content = " [" + STRING(pdfOn) + " " + STRING(pdfOff)
                                    + "] 0  d ".
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

END. /* pdf_set_dash */

PROCEDURE pdf_set_linejoin :
  /* Note:  This procedure allows you to define the Line Join Styles.  This will
            typically be used when drawing a Rectangle. Possible values are:

              0 - Miter Join
              1 - Round Join
              2 - Bevel Join
  */
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfJoin    AS INTEGER NO-UNDO.

  IF pdfJoin  < 0 OR pdfJoin > 2 THEN 0.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_linejoin","Cannot find Stream!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC"
         TT_pdf_content.obj_content = STRING(pdfJoin) + " j" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

END. /* pdf_set_linejoin */

PROCEDURE pdf_line :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromX     AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromY     AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToX       AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToY       AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfWeight    AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_line","Cannot find Stream!").
    RETURN.
  END.

  DEFINE VARIABLE L_FromY   AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_ToY     AS INTEGER NO-UNDO.

  ASSIGN L_FromY =  /* pdf_PageHeight - */ (pdfFromY) /*(pdfFromY + pdf_TopMargin) * -1 */
         L_ToY   =  /* pdf_PageHeight - */ (pdfToY ) /* (pdfToY + pdf_TopMargin) * -1 */.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC".
         TT_pdf_content.obj_content = dec2string(pdfWeight) + " w" + CHR(13)
                                    + STRING(pdfFromX) + " " + STRING(L_FromY) + " m" + CHR(13)
                                    + STRING(pdfToX) + " " + STRING(L_ToY) + " l" + CHR(13)
                                    + "S" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_GraphicX(pdfStream, pdfToX ).
  RUN pdf_set_GraphicY(pdfstream, pdfToY).

END. /* pdf_line */

PROCEDURE pdf_watermark:
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFont     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfSize     AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfRed      AS DECIMAL   NO-UNDO.
  DEFINE INPUT PARAMETER pdfGreen    AS DECIMAL   NO-UNDO.
  DEFINE INPUT PARAMETER pdfBlue     AS DECIMAL   NO-UNDO.
  DEFINE INPUT PARAMETER pdfX        AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfY        AS INTEGER   NO-UNDO.

  DEFINE VARIABLE L_Font      AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_PointSize AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_watermark","Cannot find Stream!").
    RETURN .
  END.

  RUN pdf_replace_text (INPUT-OUTPUT pdfText).

  pdfRed   = IF pdfRed < 0 THEN 0
             ELSE IF pdfRed > 1 THEN 1
             ELSE pdfRed.
  pdfGreen = IF pdfGreen < 0 THEN 0
             ELSE IF pdfGreen > 1 THEN 1
             ELSE pdfGreen.
  pdfBlue  = IF pdfBlue < 0 THEN 0
             ELSE IF pdfBlue > 1 THEN 1
             ELSE pdfBlue.

  FIND FIRST TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                           AND TT_pdf_font.font_name  = pdfFont
                           NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_font THEN DO:
    RUN pdf_error(pdfStream,"pdf_watermark","Font has not been loaded!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "WATERMARK".
         TT_pdf_content.obj_content = "BT" + CHR(13)
                                    + STRING(TT_pdf_font.font_tag) + " " + STRING(pdfSize) +  " Tf" + CHR(13)
                                    + "1 0 0 1 " + STRING(pdfX) +  " "
                                                 + STRING(pdfY) + " Tm" + CHR(13)
                                    + " " + dec2string(pdfRed) + " "
                                          + dec2string(pdfGreen) + " "
                                          + dec2string(pdfBlue) + " rg" + CHR(13)
                                    + "T* (" + pdfText  + ") Tj" + CHR(13)
                                    + "ET" + CHR(13).
         TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

END. /* pdf_watermark */

PROCEDURE pdf_text :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Width       AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text","Cannot find Stream!").
    RETURN .
  END.

  RUN pdf_replace_text (INPUT-OUTPUT pdfText).

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXT"
         TT_pdf_content.obj_content = "(" + pdfText + ") Tj".
         TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  /* Determine whether the TotalPages Function as called within this text
     function ... if so then indicate in the the PDF Content */
  IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
    TT_pdf_content.obj_function = TRUE.

  L_Width = pdf_text_width (pdfStream, pdfText).

  CASE pdf_Angle(pdfStream):
    WHEN 0 THEN
      RUN pdf_set_TextX( pdfStream, pdf_TextX(pdfStream) + L_width ).
    WHEN 180 THEN
      RUN pdf_set_TextX( pdfStream, pdf_TextX(pdfStream) - L_width ).
    WHEN 90 OR WHEN 270 THEN
      RUN pdf_set_TextY( pdfStream, L_width ).
  END CASE.

END. /* pdf_text */

PROCEDURE pdf_text_char :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue    AS INTEGER NO-UNDO.

  DEFINE VARIABLE L_Width       AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_char","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue <= 0 OR pdfValue > 377 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_char","Value must be >= 1 and <= 377!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXT"
         TT_pdf_content.obj_content = "(~\" + STRING(pdfValue) + ") Tj".
         TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  /* Increase by 1 Character Length */
  L_Width = pdf_text_width (pdfStream, "1").

  CASE pdf_Angle(pdfStream):
    WHEN 0 THEN
      RUN pdf_set_TextX( pdfStream, pdf_TextX(pdfStream) + L_width ).
    WHEN 180 THEN
      RUN pdf_set_TextX( pdfStream, pdf_TextX(pdfStream) - L_width ).
    WHEN 90 OR WHEN 270 THEN
      RUN pdf_set_TextY( pdfStream, L_width ).
  END CASE.

END. /* pdf_text_char */

PROCEDURE pdf_text_charxy :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn   AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRow      AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_charxy","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_color   AS CHARACTER NO-UNDO.
  L_color = dec2string(pdf_TextRed(pdfStream)) + " "
          + dec2string(pdf_TextGreen(pdfStream)) + " "
          + dec2string(pdf_TextBlue(pdfStream)) + " rg" + CHR(13).

  DEFINE VARIABLE L_PointSize   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.
  L_PointSize  = dec2string(pdf_PointSize(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  DEFINE VARIABLE L_Font        AS CHARACTER NO-UNDO.
  L_font = pdf_Font(pdfStream).

  FIND FIRST TT_pdf_font
       WHERE TT_pdf_font.obj_stream = pdfStream
         AND TT_pdf_font.font_name  = L_Font NO-LOCK NO-ERROR.
  L_Font = IF AVAIL TT_pdf_font THEN TT_pdf_Font.font_tag ELSE "/BF1".

  IF pdfColumn = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_charxy","Column cannot be zero!").
    RETURN .
  END.

  IF pdfRow    = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_charxy","Row cannot be zero!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXTXY"
         TT_pdf_content.obj_line    = pdfRow
         TT_pdf_content.obj_content = CHR(13) + L_Font + " " + L_PointSize + " Tf" + CHR(13)
                                    + L_color
                                    + "1 0 0 1 " + STRING(pdfColumn) + " "
                                    + STRING(pdfRow) + " Tm" + CHR(13)
                                    + "(~\" + STRING(pdfText) + ") Tj" + CHR(13).
         TT_pdf_content.obj_length  = LENGTH(pdfText, "character":u) .
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

END. /* pdf_text_charxy */

PROCEDURE pdf_text_rotate :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfAngle    AS INTEGER NO-UNDO.

  DEFINE VARIABLE L_Rotate           AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_rotate","Cannot find Stream!").
    RETURN .
  END.

  CASE pdfAngle:
    WHEN 0 THEN
      L_Rotate = "1 0 0 1 ".
    WHEN 90 THEN
      L_Rotate = "0 -1 1 0 ".
    WHEN 180 THEN
      L_Rotate = "-1 0 0 -1 ".
    WHEN 270 THEN
      L_Rotate = "0 1 -1 0 ".
  END CASE.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXT".
         TT_pdf_content.obj_content = L_Rotate + STRING(pdf_textX(pdfStream))
                                    + " " + STRING(pdf_TextY(pdfStream)) + " Tm".
  ASSIGN TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_angle(pdfstream, pdfAngle).

END. /* pdf_text_rotate */

FUNCTION pdf_GetNumFittingChars RETURNS INTEGER
                                ( INPUT pdfStream   AS CHARACTER,
                                  INPUT pdfText     AS CHARACTER,
                                  INPUT pdfFromX    AS INTEGER,
                                  INPUT pdfToX      AS INTEGER ):

  DEFINE VARIABLE L_font    AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_tot     AS INTEGER NO-UNDO.

  DEFINE VARIABLE iReqFMWidth AS INTEGER NO-UNDO. /* Required font-metric width */
  DEFINE VARIABLE iCurFMWidth AS DECIMAL NO-UNDO. /* Font-metric width of chars that fit */
  DEFINE VARIABLE iCurChar    AS INTEGER NO-UNDO. /* Char index up to */

  L_font = pdf_Font(pdfStream).

  iReqFMWidth = pdfToX - pdfFromX.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_getnumFittingChars","Cannot find Stream!").
    RETURN ERROR.
  END.

  iCurChar = 0.
  iCurFMWidth = 0.
  FIND FIRST TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                           AND TT_pdf_font.font_name  = L_Font
                           NO-LOCK NO-ERROR.
  IF AVAILABLE TT_pdf_font THEN
  DO iCurChar = 1 TO LENGTH(pdfText, "character":u):

    /* Keep looping until width is too wide, then return one less char */
    iCurFMWidth = iCurFMWidth +
                   (IF TT_pdf_Font.font_type = "FIXED" THEN
                     INTEGER(TT_pdf_font.font_width) / 1000 * pdf_PointSize(pdfStream)
                   ELSE
                     DECIMAL(INTEGER(ENTRY(ASC(SUBSTR(pdfText, iCurChar, 1, "character":u)) + 1,  /* ASCII Value of char at cur pos */
                                              TT_pdf_Font.font_width,             /* Space-seperated values list widths */
                                              " "))) / 1000 * pdf_PointSize(pdfStream)) NO-ERROR.

    IF iCurFMWidth > iReqFMWidth THEN
      RETURN iCurChar - 1.
  END. /* Loop through text */

  RETURN iCurChar.

END FUNCTION. /* pdf_GetNumFittingChars */

PROCEDURE pdf_set_font :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFont     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfSize     AS DECIMAL NO-UNDO.

  DEFINE BUFFER L_B_TT_pdf_content FOR TT_pdf_Content.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_font","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_PointSize   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.
  L_PointSize  = dec2string(pdf_PointSize(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  FIND FIRST TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                           AND TT_pdf_font.font_name = pdfFont
                           NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_font THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_font","Font (" + pdfFont + ")has not been loaded!").
    RETURN .
  END.

  FIND FIRST B_TT_pdf_content
       WHERE B_TT_pdf_content.obj_stream = pdfStream
         AND B_TT_pdf_content.obj_page   = pdf_Page(pdfStream)
         AND B_TT_pdf_content.obj_line   = pdf_TextY(pdfStream) NO-ERROR.
  IF AVAIL B_TT_pdf_content THEN DO:
    ASSIGN B_TT_pdf_content.obj_content = B_TT_pdf_content.obj_content
                                        + CHR(13) + TT_pdf_font.font_tag + " "
                                        + dec2string(pdfSize) +  " Tf" + CHR(13).
  END.

  ELSE DO:
    CREATE L_B_TT_pdf_content.
    ASSIGN L_B_TT_pdf_content.obj_stream  = pdfStream.
           L_B_TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
           L_B_TT_pdf_content.obj_type    = "TEXTFONT".
           L_B_TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
           L_B_TT_pdf_content.obj_page    = pdf_Page(pdfStream).
           L_B_TT_pdf_content.obj_content = TT_pdf_font.font_tag + " "
                                          + dec2string(pdfSize) +  " Tf".

  END.

  /* Set the Stream Font Parameter */
  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "Font" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "Font".
  END.

  TT_pdf_param.obj_value = pdfFont.

  /* Set the Stream Font Size */
  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "PointSize" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "PointSize".
  END.
  TT_pdf_param.obj_value = dec2string(pdfSize).

END. /* pdf_set_font */

PROCEDURE pdf_text_render :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRender   AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_render","Cannot find Stream!").
    RETURN .
  END.

  IF pdfRender < 0 OR pdfRender > 3 THEN pdfRender = 0.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXTRENDER"
         TT_pdf_content.obj_content = STRING(pdfRender) +  " Tr".
         TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "Render" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "Render".
  END.

  TT_pdf_param.obj_value = STRING(pdfRender).

END. /* pdf_text_render */

PROCEDURE pdf_text_color :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRed      AS DECIMAL NO-UNDO.
  DEFINE INPUT PARAMETER pdfGreen    AS DECIMAL NO-UNDO.
  DEFINE INPUT PARAMETER pdfBlue     AS DECIMAL NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_color","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_PointSize   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.
  L_PointSize  = dec2string(pdf_PointSize(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  FIND FIRST B_TT_pdf_content
       WHERE B_TT_pdf_content.obj_stream = pdfStream
         AND B_TT_pdf_content.obj_page   = pdf_Page(pdfStream)
         AND B_TT_pdf_content.obj_line   = pdf_TextY(pdfStream) NO-ERROR.
  IF AVAIL B_TT_pdf_content THEN DO:
    ASSIGN B_TT_pdf_content.obj_content = B_TT_pdf_content.obj_content + CHR(13)
                                        + dec2string(pdfRed) + " "
                                        + dec2string(pdfGreen) + " "
                                        + dec2string(pdfBlue) + " rg ".
  END.
  ELSE DO:
    CREATE TT_pdf_content.
    ASSIGN TT_pdf_content.obj_stream  = pdfStream.
           TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
    ASSIGN TT_pdf_content.obj_type    = "TEXTCOLOR".
           TT_pdf_content.obj_line    = pdf_TextY(pdfStream) .
           TT_pdf_content.obj_page    = pdf_Page(pdfStream).
           TT_pdf_content.obj_content = dec2string(pdfRed) + " "
                                      + dec2string(pdfGreen) + " "
                                      + dec2string(pdfBlue) + " rg ".
  END.

  RUN pdf_set_TextRed(pdfStream, pdfRed).
  RUN pdf_set_TextGreen(pdfStream, pdfGreen).
  RUN pdf_set_TextBlue(pdfStream, pdfBlue).

END. /* pdf_text_color */

PROCEDURE pdf_load_font :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFontName AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFontFile AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFontAFM  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFontDIF  AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_load_font","Cannot find Stream!").
    RETURN .
  END.

  IF INDEX(pdfFontName," ") > 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_load_font","Font Name cannot contains spaces!").
    RETURN .
  END.

  IF SEARCH(pdfFontFile) = ? THEN DO:
    RUN pdf_error(pdfStream,"pdf_load_font","Cannot find Font File for Loading!").
    RETURN .
  END.

  IF SEARCH(pdfFontAFM) = ? THEN DO:
    RUN pdf_error(pdfStream,"pdf_Load_Font","Cannot find Font AFM file for loading!").
    RETURN .
  END.

  IF pdfFontDIF <> "" AND SEARCH(pdfFontDIF) = ? THEN DO:
    RUN pdf_error(pdfStream,"pdf_Load_Font","Cannot find Font DIF file for loading!").
    RETURN .
  END.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.obj_stream  = pdfStream
         TT_pdf_font.font_name   = pdfFontName
         TT_pdf_font.font_file   = pdfFontFile
         TT_pdf_font.font_afm    = pdfFontAFM
         TT_pdf_font.font_dif    = pdfFontDIF.
  TT_pdf_font.font_tag    = "/" + STRING(TT_pdf_font.font_name).

  RUN pdf_ParseAFMFile
      (OUTPUT TT_pdf_font.afm_ItalicAngle,
       OUTPUT TT_pdf_font.afm_Ascender,
       OUTPUT TT_pdf_font.afm_Descender,
       OUTPUT TT_pdf_font.afm_FontBBox,
       OUTPUT TT_pdf_font.afm_FirstChar,
       OUTPUT TT_pdf_font.afm_LastChar,
       OUTPUT TT_pdf_font.afm_Widths,
       OUTPUT TT_pdf_font.afm_IsFixedPitch,
       OUTPUT TT_pdf_font.afm_Flags) NO-ERROR.

   TT_pdf_font.font_type  = IF TT_pdf_font.afm_IsFixedPitch = "0" THEN "FIXED"
                            ELSE "VARIABLE".

    /* igc - Added March 29, 2003
           - need to assign l_afm_width to TT_pdf_font.font_width so that we
             can find the actual width when using pdf_text_width */
   IF TT_pdf_font.font_type = "VARIABLE" THEN
     TT_pdf_font.font_width = TRIM(TT_pdf_font.afm_widths).
   ELSE
     TT_pdf_font.font_width = ENTRY(1, TRIM(TT_pdf_font.afm_widths)," ").

END. /* pdf_load_font */

PROCEDURE pdf_font_diff:
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFontName AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfCharNum  AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfPSName   AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_font_diff","Cannot find Stream!").
    RETURN .
  END.

  IF NOT CAN-FIND(FIRST TT_pdf_font
                  WHERE TT_pdf_font.obj_stream = pdfStream
                    AND TT_pdf_font.font_name  = pdfFontName NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_font_diff","Cannot find Font Name = " + pdfFontName).
    RETURN .
  END.

  CREATE TT_pdf_diff.
  ASSIGN TT_pdf_diff.obj_stream = pdfStream
         TT_pdf_diff.font_name  = pdfFontName
         TT_pdf_diff.char_num   = pdfCharNum
         TT_pdf_diff.PS_Name    = IF pdfPSName BEGINS "/" THEN pdfPSName
                                  ELSE "/" + pdfPSName.

END. /* pdf_font_diff */

PROCEDURE pdf_load_image :
  DEFINE INPUT PARAMETER pdfStream      AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfImageName   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfImageFile   AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_load_image","Cannot find Stream!").
    RETURN .
  END.

  IF INDEX(pdfImageName," ") > 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_load_image","Image Name cannot contain spaces!").
    RETURN .
  END.

  IF SEARCH(pdfImageFile) = ? THEN DO:
    RUN pdf_error(pdfStream,"pdf_load_image","Cannot find Image File when Loading!").
    RETURN .
  END.

  RUN pdf_get_image_wh (INPUT pdfStream,
                        INPUT pdfImageFile).


  CREATE TT_pdf_image.
  ASSIGN TT_pdf_image.obj_stream    = pdfStream
         TT_pdf_image.image_name    = pdfImageName
         TT_pdf_image.image_file    = pdfImageFile
         TT_pdf_image.image_h       = pdf_Height
         TT_pdf_image.image_w       = pdf_Width.
  TT_pdf_image.image_tag    = "/Im" + STRING(TT_pdf_image.image_name).

END. /* pdf_load_image */

PROCEDURE pdf_place_image :
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfImageName AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn    AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRow       AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfWidth     AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfHeight    AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_place_image","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_PageHeight  AS INTEGER NO-UNDO.
  L_PageHeight  = pdf_PageHeight(pdfStream).

  FIND FIRST TT_pdf_image
       WHERE TT_pdf_image.obj_stream = pdfStream
         AND TT_pdf_image.image_name = pdfImageName NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_image THEN DO:
    RUN pdf_error(pdfStream,"pdf_place_image","Cannot find Image Name for Placement!").
    RETURN .
  END.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "IMAGE"
         TT_pdf_content.obj_content = STRING(pdfWidth) + " 0 0 " + STRING(pdfHeight)
                                    + " " + STRING(pdfColumn) + " "
                                    + STRING(L_PageHeight - pdfRow) + " cm "
                                    + TT_pdf_image.image_tag + " Do" + CHR(13).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_GraphicX(pdfStream,pdfColumn).
  RUN pdf_set_GraphicY(pdfStream,pdfRow).

END. /* pdf_place_image */

PROCEDURE pdf_skip :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Spacer  AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_skip","Cannot find Stream!").
    RETURN .
  END.

  L_Spacer = INT(pdf_get_parameter(pdfStream,"LineSpacer")) NO-ERROR.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXTSKIP".
         TT_pdf_content.obj_content = "0 -" + dec2string(pdf_PointSize(pdfstream) + L_Spacer)
                                    + " TD" /* "T* */.
         TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_TextX (pdfStream, pdf_LeftMargin(pdfStream)).
  RUN pdf_set_TextY (pdfStream, pdf_TextY(pdfStream) - (pdf_PointSize(pdfStream) + L_Spacer)).

END. /* pdf_skip */

PROCEDURE pdf_skipn :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfNumber   AS INTEGER NO-UNDO.

  DEFINE VARIABLE l_Ctr              AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_skipn","Cannot find Stream!").
    RETURN .
  END.

  IF pdfNumber <= 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_skipn","Lines to skip cannot be <= zero!").
    RETURN .
  END.

  DO l_Ctr = 1 TO pdfNumber:
    RUN pdf_skip(pdfStream).
  END.

END. /* pdf_skipn */

PROCEDURE pdf_text_xy :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn   AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRow      AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_xy","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_color   AS CHARACTER NO-UNDO.
  L_color = dec2string(pdf_TextRed(pdfStream)) + " "
          + dec2string(pdf_TextGreen(pdfStream)) + " "
          + dec2string(pdf_TextBlue(pdfStream)) + " rg" + CHR(13).

  DEFINE VARIABLE L_PointSize   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.
  L_PointSize  = dec2string(pdf_PointSize(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  DEFINE VARIABLE L_Font        AS CHARACTER NO-UNDO.
  L_font = pdf_Font(pdfStream).

  FIND FIRST TT_pdf_font
       WHERE TT_pdf_font.obj_stream = pdfStream
         AND TT_pdf_font.font_name  = L_Font NO-LOCK NO-ERROR.
  L_Font = IF AVAIL TT_pdf_font THEN TT_pdf_Font.font_tag ELSE "/BF1".

  DEFINE VARIABLE L_orig_text   AS CHARACTER NO-UNDO.
  L_orig_text = pdfText.

  IF pdfColumn = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_xy","Column cannot be zero!").
    RETURN .
  END.

  IF pdfRow    = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_xy","Row cannot be zero!").
    RETURN .
  END.

  RUN pdf_replace_text (INPUT-OUTPUT pdfText).

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXTXY"
         TT_pdf_content.obj_line    = pdfRow
         TT_pdf_content.obj_content = CHR(13) 
                                    + L_Font + " " + L_PointSize + " Tf" + CHR(13)
                                    + L_color
                                    + "1 0 0 1 " + STRING(pdfColumn) + " "
                                    + STRING(pdfRow) + " Tm" + CHR(13)
                                    + "(" + pdfText + ") Tj" + CHR(13)
                                    /* + "ET" + CHR(13) */
         TT_pdf_content.obj_length  = LENGTH(L_orig_text, "character":u) .
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  /* Determine whether the TotalPages Function as called within this text
     function ... if so then indicate in the the PDF Content */
  IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
    TT_pdf_content.obj_function = TRUE.

END. /* pdf_text_xy */

PROCEDURE pdf_text_boxed_xy :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn   AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRow      AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfWidth    AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfHeight   AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfJustify  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfWeight   AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_boxed_xy","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_color   AS CHARACTER NO-UNDO.
  L_color = dec2string(pdf_TextRed(pdfStream)) + " "
          + dec2string(pdf_TextGreen(pdfStream)) + " "
          + dec2string(pdf_TextBlue(pdfStream)) + " rg" + CHR(13).

  DEFINE VARIABLE L_PointSize   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.
  L_PointSize  = dec2string(pdf_PointSize(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  DEFINE VARIABLE L_Font        AS CHARACTER NO-UNDO.
  L_Font = pdf_Font(pdfStream).
  FIND FIRST TT_pdf_font
       WHERE TT_pdf_font.obj_stream = pdfStream
         AND TT_pdf_font.font_name  = L_Font
         NO-LOCK NO-ERROR.
  L_Font = IF AVAIL TT_pdf_font THEN TT_pdf_Font.font_tag ELSE "/BF1".

  DEFINE VARIABLE L_orig_text   AS CHARACTER NO-UNDO.
  L_orig_text = pdfText.

  IF pdfColumn = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_boxed_xy","Column cannot be zero!").
    RETURN .
  END.

  IF pdfRow    = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_boxed_xy","Row cannot be zero!").
    RETURN .
  END.

  IF pdfHeight = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_boxed_xy","Height cannot be zero!").
    RETURN .
  END.

  IF pdfWidth  = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_boxed_xy","Width cannot be zero!").
    RETURN .
  END.

  IF LOOKUP(pdfJustify,"Left,Right,Center") = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_boxed_xy","Invalid Justification option passed!").
    RETURN .
  END.

  RUN pdf_replace_text (INPUT-OUTPUT pdfText).

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "TEXTXY"
         TT_pdf_content.obj_line    = pdfRow
         TT_pdf_content.obj_content = L_Font + " " + L_PointSize + " Tf" + CHR(13)
                                    + L_color
                                    + "1 0 0 1 " + STRING(pdfColumn) + " "
                                    + STRING(pdfRow) + " Tm" + CHR(13)
                                    + "(" + pdfText + ") Tj" + CHR(13)
         TT_pdf_content.obj_length  = LENGTH(L_orig_text, "character":u).
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  /* Determine whether the TotalPages Function as called within this text
     function ... if so then indicate in the the PDF Content */
  IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
    TT_pdf_content.obj_function = TRUE.

  IF pdfWeight = 0 THEN pdfWeight = 1.

  CREATE TT_pdf_content.
  ASSIGN TT_pdf_content.obj_stream  = pdfStream.
         TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
  ASSIGN TT_pdf_content.obj_type    = "GRAPHIC"
         TT_pdf_content.obj_line    = pdfRow
         TT_pdf_content.obj_content = "q" + CHR(13)
                                    + STRING(pdfWeight) + " w" + CHR(13)
                                    + STRING(pdfColumn) + " " + STRING(pdfRow) + " "
                                    + STRING(pdfWidth) + " " + STRING(pdfHeight)
                                    + " re" + CHR(13)
                                    + "B" + CHR(13)
                                    + "Q" + CHR(13)
         TT_pdf_content.obj_length  = LENGTH(L_orig_text, "character":u) .
         TT_pdf_content.obj_page    = pdf_Page(pdfStream).

  RUN pdf_set_TextY(pdfStream, pdfRow).
  RUN pdf_set_GraphicY(pdfStream, pdfRow).

END. /* pdf_text_boxed_xy */

PROCEDURE pdf_text_center:
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn   AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRow      AS INTEGER NO-UNDO.

  RUN pdf_text_xy (pdfStream,pdfText,pdfColumn
                   - INTEGER(pdf_text_width(pdfStream,pdfText) / 2),pdfRow).
END. /* pdf_text_center */

PROCEDURE pdf_text_at :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn   AS INTEGER NO-UNDO.

  DEFINE BUFFER L_B_TT_pdf_content FOR TT_pdf_content.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_at","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_TextWidth   AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_TextY       AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.
  L_TextY      = STRING(pdf_TextY(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  DEFINE VARIABLE L_orig_text   AS CHARACTER NO-UNDO.
  L_orig_text = pdfText.

  IF pdfColumn = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_at","Column cannot be zero!").
    RETURN .
  END.

  RUN pdf_replace_text (INPUT-OUTPUT pdfText).

  FIND FIRST B_TT_pdf_content
       WHERE B_TT_pdf_content.obj_stream = pdfStream
         AND B_TT_pdf_content.obj_page   = pdf_Page(pdfStream)
         AND B_TT_pdf_content.obj_line   = pdf_TextY(pdfStream) NO-ERROR.
  IF AVAIL B_TT_pdf_content THEN DO:
    ASSIGN B_TT_pdf_content.obj_content = B_TT_pdf_content.obj_content
                                        + " "
                                        + "(" + FILL(" ", pdfColumn - B_TT_pdf_content.obj_length - 1)
                                        + pdfText + ") Tj "
           B_TT_pdf_content.obj_length  = B_TT_pdf_content.obj_length
                                        + LENGTH(FILL(" ", pdfColumn - B_TT_pdf_content.obj_length - 1) + L_orig_text, "character":u).
    /* Determine whether the TotalPages Function as called within this text
       function ... if so then indicate in the the PDF Content */
    IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
      B_TT_pdf_content.obj_function = TRUE.

  END.
  ELSE DO:
    CREATE L_B_TT_pdf_content.
    ASSIGN L_B_TT_pdf_content.obj_stream  = pdfStream.
           L_B_TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
    ASSIGN L_B_TT_pdf_content.obj_type    = "TEXTAT".
           L_B_TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
           L_B_TT_pdf_content.obj_page    = pdf_Page(pdfStream).
    ASSIGN L_B_TT_pdf_content.obj_content = "1 0 0 1 " + L_LeftMargin + " "
                                            + L_TextY + " Tm" + CHR(13)
                                            + "(" + FILL(" ", pdfColumn - 1) + pdfText + ") Tj "
           L_B_TT_pdf_content.obj_length  = LENGTH( FILL(" ",pdfColumn - 1) + L_orig_text, "character":u) .

    /* Determine whether the TotalPages Function as called within this text
       function ... if so then indicate in the the PDF Content */
    IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
      L_B_TT_pdf_content.obj_function = TRUE.

  END.


  /* Increase the TextX dimension to include the new text.  The Y Dimension
     shouldn't change */
  RUN pdf_set_TextX(pdfStream,  pdf_TextX(pdfStream)
                              + pdf_text_width(pdfStream, FILL(" ", pdfColumn - 1) + pdfText)).

END. /* pdf_text_at */

PROCEDURE pdf_text_to:
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfColumn   AS INTEGER NO-UNDO.

  DEFINE BUFFER L_B_TT_pdf_content FOR TT_pdf_content.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_to","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_Spacer       AS INTEGER   NO-UNDO.
  DEFINE VARIABLE L_X            AS INTEGER   NO-UNDO.
  DEFINE VARIABLE L_Y            AS INTEGER   NO-UNDO.
  DEFINE VARIABLE L_LeftMargin   AS CHARACTER NO-UNDO.
  L_Y          = pdf_TextY(pdfStream).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  DEFINE VARIABLE L_orig_text   AS CHARACTER NO-UNDO.
  L_orig_text = pdfText.

  IF pdfColumn = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_to","Column cannot be zero!").
    RETURN .
  END.

  L_Spacer = INT(pdf_get_parameter(pdfstream,"LineSpacer")) NO-ERROR.

  RUN pdf_replace_text (INPUT-OUTPUT pdfText).

  FIND FIRST B_TT_pdf_content
       WHERE B_TT_pdf_content.obj_stream = pdfStream
         AND B_TT_pdf_content.obj_type   = "TEXTTO"
         AND B_TT_pdf_content.obj_page   = pdf_Page(pdfStream)
         AND B_TT_pdf_content.obj_line   = pdf_TextY(pdfStream) NO-ERROR.
  IF AVAIL B_TT_pdf_content THEN DO:
    ASSIGN B_TT_pdf_content.obj_content = B_TT_pdf_content.obj_content
                                        + " "
                                        + "(" + FILL(" ", pdfColumn - (B_TT_pdf_content.obj_length + LENGTH(L_Orig_text, "character":u)))
                                        + pdfText + ") Tj "
           B_TT_pdf_content.obj_length  = B_TT_pdf_content.obj_length
                                        + LENGTH(FILL(" ", pdfColumn - (B_TT_pdf_content.obj_length + LENGTH(L_orig_text, "character":u))) + L_orig_text, "character":u).

    /* Determine whether the TotalPages Function as called within this text
       function ... if so then indicate in the the PDF Content */
    IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
      B_TT_pdf_content.obj_function = TRUE.

    /* Set the TextX dimension to the new text column.  The Y Dimension
     shouldn't change */
    RUN pdf_set_TextX(pdfStream,  pdf_TextX(pdfStream)
                              + pdf_text_width(pdfStream, FILL(" ", pdfColumn - (B_TT_pdf_content.obj_length
                                                                    + LENGTH(L_Orig_text, "character":u)))
                                                                    + pdfText)).

  END.
  ELSE DO:
    IF pdf_FontType(pdfStream) = "FIXED" THEN DO:
      CREATE L_B_TT_pdf_content.
      ASSIGN L_B_TT_pdf_content.obj_stream  = pdfStream.
             L_B_TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
      ASSIGN L_B_TT_pdf_content.obj_type    = "TEXTTO"
             L_B_TT_pdf_content.obj_content = "1 0 0 1 " + L_LeftMargin + " "
                                              + STRING(L_Y + L_Spacer) + " Tm" + CHR(13)
                                              + "(" + FILL(" ", pdfColumn - LENGTH(L_orig_text, "character":u) - 1) + pdfText + ") Tj "
             L_B_TT_pdf_content.obj_length  = LENGTH( FILL(" ",pdfColumn - LENGTH(L_orig_text, "character":u) - 1) + L_orig_text, "character":u) .
             L_B_TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
             L_B_TT_pdf_content.obj_page    = pdf_Page(pdfStream).

      /* Determine whether the TotalPages Function as called within this text
         function ... if so then indicate in the the PDF Content */
      IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
        L_B_TT_pdf_content.obj_function = TRUE.

      /* Set the TextX dimension to the new text column.  The Y Dimension
         shouldn't change */
      RUN pdf_set_TextX(pdfStream,  pdf_TextX(pdfStream)
                                  + pdf_text_width(pdfStream, FILL(" ", pdfColumn - LENGTH(L_orig_text, "character":u) - 1)
                                                              + pdfText)).

    END. /* Fixed */

    /* Variable - proportional fonts */
    ELSE DO:
      L_X = pdf_text_width(pdfStream,FILL(" ",pdfColumn))
          - pdf_text_width(pdfStream, L_orig_text).

      CREATE L_B_TT_pdf_content.
      ASSIGN L_B_TT_pdf_content.obj_stream  = pdfStream.
             L_B_TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
      ASSIGN L_B_TT_pdf_content.obj_type    = "TEXTTO".
             L_B_TT_pdf_content.obj_content = "1 0 0 1 "
                                            + STRING(L_X) + " "
                                            + STRING(L_Y + L_Spacer) + " Tm" + CHR(13)
                                            + "(" + L_orig_text + ") Tj ".
             L_B_TT_pdf_content.obj_length  = LENGTH( FILL(" ",pdfColumn - LENGTH(L_orig_text, "character":u) - 1) + L_orig_text, "character":u) .
             L_B_TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
             L_B_TT_pdf_content.obj_page    = pdf_Page(pdfStream).

      /* Determine whether the TotalPages Function as called within this text
         function ... if so then indicate in the the PDF Content */
      IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
         L_B_TT_pdf_content.obj_function = TRUE.

      /* Set the TextX dimension to the new text column.  The Y Dimension
         shouldn't change */
      RUN pdf_set_TextX(pdfStream,  L_X + pdf_text_width(pdfstream,L_Orig_text)).

    END. /* Variable */
  END.

END. /* pdf_text_to */

PROCEDURE pdf_text_align:
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfAlign    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfX        AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfY        AS INTEGER NO-UNDO.

  DEFINE VARIABLE L_color       AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_Font        AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_PointSize   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_LeftMargin  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Width       AS DECIMAL NO-UNDO.

  /* Begin Procedure Validation */
  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_align","Cannot find Stream!").
    RETURN .
  END.

  IF LOOKUP(pdfAlign,"LEFT,RIGHT,CENTER":U) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_align","Invalid Alignment option passed!").
    RETURN .
  END.

  IF pdfX = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_align","X location cannot be zero!").
    RETURN .
  END.

  IF pdfY = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_text_align","Y location cannot be zero!").
    RETURN .
  END.

  /* end of validation */
  L_color = dec2string(pdf_TextRed(pdfStream)) + " "
          + dec2string(pdf_TextGreen(pdfStream)) + " "
          + dec2string(pdf_TextBlue(pdfStream)) + " rg" + CHR(13).

  L_PointSize  = dec2string(pdf_PointSize(pdfStream)).
  L_LeftMargin = STRING(pdf_LeftMargin(pdfStream)).

  L_font = pdf_Font(pdfStream).

  FIND FIRST TT_pdf_font
       WHERE TT_pdf_font.obj_stream = pdfStream
         AND TT_pdf_font.font_name  = L_Font NO-LOCK NO-ERROR.
  L_Font = IF AVAIL TT_pdf_font THEN TT_pdf_Font.font_tag ELSE "/BF1".

  pdfText = TRIM(pdfText).
  RUN pdf_replace_text (INPUT-OUTPUT pdftext).

  L_Width = pdf_text_widthdec(pdfStream, pdftext).

  CASE pdfAlign:
    WHEN "RIGHT" THEN DO:
      CREATE TT_pdf_content.
      ASSIGN TT_pdf_content.obj_stream  = pdfStream.
             TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
      ASSIGN TT_pdf_content.obj_type    = "TEXTXY".
             TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
      ASSIGN TT_pdf_content.obj_content = CHR(13) 
                                        + L_Font + " " + L_PointSize + " Tf" + CHR(13)
                                        + L_color
                                        + "1 0 0 1 " + dec2string(pdfX - L_Width) + " "
                                        + STRING(pdfY) + " Tm" + CHR(13)
                                        + "(" + pdfText + ") Tj" + CHR(13)
             TT_pdf_content.obj_length  = LENGTH(pdfText, "character":u) .
             TT_pdf_content.obj_page    = pdf_Page(pdfStream).

      RUN pdf_set_TextX(pdfStream,  pdfX ).
    END. /* Right Alignment */

    WHEN "LEFT" THEN DO:
      CREATE TT_pdf_content.
      ASSIGN TT_pdf_content.obj_stream  = pdfStream.
             TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
      ASSIGN TT_pdf_content.obj_type    = "TEXTXY".
             TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
      ASSIGN TT_pdf_content.obj_content = CHR(13) 
                                        + L_Font + " " + L_PointSize + " Tf" + CHR(13)
                                        + L_color
                                        + "1 0 0 1 " +  STRING(pdfX) + " "
                                        + STRING(pdfY) + " Tm" + CHR(13)
                                        + "(" + pdfText + ") Tj" + CHR(13)
             TT_pdf_content.obj_length  = LENGTH(pdfText, "character":u) .
             TT_pdf_content.obj_page    = pdf_Page(pdfStream).

      RUN pdf_set_TextX(pdfStream,  INT(pdfX + L_Width)).
    END. /* Left Alignment */

    WHEN "CENTER" THEN DO:
      L_Width = L_Width / 2.

      CREATE TT_pdf_content.
      ASSIGN TT_pdf_content.obj_stream  = pdfStream.
             TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
      ASSIGN TT_pdf_content.obj_type    = "TEXTXY".
             TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
      ASSIGN TT_pdf_content.obj_content = CHR(13) 
                                        + L_Font + " " + L_PointSize + " Tf" + CHR(13)
                                        + L_color
                                        + "1 0 0 1 " +  dec2string(pdfX - L_Width) + " "
                                        + STRING(pdfY) + " Tm" + CHR(13)
                                        + "(" + pdfText + ") Tj" + CHR(13)
             TT_pdf_content.obj_length  = LENGTH(pdfText, "character":u) .
             TT_pdf_content.obj_page    = pdf_Page(pdfStream).

      RUN pdf_set_TextX(pdfStream,  INT(pdfX + L_Width)).
    END. /* Centered Alignment */

  END CASE.
  
  /* Determine whether the TotalPages Function as called within this text
     function ... if so then indicate in the the PDF Content */
  IF INDEX(pdfText,"@@TOTALPages-" + pdfStream) > 0 THEN
    TT_pdf_content.obj_function = TRUE.

END. /* pdf_text_align */

PROCEDURE pdf_set_Angle :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_Angle","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_option  AS CHARACTER NO-UNDO.
  L_Option = "0,90,180,270".

  IF LOOKUP(STRING(pdfValue),L_option) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_Angle","Invalid Angle option passed!").
    RETURN .
  END.

  FIND FIRST B_TT_pdf_param
       WHERE B_TT_pdf_param.obj_stream    = pdfStream
         AND B_TT_pdf_param.obj_parameter = "Angle" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_param THEN DO:
    CREATE B_TT_pdf_param.
    ASSIGN B_TT_pdf_param.obj_stream    = pdfStream
           B_TT_pdf_param.obj_parameter = "Angle".
  END.

  B_TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_Angle */

PROCEDURE pdf_set_Orientation :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_width   AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_height  AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_Orientation","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_option  AS CHARACTER NO-UNDO.
  L_Option = "Portrait,Landscape".

  IF LOOKUP(pdfValue,L_option) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_Orientation","Invalid Orientation option passed!").
    RETURN .
  END.

  FIND FIRST B_TT_pdf_param
       WHERE B_TT_pdf_param.obj_stream    = pdfStream
         AND B_TT_pdf_param.obj_parameter = "Orientation" NO-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_param THEN DO:
    CREATE B_TT_pdf_param.
    ASSIGN B_TT_pdf_param.obj_stream    = pdfStream
           B_TT_pdf_param.obj_parameter = "Orientation".
  END.

  IF B_TT_pdf_param.obj_value <> pdfValue THEN DO:
    L_width  = pdf_PageWidth(pdfStream).
    L_height = pdf_PageHeight(pdfStream).

    IF pdfValue = "Landscape" THEN DO:
      RUN pdf_set_PageWidth(pdfStream,L_height).
      RUN pdf_set_PageHeight(pdfStream,L_width).
    END.
    ELSE DO:
      RUN pdf_set_PageWidth(pdfStream,L_width).
      RUN pdf_set_PageHeight(pdfStream,L_height).
    END.
  END.

  B_TT_pdf_param.obj_value = pdfValue.

END. /* pdf_set_Orientation */

PROCEDURE pdf_set_VerticalSpace :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_VerticalSpace","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_VerticalSpace","Vertical Space cannot be zero!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "VerticalSpace" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "VerticalSpace".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_VerticalSpace */

PROCEDURE pdf_set_LeftMargin :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_LeftMargin","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_LeftMargin","Left Margin cannot be zero!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "LeftMargin" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "LeftMargin".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_LeftMargin */

PROCEDURE pdf_set_TopMargin :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TopMargin","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_TopMargin","Top Margin cannot be zero!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "TopMargin" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "TopMargin".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_TopMargin */

PROCEDURE pdf_set_BottomMargin :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS INTEGER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_BottomMargin","Cannot find Stream!").
    RETURN .
  END.

  IF pdfValue <= 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_BottomMargin","Bottom Margin cannot be <= zero!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "BottomMargin" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "BottomMargin".
  END.

  TT_pdf_param.obj_value = STRING(pdfValue).

END. /* pdf_set_BottomMargin */

FUNCTION pdf_Orientation RETURN CHARACTER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_Orientation","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "Orientation"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN RETURN TT_pdf_param.obj_value.
  ELSE RETURN "Portrait".

END. /* pdf_Orientation */

PROCEDURE pdf_set_PaperType :
  DEFINE INPUT PARAMETER pdfStream          AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue           AS CHARACTER NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PaperType","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_option  AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_width   AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_height  AS INTEGER NO-UNDO.

  L_Option = "A0,A1,A2,A3,A4,A5,A6,B5,LETTER,LEGAL,LEDGER".

  IF LOOKUP(pdfValue,L_option) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_PaperType","Invalid Paper Type option passed!").
    RETURN .
  END.

  FIND FIRST TT_pdf_param
       WHERE TT_pdf_param.obj_stream    = pdfStream
         AND TT_pdf_param.obj_parameter = "PaperType" EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_param THEN DO:
    CREATE TT_pdf_param.
    ASSIGN TT_pdf_param.obj_stream    = pdfStream
           TT_pdf_param.obj_parameter = "PaperType".
  END.

  /* Set the Paper Type */
  TT_pdf_param.obj_value = pdfValue.

  /* Determine the Paper Height and Width */
  CASE pdfValue:
    WHEN "A0" THEN
      ASSIGN L_width  = 2380
             L_height = 3368.
    WHEN "A1" THEN
      ASSIGN L_width  = 1684
             L_height = 2380.
    WHEN "A2" THEN
      ASSIGN L_width  = 1190
             L_height = 1684.
    WHEN "A3" THEN
      ASSIGN L_width  = 842
             L_height = 1190.
    WHEN "A4" THEN
      ASSIGN L_width  = 595
             L_height = 842.
    WHEN "A5" THEN
      ASSIGN L_width  = 421
             L_height = 595.
    WHEN "A6" THEN
      ASSIGN L_width  = 297
             L_height = 421.
    WHEN "B5" THEN
      ASSIGN L_width  = 501
             L_height = 709.
    WHEN "LETTER" THEN
      ASSIGN L_width  = 612
             L_height = 792.
    WHEN "LEGAL" THEN
      ASSIGN L_width  = 612
             L_height = 1008.
    WHEN "LEDGER" THEN
      ASSIGN L_width  = 1224
             L_height = 792.
    OTHERWISE
      ASSIGN L_width  = 612
             L_height = 792.
  END CASE.

  /* Now Set the Page Height and Width Parameters */
  IF pdf_Orientation(pdfStream) = "Portrait" THEN DO:
    RUN pdf_set_PageWidth(pdfStream,L_width).
    RUN pdf_set_PageHeight(pdfStream,L_height).
  END.
  ELSE DO:
    RUN pdf_set_PageWidth(pdfStream,L_height).
    RUN pdf_set_PageHeight(pdfStream,L_width).
  END.

END. /* pdf_set_PaperType */

FUNCTION pdf_PaperType RETURN CHARACTER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_PaperType","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "PaperType"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN RETURN TT_pdf_param.obj_value.
  ELSE RETURN "LETTER".

END. /* pdf_PaperType */

FUNCTION pdf_Render RETURN INTEGER ( INPUT pdfStream AS CHARACTER):

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_Render","Cannot find Stream!").
    RETURN ERROR.
  END.

  FIND FIRST TT_pdf_param WHERE TT_pdf_param.obj_stream    = pdfStream
                            AND TT_pdf_param.obj_parameter = "Render"
                            NO-LOCK NO-ERROR.
  IF AVAIL TT_pdf_param THEN RETURN INT(TT_pdf_param.obj_value).
  ELSE RETURN 0.

END. /* pdf_Render */

FUNCTION pdf_get_wrap_length RETURNS INTEGER ( INPUT pdfStream   AS CHARACTER,
                                               INPUT pdfText AS CHARACTER,
                                               INPUT pdfWidth AS INTEGER ):
  DEFINE VAR pdfHeight  AS INTEGER NO-UNDO.

  DEFINE VAR v-thisline AS CHAR NO-UNDO.
  DEFINE VAR v-lastline AS CHAR NO-UNDO.
  DEFINE VAR v-fontsize AS DECIMAL NO-UNDO.
  DEFINE VAR i          AS INTEGER NO-UNDO.
  DEFINE VAR v-pointsize AS INTEGER NO-UNDO.

  ASSIGN v-fontsize = PDF_text_width(pdfStream,"i") * .81
         v-pointsize = PDF_PointSize(pdfStream).

  ASSIGN pdfText = REPLACE(pdfText,"|","&pipe;").
  /* Spaces */
  ASSIGN pdfText = REPLACE(pdfText," ","| ").
  /* Hyphens */
  ASSIGN pdfText = REPLACE(pdfText,"-","-|").
  /* Commas */
  ASSIGN pdfText = REPLACE(pdfText,",",",|").

  /* Divide up the pdf text into lines of width less than the
     available width */
  DO i = 1 TO NUM-ENTRIES(pdfText,"|"):
    ASSIGN v-lastline = v-thisline.
    ASSIGN v-thisline = v-thisline
                      + REPLACE(ENTRY(i,pdfText,"|"),"&pipe;","|").

    IF PDF_text_width(pdfStream,TRIM(v-thisline)) >= (pdfWidth * v-fontSize)
    THEN DO:
        ASSIGN pdfHeight = pdfHeight + (v-pointsize * 2).
        ASSIGN v-thisline = TRIM(REPLACE(ENTRY(i,pdfText,"|"),"&pipe;","|")).
    END.
  END.

  IF v-thisline NE "" THEN
    ASSIGN pdfHeight = pdfHeight + v-pointsize.

  RETURN pdfHeight.
END FUNCTION. /* pdf_get_wrap_length */

/* ---------------------- Define INTERNAL PROCEDURES ----------------------- */
PROCEDURE pdf_init_param:
  DEFINE INPUT PARAMETER pdfStream AS CHARACTER NO-UNDO.

  /* Create Parameters */
  RUN pdf_set_Orientation(pdfStream,"Portrait").
  RUN pdf_set_PaperType(pdfStream,"LETTER").  /* This also default the PageWidth
                                             and PageHeight Parameters */
  RUN pdf_set_Font(pdfStream,"Courier",10.0). /* This also sets the PointSize */
  RUN pdf_set_VerticalSpace(pdfStream,10).
  RUN pdf_set_LeftMargin(pdfStream,10).
  RUN pdf_set_TopMargin(pdfStream,50).
  RUN pdf_set_BottomMargin(pdfStream,1).
  RUN pdf_set_Angle(pdfStream,0).
  RUN pdf_set_TextY(pdfStream,pdf_PageHeight(pdfStream) - pdf_TopMargin(pdfStream)).
  RUN pdf_set_TextX(pdfStream, pdf_LeftMargin(pdfStream)).
  RUN pdf_set_GraphicY(pdfStream,0).
  RUN pdf_set_GraphicX(pdfStream,0).
  RUN pdf_set_TextRed(pdfStream,.0).
  RUN pdf_set_TextGreen(pdfStream,.0).
  RUN pdf_set_TextBlue(pdfStream,.0).
  RUN pdf_set_FillRed(pdfStream,1.0).
  RUN pdf_set_FillGreen(pdfStream,1.0).
  RUN pdf_set_FillBlue(pdfStream,1.0).
  RUN pdf_set_PageRotate(pdfStream,0).    /* Default to zero rotation */

END. /* pdf_init_param */

PROCEDURE pdf_Header :
  DEFINE INPUT PARAMETER P_Stream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER p_Encrypt  AS LOGICAL NO-UNDO.

  DEFINE VARIABLE l_CreationDate  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE m_EncryptKey    AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_Author        AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_CreationDate  AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_Creator       AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_Producer      AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_Subject       AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_Title         AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_Keywords      AS MEMPTR NO-UNDO.

  /* Version Compatibilities */
  PUT STREAM S_pdf_inc UNFORMATTED
      "%PDF-1.4" {&pdfSKIP}.

  /* Output 4 Binary Characters (greater than ASCII 128) to indicate to a binary
     file -- randomly selected codes */
  PUT STREAM S_pdf_inc UNFORMATTED
      "%" CHR(228) CHR(227) CHR(207) CHR(210) {&pdfSKIP}.

  l_CreationDate = "D:" + TRIM(STRING(YEAR(TODAY),"9999"))
                 + TRIM(STRING(MONTH(TODAY),"99"))
                 + TRIM(STRING(DAY(TODAY),"99"))
                 + REPLACE(STRING(TIME,"hh:mm:ss"),":","") + "-0800".

  ObjectSequence( p_Stream,1, "Info", 0 ).
  /* Display Creation, Title, Producer etc Information */
  PUT STREAM S_pdf_inc UNFORMATTED
      "1 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}.

  IF p_Encrypt THEN DO:
    SET-SIZE(m_EncryptKey) = 10.
    RUN GetEncryptKey IN h_PDF-Encrypt
        (INPUT  TT_pdf_stream.obj_UniqueID,
         INPUT  1,
         INPUT  0,
         INPUT  pdf-EncryptKeyMemPtr,
         OUTPUT m_EncryptKey).

    IF LENGTH(pdf_get_info(p_Stream,"Author"), "character":u) > 0 THEN DO:
      SET-SIZE(m_Author) = LENGTH(pdf_get_info(P_Stream,"Author"), "character":u) + 1.
      PUT-STRING(m_Author,1) = pdf_get_info(P_Stream,"Author").

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  m_Author,
                          OUTPUT m_Author).

      PUT STREAM S_pdf_inc UNFORMATTED "/Author <".

      RUN OutputMemPtrAsHex (m_Author).
      SET-SIZE(m_Author) = 0.

      PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    END. /* Show Author */

    /* Show Creation Date */
    SET-SIZE(m_CreationDate) = LENGTH(l_CreationDate, "character":u) + 1.
    PUT-STRING(m_CreationDate,1) = l_CreationDate.

    RUN EncryptContent IN h_PDF-Encrypt
                       (INPUT  m_EncryptKey,
                        INPUT  m_CreationDate,
                        OUTPUT m_CreationDate).

    PUT STREAM S_pdf_inc UNFORMATTED "/CreationDate<".

    RUN OutputMemPtrAsHex (m_CreationDate).
    SET-SIZE(m_CreationDate) = 0.
    PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    /* Show Creation Date */

    IF LENGTH(pdf_get_info(p_Stream,"Creator"), "character":u) > 0 THEN DO:
      SET-SIZE(m_Creator) = LENGTH(pdf_get_info(P_Stream,"Creator"), "character":u) + 1.
      PUT-STRING(m_Creator,1) = pdf_get_info(P_Stream,"Creator").

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  m_Creator,
                          OUTPUT m_Creator).

      PUT STREAM S_pdf_inc UNFORMATTED "/Creator <".

      RUN OutputMemPtrAsHex (m_Creator).
      SET-SIZE(m_Creator) = 0.
      PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    END. /* Show Creator */

    IF LENGTH(pdf_get_info(p_Stream,"Producer"), "character":u) > 0 THEN DO:
      SET-SIZE(m_Producer) = LENGTH(pdf_get_info(P_Stream,"Producer"), "character":u) + 1.
      PUT-STRING(m_Producer,1) = pdf_get_info(P_Stream,"Producer").

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  m_Producer,
                          OUTPUT m_Producer).

      PUT STREAM S_pdf_inc UNFORMATTED "/Producer <".

      RUN OutputMemPtrAsHex (m_Producer).
      SET-SIZE(m_Producer) = 0.
      PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    END. /* Show Producer */

    IF LENGTH(pdf_get_info(p_Stream,"Subject"), "character":u) > 0 THEN DO:
      SET-SIZE(m_Subject) = LENGTH(pdf_get_info(P_Stream,"Subject"), "character":u) + 1.
      PUT-STRING(m_Subject,1) = pdf_get_info(P_Stream,"Subject").

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  m_Subject,
                          OUTPUT m_Subject).

      PUT STREAM S_pdf_inc UNFORMATTED "/Subject <".

      RUN OutputMemPtrAsHex (m_Subject).
      SET-SIZE(m_Subject) = 0.
      PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    END. /* Show Subject */

    IF LENGTH(pdf_get_info(p_Stream,"Title"), "character":u) > 0 THEN DO:
      SET-SIZE(m_Title) = LENGTH(pdf_get_info(P_Stream,"Title"), "character":u) + 1.
      PUT-STRING(m_Title,1) = pdf_get_info(P_Stream,"Title").

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  m_Title,
                          OUTPUT m_Title).

      PUT STREAM S_pdf_inc UNFORMATTED "/Title <".

      RUN OutputMemPtrAsHex (m_Title).
      SET-SIZE(m_Title) = 0.
      PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    END. /* Show Title */

    IF LENGTH(pdf_get_info(p_Stream,"Keywords"), "character":u) > 0 THEN DO:
      SET-SIZE(m_Keywords) = LENGTH(pdf_get_info(P_Stream,"Keywords"), "character":u) + 1.
      PUT-STRING(m_Keywords,1) = pdf_get_info(P_Stream,"Keywords").

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  m_Keywords,
                          OUTPUT m_Keywords).

      PUT STREAM S_pdf_inc UNFORMATTED "/Keywords <".

      RUN OutputMemPtrAsHex (m_Keywords).
      SET-SIZE(m_Keywords) = 0.
      PUT STREAM S_pdf_inc UNFORMATTED ">" {&pdfSKIP}.
    END. /* Show Title */

    SET-SIZE(m_EncryptKey) = 0.
  END. /* If Encrypted PDF */

  ELSE DO: /* Not Encrypted */
    PUT STREAM S_pdf_inc UNFORMATTED
        "/Author (" pdf_get_info(p_Stream,"Author") ")" {&pdfSKIP}
        "/CreationDate (" l_CreationDate ")" {&pdfSKIP}
        "/Producer (" pdf_get_info(P_Stream,"Producer") ")" {&pdfSKIP}
        "/Creator (" pdf_get_info(P_Stream,"Creator") ")" {&pdfSKIP}
        "/Subject (" pdf_get_info(P_Stream,"Subject") ")" {&pdfSKIP}
        "/Title (" pdf_get_info(P_Stream,"Title") ")" {&pdfSKIP}
        "/Keywords (" pdf_get_info(P_Stream,"Keywords") ")" {&pdfSKIP}.
  END. /* Not Encrypted */

  PUT STREAM S_pdf_inc UNFORMATTED
      " >>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_header */

PROCEDURE pdf_LoadBase14:

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Object          AS INTEGER NO-UNDO.

  L_Object = 5.

  /* ---- Beginning of Courier Fonts ---- */

  /* Create Associated Object */
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Courier"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF1"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "FIXED"
         TT_pdf_font.font_width = "600".
  L_Object = L_Object + 1.

  /* Create Associated Object */
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Courier-Oblique"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF2"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "FIXED"
         TT_pdf_font.font_width = "600".
  L_Object = L_Object + 1.

  /* Create Associated Object */
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Courier-Bold"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF3"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "FIXED"
         TT_pdf_font.font_width = "600".
  L_Object = L_Object + 1.

  /* Create Associated Object */
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Courier-BoldOblique"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF4"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "FIXED"
         TT_pdf_font.font_width = "600".
  L_Object = L_Object + 1.

  /* ---- End of Courier Fonts ---- */

  /* ---- Beginning of Helvetica Fonts ---- */

  /* Create Associated Object */
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Helvetica"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF5"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width = FILL("788 ", 32) + "278 278 355 556 556 889 "
                                + "667 222 333 333 389 584 278 333 278 278 556 "
                                + "556 556 556 556 556 556 556 556 556 278 278 "
                                + "584 584 584 556 1015 667 667 722 722 667 611 778 "
                                + "722 278 500 667 556 833 722 778 667 778 722 "
                                + "667 611 722 667 944 667 667 611 278 278 278 "
                                + "469 556 222 556 556 500 556 556 278 556 556 "
                                + "222 222 500 222 833 556 556 556 556 333 500 "
                                + "278 556 500 722 500 500 500 334 260 334 584 "
                                + "333 556 556 167 556 556 556 556 191 333 556 "
                                + "333 333 500 500 556 556 556 278 537 350 222 "
                                + "333 333 556 1000 1000 611 333 333 333 333 333 "
                                + "333 333 333 333 333 333 333 333 1000 1000 370 "
                                + "556 778 1000 365 889 278 222 611 944 611".
  L_Object = L_Object + 1.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Helvetica-Oblique"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF6"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width = FILL("788 ", 32) + "278 278 355 556 556 889 "
                                + "667 222 333 333 389 584 278 333 278 278 556 "
                                + "556 556 556 556 556 556 556 556 556 278 278 "
                                + "584 584 584 556 1015 667 667 722 722 667 611 "
                                + "778 722 278 500 667 556 833 722 778 667 778 "
                                + "722 667 611 722 667 944 667 667 611 278 278 "
                                + "278 469 556 222 556 556 500 556 556 278 556 "
                                + "556 222 222 500 222 833 556 556 556 556 333 "
                                + "500 278 556 500 722 500 500 500 334 260 334 "
                                + "584 333 556 556 167 556 556 556 556 191 333 "
                                + "556 333 333 500 500 556 556 556 278 537 350 "
                                + "222 333 333 556 1000 1000 611 333 333 333 333 "
                                + "333 333 333 333 333 333 333 333 333 1000 "
                                + "1000 370 556 778 1000 365 889 278 222 611 "
                                + "944 611".
  L_Object = L_Object + 1.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Helvetica-Bold"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF7"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width = FILL("788 ", 32) + "278 333 474 556 556 889 "
                                + "722 278 333 333 389 584 278 333 278 278 556 "
                                + "556 556 556 556 556 556 556 556 556 333 333 "
                                + "584 584 584 611 975 722 722 722 722 667 611 "
                                + "778 722 278 556 722 611 833 722 778 667 778 "
                                + "722 667 611 722 667 944 667 667 611 333 278 "
                                + "333 584 556 278 556 611 556 611 556 333 611 "
                                + "611 278 278 556 278 889 611 611 611 611 389 "
                                + "556 333 611 556 778 556 556 500 389 280 389 "
                                + "584 333 556 556 167 556 556 556 556 238 500 "
                                + "556 333 333 611 611 556 556 556 278 556 350 "
                                + "278 500 500 556 1000 1000 611 333 333 333 333 "
                                + "333 333 333 333 333 333 333 333 333 1000 "
                                + "1000 370 611 778 1000 365 889 278 278 611 "
                                + "944 611".
  L_Object = L_Object + 1.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Helvetica-BoldOblique"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF8"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width = FILL("788 ", 32) + "278 333 474 556 556 889 "
                                + "722 278 333 333 389 584 278 333 278 278 556 "
                                + "556 556 556 556 556 556 556 556 556 333 333 "
                                + "584 584 584 611 975 722 722 722 722 667 611 "
                                + "778 722 278 556 722 611 833 722 778 667 778 "
                                + "722 667 611 722 667 944 667 667 611 333 278 "
                                + "333 584 556 278 556 611 556 611 556 333 611 "
                                + "611 278 278 556 278 889 611 611 611 611 389 "
                                + "556 333 611 556 778 556 556 500 389 280 389 "
                                + "584 333 556 556 167 556 556 556 556 238 500 "
                                + "556 333 333 611 611 556 556 556 278 556 350 "
                                + "278 500 500 556 1000 1000 611 333 333 333 "
                                + "333 333 333 333 333 333 333 333 333 333 "
                                + "1000 1000 370 611 778 1000 365 889 278 278 "
                                + "611 944 611".
  L_Object = L_Object + 1.

  /* ---- End of Helvetica Fonts ---- */

  /* ---- Beginning of Times Roman Fonts ---- */

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Times-Roman"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF9"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width =   FILL("788 ", 32) + "250 " + "333 " + "408 "
                                + "500 500 833 778 180 333 333 500 564 250 "
                                + "333 250 278 500 500 500 500 500 500 500 "
                                + "500 500 500 278 278 564 564 564 444 921 "
                                + "722 667 667 722 611 556 722 722 333 389 "
                                + "722 611 889 722 722 556 722 667 556 611 "
                                + "722 722 944 722 722 611 333 278 333 469 "
                                + "500 333 444 500 444 500 444 333 500 500 "
                                + "278 278 500 278 778 500 500 500 500 333 "
                                + "389 278 500 500 722 500 500 444 480 200 "
                                + "480 541 778 500 578 333 500 444 1000 500 "
                                + "500 333 1000 556 333 889 667 611 722 444 "
                                + "333 333 444 444 350 500 1000 333 980 389 "
                                + "333 722 486 444 722 250 333 500 500 500 "
                                + "500 200 500 333 760 276 500 564 333 760 "
                                + "500 400 549 300 300 333 576 453 250 333 "
                                + "300 310 500 750 750 750 444 722 722 722 "
                                + "722 722 722 889 667 611 611 611 611 333 "
                                + "333 333 333 722 722 722 722 722 722 722 "
                                + "564 722 722 722 722 722 722 556 500 444 "
                                + "444 444 444 444 444 667 444 444 444 444 "
                                + "444 278 278 278 278 500 500 500 500 500 "
                                + "500 500 549 500 500 500 500 500 500 500 "
                                + "500".
  L_Object = L_Object + 1.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Times-Italic"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF10"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width =   FILL("788 ", 32) + "250 333 420 500 500 "
                                + "833 778 333 333 333 500 675 250 333 250 278 "
                                + "500 500 500 500 500 500 500 500 500 500 333 "
                                + "333 675 675 675 500 920 611 611 667 722 611 "
                                + "611 722 722 333 444 667 556 833 667 722 611 "
                                + "500 559 722 611 833 611 556 556 389 278 389 "
                                + "422 500 333 500 500 444 500 444 278 500 500 "
                                + "278 278 444 278 722 500 500 500 500 389 389 "
                                + "278 500 444 667 444 444 389 400 278 400 541 "
                                + "389 500 500 167 500 500 500 500 214 556 500 "
                                + "333 333 500 500 500 500 500 250 523 350 333 "
                                + "556 556 500 889 1000 500 333 333 333 333 333 "
                                + "333 333 333 333 333 333 333 333 889 889 276"
                                + "556 722 944 310 667 278 278 500 667 500".
  L_Object = L_Object + 1.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Times-Bold"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF11"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width =  FILL("788 ", 32) + "250 333 555 500 500 "
                                + "1000 833 333 333 333 500 570 250 333 250 "
                                + "278 500 500 500 500 500 500 500 500 500 500 "
                                + "333 333 570 570 570 500 930 722 667 722 722 "
                                + "667 611 778 778 389 500 778 667 944 722 778 "
                                + "611 778 722 556 667 722 722 1000 722 722 667 "
                                + "333 278 333 581 500 333 500 556 444 556 444 "
                                + "333 500 556 278 333 556 278 833 556 500 556 "
                                + "556 444 389 333 556 500 722 500 500 444 394 "
                                + "220 394 520 333 500 500 167 500 500 500 500 "
                                + "278 500 500 333 333 556 556 500 500 500 250 "
                                + "540 350 333 500 500 500 1000 1000 500 333 333 "
                                + "333 333 333 333 333 333 333 333 333 333 333 "
                                + "1000 1000 300 667 778 1000 330 722 278 278 "
                                + "500 722 556".
  L_Object = L_Object + 1.

  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Times-BoldItalic"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF12"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "VARIABLE"
         TT_pdf_font.font_width = FILL("788 ", 32) + "250 389 555 500 500 833 "
                                + "778 333 333 333 500 570 250 333 250 278 500 "
                                + "500 500 500 500 500 500 500 500 500 333 333 "
                                + "570 570 570 500 832 667 667 667 722 667 667 "
                                + "722 778 389 500 667 611 889 722 722 611 722 "
                                + "667 556 611 722 667 889 667 611 611 333 278 "
                                + "333 570 500 333 500 500 444 500 444 333 500 "
                                + "556 278 278 500 278 778 556 500 500 500 389 "
                                + "278 556 444 667 500 444 389 348 220 348 570 "
                                + "389 500 500 167 500 500 500 500 278 500 500 "
                                + "333 333 556 556 500 500 500 250 500 350 333 "
                                + "500 500 500 1000 1000 500 333 333 333 333 333 "
                                + "333 333 333 333 333 333 333 333 1000 944 266 "
                                + "611 722 944 300 722 278 278 500 722 500".

  /* ---- End of Times Roman Fonts ---- */

  
  
  /* ---- Beginning of Symbol Font ---- */

  /* Create Associated Object */
  L_Object = L_Object + 1.
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "Symbol"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF13"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "FIXED"
         TT_pdf_font.font_width = "600".
  
  /* ---- End of Symbol Font ---- */

  /* ---- Beginning of ZapfDingbats Font ---- */

  /* Create Associated Object */
  L_Object = L_Object + 1.
  CREATE TT_pdf_font.
  ASSIGN TT_pdf_font.font_name  = "ZapfDingbats"
         TT_pdf_font.font_file  = "PDFBASE14"
         TT_pdf_font.font_afm   = ""
         TT_pdf_font.font_dif   = ""
         TT_pdf_font.font_obj   = L_Object
         TT_pdf_font.font_tag   = "/BF14"
         TT_pdf_font.obj_stream = pdfStream
         TT_pdf_font.font_type  = "FIXED"
         TT_pdf_font.font_width = "600".
  
  /* ---- End of Symbol Font ---- */

  pdf_inc_ObjectSequence = L_Object.

END. /* pdf_LoadBase14 */

PROCEDURE pdf_Encoding:
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  ObjectSequence(pdfStream,4, "Encoding", 0).
  PUT STREAM S_pdf_inc UNFORMATTED
      "4 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "/Type /Encoding" {&pdfSKIP}
      "/BaseEncoding /WinAnsiEncoding" {&pdfSKIP}
      ">>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_Encoding */

PROCEDURE pdf_Resources:
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_fontobj AS CHARACTER NO-UNDO.

  FOR EACH TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream NO-LOCK:
    ASSIGN L_fontobj = L_fontobj + " " + TT_pdf_font.font_tag + " "
                     + STRING(TT_pdf_font.font_obj) + " 0 R".
  END.

  Vuse-font = "/BF1".

  ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Resource", 0).
  pdf-Res-Object = pdf_inc_ObjectSequence.
  PUT STREAM S_pdf_inc UNFORMATTED
      pdf_inc_ObjectSequence " 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "  /Font << " L_fontobj " >>" {&pdfSKIP}
      "  /ProcSet [ /PDF /Text /ImageC ]" {&pdfSKIP}
      "  /XObject << " {&pdfSKIP}.

  /* Output Image Definitions */
  FOR EACH TT_pdf_image WHERE TT_pdf_image.obj_stream = pdfStream:
    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_image.image_tag " " TT_pdf_image.image_obj " 0 R " {&pdfSKIP}.
  END.

  PUT STREAM S_pdf_inc UNFORMATTED
      " >>" {&pdfSKIP}
      ">>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_Resources */

PROCEDURE pdf_Content:
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfEncrypt AS LOGICAL   NO-UNDO.

  DEFINE VARIABLE L_font        AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_StreamKey   AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_SourceMem   AS MEMPTR NO-UNDO.
  DEFINE VARIABLE L_DestMem     AS MEMPTR NO-UNDO.
  DEFINE VARIABLE L_EncryptMem  AS MEMPTR NO-UNDO.
  DEFINE VARIABLE L_EncryptKey  AS MEMPTR NO-UNDO.

  DEFINE VARIABLE L_Loop        AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Page        AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Stat        AS INTEGER NO-UNDO.

  DEFINE VARIABLE vDest         AS MEMPTR NO-UNDO.
  DEFINE VARIABLE vDestLen      AS INTEGER NO-UNDO.
  DEFINE VARIABLE vSource       AS MEMPTR NO-UNDO.

  /* Produce each Page one at a time */
  DO L_Loop = 1 TO pdf_Page(pdfStream):
    IF CAN-FIND(FIRST TT_pdf_content WHERE TT_pdf_content.obj_page = L_Loop)
    THEN DO:

      L_page = L_page + 1.
      /* Start Page Definition */
      RUN pdf_Definition (pdfStream ,L_page, IF L_page = 1 THEN FALSE ELSE TRUE).
      ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Content", 0).

      PUT STREAM S_pdf_inc UNFORMATTED
          pdf_inc_ObjectSequence " 0 obj" {&pdfSKIP}
          "<<" {&pdfSKIP}
          "/Length " (pdf_inc_ObjectSequence + 1) " 0 R" {&pdfSKIP}.

      /* Only add the Filter if 'Compress' is turned on */
      IF pdf_get_parameter(pdfStream,"Compress") = "TRUE" THEN
        PUT STREAM S_pdf_inc UNFORMATTED
            "/Filter /FlateDecode" {&pdfSKIP}.

      PUT STREAM S_pdf_inc UNFORMATTED
          ">>" {&pdfSKIP}
          "stream" {&pdfSKIP} .

      pdf-Stream-Start = SEEK(S_pdf_inc).

      SET-SIZE(mContent) = 1.

      OUTPUT TO VALUE(SESSION:TEMP-DIR + TT_pdf_stream.obj_UniqueID  + "-C.txt") NO-ECHO BINARY NO-CONVERT.
      /* Process Watermark Content */
      IF CAN-FIND(FIRST TT_pdf_content
                  WHERE TT_pdf_content.obj_stream  = pdfStream
                    AND TT_pdf_content.obj_page    = L_page
                    AND TT_pdf_content.obj_type    = "WATERMARK" NO-LOCK)
      THEN DO:
        FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream  = pdfStream
                                  AND TT_pdf_content.obj_page    = L_page
                                  AND TT_pdf_content.obj_type    = "WATERMARK"
                                  NO-LOCK
                                  BREAK BY TT_pdf_content.obj_seq:
          TT_pdf_content.obj_content = TT_pdf_content.obj_content + CHR(13).

          RUN AddCharContent(TT_pdf_content.obj_content).
        END.
      END. /* Can find Watermark Content */

      /* Process Rectangle Content */
      IF CAN-FIND(FIRST TT_pdf_content
                  WHERE TT_pdf_content.obj_stream  = pdfStream
                    AND TT_pdf_content.obj_page    = L_page
                    AND TT_pdf_content.obj_type BEGINS "GRAPHIC" NO-LOCK)
      THEN DO:

        RUN AddCharContent("q" + CHR(13)).

        FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream      = pdfStream
                                  AND TT_pdf_content.obj_page        = L_page
                                  AND TT_pdf_content.obj_type BEGINS "GRAPHIC"
                                  NO-LOCK
                                  BREAK BY TT_pdf_content.obj_seq:

          TT_pdf_content.obj_content = TT_pdf_content.obj_content + CHR(13).

          RUN AddCharContent(TT_pdf_content.obj_content).
        END.

        RUN AddCharContent("Q" + CHR(13)).

      END. /* Cand find Graphic Content */

      /* Process Image Content */
      FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream      = pdfStream
                                AND TT_pdf_content.obj_page        = L_page
                                AND TT_pdf_content.obj_type BEGINS "IMAGE"
                                NO-LOCK
                                BREAK BY TT_pdf_content.obj_seq:
        RUN AddCharContent ("q " + TT_pdf_content.obj_content + " Q" + CHR(13)).
      END.

      /* Now Process Text Content */
      RUN AddCharContent(  CHR(13) + "BT" + CHR(13)
                         + Vuse-font + " " + dec2string(pdf_PointSize(pdfStream)) + " Tf" + CHR(13)
                         + "1 0 0 1 " + STRING(pdf_LeftMargin(pdfStream)) + " "
                         + STRING((pdf_PageHeight(pdfStream) - pdf_TopMargin(pdfStream)))
                         + " Tm" + CHR(13)).

      FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream  = pdfStream
                                AND TT_pdf_content.obj_page    = L_page
                                AND TT_pdf_content.obj_type BEGINS "TEXT"
                                AND TT_pdf_content.obj_type <> "TEXTXY" NO-LOCK
          BREAK BY TT_pdf_content.obj_line DESC
                BY TT_pdf_content.obj_seq:

        /* Replace the @@TotalPages call with the actual Total Page # */
        IF TT_pdf_Content.obj_function THEN
          TT_pdf_content.obj_content = REPLACE(TT_pdf_content.obj_content,
                                               "@@TotalPages-" + pdfStream,
                                               STRING(pdf_Page(pdfStream))).

        TT_pdf_content.obj_content = TT_pdf_content.obj_content + CHR(13).
        RUN AddCharContent(TT_pdf_content.obj_content).

      END. /* each text object on page - not-TEXTXY*/

      RUN AddCharContent(  "ET" + CHR(13)
                         + "BT" + CHR(13)
                         + Vuse-font + " " + STRING(pdf_PointSize(pdfStream)) + " Tf" + CHR(13)
                         + "1 0 0 1 " + STRING(pdf_LeftMargin(pdfStream)) + " "
                         + STRING(pdf_PageHeight(pdfStream) - pdf_TopMargin(pdfStream)) + " Tm"
                         + CHR(13)).

      FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream  = pdfStream
                                AND TT_pdf_content.obj_page    = L_page
                                AND TT_pdf_content.obj_type    = "TEXTXY" NO-LOCK
          BREAK BY TT_pdf_content.obj_line DESC
                BY TT_pdf_content.obj_seq:

        /* Replace the @@TotalPages call with the actual Total Page # */
        IF TT_pdf_Content.obj_function THEN
          TT_pdf_content.obj_content = REPLACE(TT_pdf_content.obj_content,
                                               "@@TotalPages-" + pdfStream,
                                               STRING(pdf_Page(pdfStream))).

        TT_pdf_content.obj_content = TT_pdf_content.obj_content + CHR(13).
        RUN AddCharContent(TT_pdf_content.obj_content).

      END. /* each textxy object on page */

      RUN AddCharContent("ET").

      OUTPUT CLOSE.

      FILE-INFO:FILE-NAME = SESSION:TEMP-DIR + TT_pdf_stream.obj_UniqueID  + "-C.txt".

      SET-SIZE(mContent) = 0.
      SET-SIZE(mContent) = FILE-INFO:FILE-SIZE.
      INPUT FROM VALUE(FILE-INFO:FILE-NAME) NO-ECHO BINARY NO-CONVERT.
        IMPORT mContent.
      INPUT CLOSE.

      mHolder = mContent.

      /** Compression happens before Encryption **/
      IF pdf_get_parameter(pdfStream,"Compress") = "TRUE" THEN DO:
        L_Stat = compressbuffer( mHolder,
                                 INPUT-OUTPUT vDest,
                                 OUTPUT vDestLen).

        SET-SIZE(mHolder) = 0.
        SET-SIZE(mHolder) = GET-SIZE(vDest).
        mHolder = Vdest.
      END.

      IF NOT pdfEncrypt THEN
        EXPORT STREAM S_pdf_inc mHolder.

      ELSE DO:
        SET-SIZE(L_EncryptKey) = 10.
        RUN GetEncryptKey IN h_PDF-Encrypt
            (INPUT  TT_pdf_stream.obj_UniqueID,
             INPUT  pdf_inc_ObjectSequence,
             INPUT  0,
             INPUT  pdf-EncryptKeyMemPtr,
             OUTPUT L_EncryptKey).


        RUN EncryptContent IN h_PDF-Encrypt
                           (INPUT  L_EncryptKey,
                            INPUT  mHolder,
                            OUTPUT L_EncryptMem).

        RUN OutputMemPtr (L_EncryptMem).

        SET-SIZE(L_EncryptMem) = 0.
        SET-SIZE(L_EncryptKey) = 0.

      END. /* Encrypted Content */

      SET-SIZE(mHolder) = 0.
      SET-SIZE(mContent) = 0.

      /* igc - left for compression debugging
      RUN PutMemPtr (vDest, "c:\gord\zlib\dest.txt").

      L_Stat = decompressbuffer( vDest,
                                 OUTPUT vSource,
                                 OUTPUT vDestLen).

      RUN PutMemPtr (vSource, "c:\gord\zlib\source.txt").
      */

      SET-SIZE(vSource) = 0.
      SET-SIZE(vDest)   = 0.

      /* End Page Definition */
      pdf-Stream-End = SEEK(S_pdf_inc).
      PUT STREAM S_pdf_inc UNFORMATTED
          "endstream" {&pdfSKIP}
          "endobj" {&pdfSKIP}.

      /* Output Length */
      RUN pdf_length (pdfStream, pdf-Stream-End - pdf-Stream-Start).

    END. /* Can-find page content */

  END. /* loop for each page */

  OS-DELETE VALUE(SESSION:TEMP-DIR + TT_pdf_stream.obj_UniqueID + "-C.txt").

  SET-SIZE(mHolder) = 0.
  SET-SIZE(mContent) = 0.
  SET-SIZE(vSource) = 0.
  SET-SIZE(vDest)   = 0.

  /* Load Bookmarks */
  IF CAN-FIND( FIRST TT_pdf_bookmark WHERE TT_pdf_bookmark.obj_Stream = pdfStream NO-LOCK)
  THEN RUN pdf_Load_Bookmarks (pdfStream).

  /* This will set the PDF Page to the max actual Page number */
  RUN pdf_set_Page(pdfStream, L_page).

  RUN pdf_catalog (pdfStream).
  RUN pdf_ListPages (pdfStream).

END. /* pdf_Content */

PROCEDURE pdf_definition:
  DEFINE INPUT PARAMETER pdfStream      AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER P_page         AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER P_incl-annot   AS LOGICAL NO-UNDO.

  FIND TT_pdf_page WHERE TT_pdf_page.obj_stream = pdfStream
                     AND TT_pdf_page.page_nbr   = p_Page NO-ERROR.
  IF NOT AVAIL TT_pdf_page THEN DO:
    MESSAGE "Error occurred - could not find TT_pdf_page for Page=" p_Page
            VIEW-AS ALERT-BOX.
    RETURN.
  END.

  ObjectSequence (pdfStream, pdf_inc_ObjectSequence + 1, "PageDefinition", p_Page).
  PUT STREAM S_pdf_inc UNFORMATTED
      pdf_inc_ObjectSequence  " 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "/Type /Page" {&pdfSKIP}
      "/Parent 3 0 R" {&pdfSKIP}
      "/Resources " pdf-Res-Object " 0 R" {&pdfSKIP}
      "/Contents " (pdf_inc_ObjectSequence + 1) " 0 R" {&pdfSKIP}
      "/Rotate " STRING(TT_pdf_page.page_rotate) {&pdfSKIP}.

  IF TT_pdf_page.page_rotate = 0 OR TT_pdf_page.page_rotate = 180 THEN
    PUT STREAM S_pdf_inc UNFORMATTED
        "/MediaBox [ 0 0 " pdf_PageWidth(pdfStream) " " pdf_PageHeight(pdfStream)
        "]" {&pdfSKIP}.
  ELSE
    PUT STREAM S_pdf_inc UNFORMATTED
        "/MediaBox [ 0 0 " pdf_PageHeight(pdfStream) " " pdf_PageWidth(pdfStream)
        "]" {&pdfSKIP}.

  PUT STREAM S_pdf_inc UNFORMATTED
      "/Annots [ " {&pdfSKIP}.


      /* Output Link Definitions */
  FOR EACH TT_pdf_link WHERE TT_pdf_link.obj_stream = pdfStream
                         AND TT_pdf_link.link_page  = p_Page:
    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_link.link_obj " 0 R" {&pdfSKIP}.
  END.

  PUT STREAM S_pdf_inc UNFORMATTED
      " ]" {&pdfSKIP}
      ">>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_definition */

PROCEDURE pdf_Length:

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER P_length   AS INTEGER NO-UNDO.

  ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Length", 0).
  PUT STREAM S_pdf_inc UNFORMATTED
      pdf_inc_ObjectSequence " 0 obj" {&pdfSKIP}
      P_length {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_Length */

PROCEDURE pdf_Catalog :
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  ObjectSequence(pdfStream, 2, "Catalog", 0).
  PUT STREAM S_pdf_inc UNFORMATTED
      "2 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "/Type /Catalog" {&pdfSKIP}
      "/Pages 3 0 R" {&pdfSKIP}
      "/Outlines " pdf_OutlinesDict " 0 R" {&pdfSKIP}
      ">>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_Catalog */

PROCEDURE pdf_ListPages :

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  ObjectSequence(pdfStream, 3, "Pages", 0).
  PUT STREAM S_pdf_inc UNFORMATTED
      "3 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "/Type /Pages" {&pdfSKIP}
      "/Count " pdf_Page(pdfStream) {&pdfSKIP}
      "/MediaBox [ 0 0 " pdf_PageWidth(pdfStream)
      " " pdf_PageHeight(pdfStream) " ]" {&pdfSKIP}
      "/Kids [ " {&pdfSKIP}.

  FOR EACH TT_pdf_object WHERE TT_pdf_object.obj_stream = pdfStream
                           AND TT_pdf_object.obj_desc = "Content" NO-LOCK:
    PUT STREAM S_pdf_inc UNFORMATTED
        (TT_pdf_object.obj_nbr - 1) " 0 R " {&pdfSKIP}.
  END. /* Display Pages */

  PUT STREAM S_pdf_inc UNFORMATTED
      "]"  {&pdfSKIP}
      ">>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_ListPages */

PROCEDURE pdf_xref :

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfEncrypt AS LOGICAL   NO-UNDO.

  DEFINE VARIABLE L_ctr AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_obj AS INTEGER NO-UNDO.

  ObjectSequence(pdfStream, 0, "Xref", 0).
  FOR EACH TT_pdf_object WHERE TT_pdf_object.obj_stream = pdfStream:
    L_ctr = L_ctr + 1.
  END.

  /* Get the Xref start point */
  pdf-Stream-Start = SEEK(S_pdf_inc).

  PUT STREAM S_pdf_inc UNFORMATTED
      "xref" {&pdfSKIP}
      "0 " L_ctr {&pdfSKIP}.

  FOR EACH TT_pdf_object WHERE TT_pdf_object.obj_stream = pdfStream
      BREAK BY TT_pdf_object.obj_nbr:
    IF FIRST( TT_pdf_object.obj_nbr) THEN
      PUT STREAM S_pdf_inc CONTROL
          "0000000000"
          " "
          TT_pdf_object.gen_nbr " "
          TT_pdf_object.obj_type CHR(13) CHR(10).
    ELSE DO:
      TT_pdf_object.obj_off = TT_pdf_object.obj_off - 1.
      PUT STREAM S_pdf_inc CONTROL
          STRING(TT_pdf_object.obj_off, "9999999999")
          " "
          STRING(TT_pdf_object.gen_nbr, "99999") " "
          TT_pdf_object.obj_type CHR(13) CHR(10).
    END.
  END. /* each Object */

  FIND TT_pdf_stream WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK NO-ERROR.

  FIND LAST TT_pdf_object WHERE TT_pdf_object.obj_stream = pdfStream NO-LOCK NO-ERROR.

  PUT STREAM S_pdf_inc UNFORMATTED
      "trailer" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "/Size " (TT_pdf_object.obj_nbr + 1) {&pdfSKIP}
      "/Root 2 0 R" {&pdfSKIP}
      "/Info 1 0 R" {&pdfSKIP}.

  IF pdfEncrypt THEN DO:
    PUT STREAM S_pdf_inc UNFORMATTED
        "/Encrypt " TT_pdf_stream.obj_EncryptDict " 0 R" {&pdfSKIP}
        "/ID [<" TT_pdf_stream.obj_ID "><" TT_pdf_stream.obj_ID ">]" {&pdfSKIP}.
  END.

  PUT STREAM S_pdf_inc UNFORMATTED
      ">>" {&pdfSKIP}
      "startxref" {&pdfSKIP}
      pdf-Stream-Start {&pdfSKIP}
      "%%EOF" {&pdfSKIP}.

END. /* pdf_xref */

PROCEDURE pdf_Fonts:

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  FOR EACH TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                         AND TT_pdf_font.font_file  = "PDFBASE14"
      BY TT_pdf_font.font_obj:

    ObjectSequence(pdfStream, TT_pdf_font.font_obj, "Font", 0).
    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_font.FONT_obj " 0 obj" {&pdfSKIP}
        "<<" {&pdfSKIP}
        "/Type /Font" {&pdfSKIP}
        "/Subtype /Type1" {&pdfSKIP}
        "/Name " TT_pdf_font.font_tag {&pdfSKIP}.

    IF TT_pdf_font.font_name= "Symbol" 
    OR TT_pdf_font.font_name = "ZapfDingbats" THEN
      PUT STREAM S_pdf_inc UNFORMATTED
          "/Encoding << /BaseEncoding /StandardEncoding >>" {&pdfSKIP}.
    ELSE
      PUT STREAM S_pdf_inc UNFORMATTED
        "/Encoding 4 0 R" {&pdfSKIP}.

      PUT STREAM S_pdf_inc UNFORMATTED
        "/BaseFont /" TT_pdf_font.font_name {&pdfSKIP}
        ">>" {&pdfSKIP}
        "endobj" {&pdfSKIP}.
  END.

END. /* pdf_Fonts */

PROCEDURE pdf_Load_fonts :
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfEncrypt AS LOGICAL   NO-UNDO.

  DEFINE VARIABLE L_data  AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_count AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_start AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_end   AS INTEGER NO-UNDO.

  DEFINE VARIABLE mFont         AS MEMPTR NO-UNDO.
  DEFINE VARIABLE m_EncryptKey  AS MEMPTR NO-UNDO.

  FOR EACH TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                         AND TT_pdf_font.font_file <> "PDFBASE14":

    FILE-INFO:FILE-NAME = TT_pdf_font.font_file.

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "FontDescriptor", 0).
    TT_pdf_font.font_descr  = pdf_inc_ObjectSequence.

    /* Output the Font Descriptor */
    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_font.font_descr " 0 obj" {&pdfSKIP}
        "<< /Type /FontDescriptor" {&pdfSKIP}
        "   /Ascent " TT_pdf_font.afm_Ascender {&pdfSKIP}
        "   /Descent " TT_pdf_font.afm_Descender {&pdfSKIP}
        "   /Flags " TT_pdf_font.afm_Flags {&pdfSKIP}
        "   /FontBBox [ " TT_pdf_font.afm_FontBBox " ]" {&pdfSKIP}
        "   /FontName /" TT_pdf_font.font_name {&pdfSKIP}
        "   /ItalicAngle " TT_pdf_font.afm_ItalicAngle  {&pdfSKIP}
        "   /FontFile2 " (TT_pdf_font.font_descr + 3) " 0 R" {&pdfSKIP}
        ">>" {&pdfSKIP}
        "endobj" {&pdfSKIP}.

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Font", 0).
    TT_pdf_font.font_obj    = pdf_inc_ObjectSequence.

    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_font.font_obj " 0 obj" {&pdfSKIP}
        "<<" {&pdfSKIP}
        "/Type /Font" {&pdfSKIP}
        "/Subtype /TrueType" {&pdfSKIP}
        "/FirstChar " TT_pdf_font.afm_FirstChar {&pdfSKIP}
        "/LastChar " TT_pdf_font.afm_LastChar {&pdfSKIP}
        "/Widths [ " TT_pdf_font.afm_widths " ]" {&pdfSKIP}
        "/Encoding " (pdf_inc_ObjectSequence + 1) " 0 R" {&pdfSKIP}
        "/BaseFont /" TT_pdf_font.font_name {&pdfSKIP}
        "/FontDescriptor " TT_pdf_font.font_descr " 0 R" {&pdfSKIP}
        ">>" {&pdfSKIP}
        "endobj" {&pdfSKIP}.

    /* igc - Aug 28 - Added this code to allow for remapping of characters */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Encoding", 0).
    TT_pdf_font.font_encoding  = pdf_inc_ObjectSequence.

    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_font.font_encoding " 0 obj" {&pdfSKIP}
        "<< /Type /Encoding" {&pdfSKIP}
        "/BaseEncoding /WinAnsiEncoding" {&pdfSKIP}.

    IF TT_pdf_font.font_dif <> ""
    OR CAN-FIND(FIRST TT_pdf_diff
                WHERE TT_pdf_diff.obj_stream = pdfStream
                  AND TT_pdf_diff.font_name  = TT_pdf_font.font_name NO-LOCK)
    THEN DO:
      PUT STREAM S_pdf_inc UNFORMATTED "/Differences [ " {&pdfSKIP}.

      IF TT_pdf_font.font_dif <> "" THEN DO:
        INPUT STREAM S_pdf_inp FROM VALUE(SEARCH(TT_pdf_font.font_dif)) NO-CONVERT NO-MAP NO-ECHO.

          REPEAT:
            IMPORT STREAM S_pdf_inp UNFORMATTED L_data.
            PUT STREAM S_pdf_inc UNFORMATTED
                L_Data {&pdfSKIP}.
          END.

        INPUT STREAM S_pdf_inp CLOSE.
      END. /* Differences File */

      FOR EACH TT_pdf_diff WHERE TT_pdf_diff.obj_stream = pdfStream:
        PUT STREAM S_pdf_inc UNFORMATTED
            TT_pdf_diff.char_num
            " "
            TT_pdf_diff.PS_name {&pdfSKIP}.
      END.

      PUT STREAM S_pdf_inc UNFORMATTED "]". /* Close the Differences Array */

    END. /* Font Difference File exists */

    PUT STREAM S_pdf_inc UNFORMATTED
        ">>" {&pdfSKIP}
        "endobj" {&pdfSKIP}.

    /* igc - Aug 28 - end of character mapping code */

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "FontStream", 0).
    TT_pdf_font.font_stream = pdf_inc_ObjectSequence.

    /* Display Embedded Font Stream */
    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_font.font_stream " 0 obj" {&pdfSKIP}
        "<< /Length " (TT_pdf_font.font_stream + 1) " 0 R /Length1 "
                      (TT_pdf_font.font_stream + 1) " 0 R >>" {&pdfSKIP}
                      "stream" {&pdfSKIP}.

    /* Get PDF Stream Start Offset */
    L_start = SEEK(S_pdf_inc).

    SET-SIZE(mFont) = FILE-INFO:FILE-SIZE. /* Initialize Memory */
    INPUT FROM VALUE(SEARCH(FILE-INFO:FILE-NAME)) NO-ECHO NO-CONVERT BINARY.
      IMPORT mFont.
    INPUT CLOSE.

    IF pdfEncrypt THEN DO:
      SET-SIZE(m_EncryptKey) = 10.
      RUN GetEncryptKey IN h_PDF-Encrypt
          (INPUT  TT_pdf_stream.obj_UniqueID,
           INPUT  pdf_inc_ObjectSequence,
           INPUT  0,
           INPUT  pdf-EncryptKeyMemPtr,
           OUTPUT m_EncryptKey).

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  mFont,
                          OUTPUT mFont).

    END.

    RUN OutputMemPtr (mFont).

    SET-SIZE(mFont) = 0. /* Release memory */

    L_end = SEEK(S_pdf_inc).

    PUT STREAM S_pdf_inc UNFORMATTED
        {&pdfSKIP} {&pdfSKIP} "endstream" {&PDFSKIP}
                  "endobj" {&pdfSKIP}.

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "FontLength", 0).
    TT_pdf_font.font_len = pdf_inc_ObjectSequence.

    /* Put out Length */
    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_font.font_len " 0 obj" {&pdfSKIP}
        "  " (L_end - L_start) {&pdfSKIP}
        "endobj" {&pdfSKIP}.

    SET-SIZE(m_EncryptKey) = 0.

  END. /* each TTfont */

END. /* pdf_Load_fonts */

PROCEDURE pdf_ParseAFMFile:
  DEFINE OUTPUT PARAMETER P_afm_ItalicAngle  AS INTEGER   NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_Ascender     AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_Descender    AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_FontBBox     AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_FirstChar    AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_LastChar     AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_Widths       AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER P_afm_IsFixedPitch AS CHARACTER NO-UNDO INIT "0".
  DEFINE OUTPUT PARAMETER P_afm_flags        AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_data  AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_key   AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_flag  AS CHARACTER NO-UNDO
         INIT "00000000000000000000000000100010".
  /* Bit 6 (above) is set to identify NonSymbolic Fonts -- or Fonts that use
     the Standard Latin Character Set */

  DEFINE VARIABLE L_int   AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Loop  AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_exp   AS INTEGER NO-UNDO.

  ASSIGN P_afm_ItalicAngle  = 0
         P_afm_Descender    = ""
         P_afm_FontBBox     = ""
         P_afm_IsFixedPitch = ""
         P_afm_FirstChar    = ""
         P_afm_LastChar     = ""
         P_afm_Widths       = "".

  INPUT STREAM S_pdf_inp FROM VALUE(SEARCH(TT_pdf_font.font_afm)) BINARY NO-CONVERT NO-MAP NO-ECHO.

    REPEAT:
      IMPORT STREAM S_pdf_inp UNFORMATTED L_data.
      L_Key = ENTRY(1, L_data, " ") NO-ERROR.
      IF ERROR-STATUS:ERROR THEN NEXT.

      CASE L_key:
        WHEN "ItalicAngle" THEN
          P_afm_ItalicAngle = INT(ENTRY( 2, L_data, " ")) NO-ERROR.

        WHEN "Ascender" THEN
          P_afm_Ascender = ENTRY( 2, L_data, " ") NO-ERROR.

        WHEN "Descender" THEN
          P_afm_Descender = ENTRY( 2, L_data, " ") .

        WHEN "FontBBox" THEN
          ASSIGN P_afm_FontBBox = REPLACE(L_data,"FontBBox ","").

        WHEN "IsFixedPitch" THEN
          P_afm_IsFixedPitch = IF ENTRY(2,L_data, " ") = "True" THEN "0" ELSE "1".

        WHEN "C" THEN DO:
          IF P_afm_FirstChar = "" THEN
            P_afm_FirstChar = ENTRY(2, L_data, " ") NO-ERROR.

          ASSIGN P_afm_Widths = P_afm_widths + " "
                              + ENTRY(5, L_data, " ") NO-ERROR.

          P_afm_LastChar = ENTRY(2, L_data, " ") NO-ERROR.
        END.
      END CASE.
    END. /* REPEAT */

  INPUT STREAM S_pdf_inp CLOSE.

  /* Determine Font Flags */
  IF P_afm_IsFixedPitch = "0" THEN
    OVERLAY( L_Flag, 32, 1, "CHARACTER") = "1".

  DO L_loop = LENGTH(L_Flag, "character":u) TO 1 BY -1 :
    IF SUBSTR(L_flag,L_loop,1, "character":u) = "1" THEN
      L_int = L_int + EXP(2, L_exp).

    L_exp = L_exp + 1.
  END.

  P_afm_Flags = STRING( L_int ).

END. /* pdf_ParseAFMFile */

PROCEDURE pdf_load_images:

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfEncrypt AS LOGICAL   NO-UNDO.

  DEFINE VARIABLE L_data  AS RAW NO-UNDO.
  DEFINE VARIABLE L_count AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_start AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_end   AS INTEGER NO-UNDO.

  DEFINE VARIABLE m_EncryptKey  AS MEMPTR NO-UNDO.
  DEFINE VARIABLE mImage        AS MEMPTR  NO-UNDO.

  FOR EACH TT_pdf_image WHERE TT_pdf_image.obj_stream = pdfStream:

    FILE-INFO:FILE-NAME = TT_pdf_image.image_file.

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Image", 0).
    TT_pdf_image.image_obj = pdf_inc_ObjectSequence.

    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_image.image_obj " 0 obj" {&pdfSKIP}
        "<<" {&pdfSKIP}
        "/Type /XObject" {&pdfSKIP}
        "/Subtype /Image" {&pdfSKIP}
        "/Name " TT_pdf_image.image_tag {&pdfSKIP}
        "/Width " TT_pdf_image.image_w {&pdfSKIP}
        "/Height " TT_pdf_image.image_h {&pdfSKIP}
        "/BitsPerComponent 8" {&pdfSKIP}
        "/ColorSpace /DeviceRGB" {&pdfSKIP}
        "/Length " (TT_pdf_image.image_obj + 1) " 0 R" {&pdfSKIP}
        "/Filter /DCTDecode" {&pdfSKIP}
        ">>" {&pdfSKIP}
        "stream" {&pdfSKIP}.

    /* Get PDF Stream Start Offset */
    L_start = SEEK(S_pdf_inc).

    SET-SIZE(mImage) = FILE-INFO:FILE-SIZE. /* Initialize Memory */
    INPUT FROM VALUE(SEARCH(FILE-INFO:FILE-NAME)) NO-ECHO NO-CONVERT BINARY.
      IMPORT mImage.
    INPUT CLOSE.

    IF pdfEncrypt THEN DO:

      SET-SIZE(m_EncryptKey) = 10.
      RUN GetEncryptKey IN h_PDF-Encrypt
          (INPUT  TT_pdf_stream.obj_UniqueID,
           INPUT  pdf_inc_ObjectSequence,
           INPUT  0,
           INPUT  pdf-EncryptKeyMemPtr,
           OUTPUT m_EncryptKey).

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  mImage,
                          OUTPUT mImage).
    END.

    RUN OutputMemPtr (mImage).

    SET-SIZE(mImage) = 0. /* Release memory */

    L_end = SEEK(S_pdf_inc).

    PUT STREAM S_pdf_inc UNFORMATTED
        {&pdfSKIP} {&pdfSKIP} "endstream" {&pdfSKIP}
                    "endobj" {&pdfSKIP}.

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "ImageLen", 0).
    TT_pdf_image.image_len = pdf_inc_ObjectSequence.

    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_image.image_len " 0 obj" {&pdfSKIP}
        "  " (L_end - L_start) {&pdfSKIP}
        "endobj" {&pdfSKIP}.

    SET-SIZE(m_EncryptKey) = 0.

  END. /* each TT_pdf_image */

END. /* pdf_load_images */

PROCEDURE pdf_load_links:

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfEncrypt AS LOGICAL   NO-UNDO.

  DEFINE VARIABLE L_data  AS RAW NO-UNDO.
  DEFINE VARIABLE L_start AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_end   AS INTEGER NO-UNDO.

  DEFINE VARIABLE m_EncryptKey  AS MEMPTR NO-UNDO.
  DEFINE VARIABLE mLink         AS MEMPTR NO-UNDO.

  FOR EACH TT_pdf_link WHERE TT_pdf_link.obj_stream = pdfStream:

    /* igc - Added Sept 10, 2002 */
    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Link", 0).
    TT_pdf_link.link_obj = pdf_inc_ObjectSequence.

    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_link.link_obj " 0 obj" CHR(13)
        "<<" CHR(13)
        "/Type /Annot" CHR(13)
        "/Subtype /Link" CHR(13)
        "/Rect [ " STRING(TT_pdf_link.link_x) " "
                   STRING(TT_pdf_link.link_y) " "
                   STRING(TT_pdf_link.link_w) " "
                   STRING(TT_pdf_link.link_h) " "
                   "]" CHR(13)
        "/Border [0 0 " TT_pdf_link.link_border "]" CHR(13)
        "/H /" TT_pdf_link.link_style CHR(13)
        "/C [ " TT_pdf_link.link_color " ]" CHR(13)
        "/A << " CHR(13)
        "  /Type /Action" CHR(13)
        "  /S /URI" CHR(13)
        "  /URI ".

    IF pdfEncrypt THEN DO:
      SET-SIZE(m_EncryptKey) = 10.
      RUN GetEncryptKey IN h_PDF-Encrypt
          (INPUT  TT_pdf_stream.obj_UniqueID,
           INPUT  pdf_inc_ObjectSequence,
           INPUT  0,
           INPUT  pdf-EncryptKeyMemPtr,
           OUTPUT m_EncryptKey).

      SET-SIZE(mLink) = LENGTH(TT_pdf_link.link_name, "character":u) + 1.
      PUT-STRING(mlink,1) = TT_pdf_link.link_name.

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  mLink,
                          OUTPUT mLink).

      PUT STREAM S_pdf_inc UNFORMATTED "<".

      RUN OutputMemPtrAsHex (mLink).
      SET-SIZE(mLink) = 0.

      PUT STREAM S_pdf_inc UNFORMATTED ">" CHR(13).
    END.

    ELSE DO:
      RUN pdf_replace_text(INPUT-OUTPUT TT_pdf_link.link_name).

      PUT STREAM S_pdf_inc UNFORMATTED
        "(" TT_pdf_link.link_Name ")" CHR(13).
    END.

    PUT STREAM S_pdf_inc UNFORMATTED
        "  >> " CHR(13)
        ">>" CHR(13)
        "endobj" CHR(13).

    SET-SIZE(m_EncryptKey) = 0.

  END. /* each TT_pdf_link */

END. /* pdf_load_links */

PROCEDURE pdf_get_image_wh:
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER Pimage     AS CHARACTER NO-UNDO.

  DEFINE VARIABLE pdf_marker    AS INTEGER NO-UNDO.

  INPUT STREAM S_pdf_inp FROM VALUE(SEARCH(Pimage)) BINARY NO-MAP NO-CONVERT UNBUFFERED.

    IF NOT first_marker() THEN DO:
      RUN pdf_error(pdfStream,"pdf_get_Image_wh","Cannot find Stream!").
      RETURN.
    END.

    DO WHILE TRUE:
      pdf_marker = next_marker().

      CASE hex(pdf_marker):
        WHEN {&M_SOF0} OR WHEN {&M_SOF1} OR WHEN {&M_SOF2} OR WHEN {&M_SOF3}
        OR WHEN {&M_SOF5} OR WHEN {&M_SOF6} OR WHEN {&M_SOF7} OR WHEN {&M_SOF9}
        OR WHEN {&M_SOF10} OR WHEN {&M_SOF11} OR WHEN {&M_SOF13}
        OR WHEN {&M_SOF14} OR WHEN {&M_SOF15} THEN DO:
          process_SOF().
          LEAVE.
        END.
        WHEN {&M_SOS} OR WHEN {&M_EOI} THEN
          LEAVE.
        OTHERWISE
          skip_variable().
      END CASE.
    END. /* true loop */

  INPUT STREAM S_pdf_inp CLOSE.

END.

PROCEDURE pdf_replace_text:
  DEFINE INPUT-OUTPUT PARAMETER pdfText AS CHARACTER NO-UNDO.

  /* Replace any special characters in the data string since this
     will create a bad PDF doccument */
  pdfText = REPLACE(pdfText,"~\","~~\\").
  pdfText = REPLACE(pdfText,"(","~\(").
  pdfText = REPLACE(pdfText,")","~\)").
  pdfText = REPLACE(pdfText,"[","~\[").
  pdfText = REPLACE(pdfText,"]","~\]").

END. /* pdf_replace_text */

PROCEDURE pdf_reset_all:
  /* clear out all streams, and reset variables as required */

  /* These are the only two variables that don't apear to be reset anywhere */
  ASSIGN pdf_inc_ContentSequence = 0
         pdf_inc_ObjectSequence  = 0.

  /* Clear all temp-tables */
  &IF NOT PROVERSION BEGINS "8" &THEN /* Assume progress versions 8 or above */
    EMPTY TEMP-TABLE TT_pdf_stream.
    EMPTY TEMP-TABLE TT_pdf_page.
    EMPTY TEMP-TABLE TT_pdf_bookmark.
    EMPTY TEMP-TABLE TT_pdf_ReplaceTxt.
    EMPTY TEMP-TABLE TT_pdf_param.
    EMPTY TEMP-TABLE TT_pdf_error.
    EMPTY TEMP-TABLE TT_pdf_object.
    EMPTY TEMP-TABLE TT_pdf_content.
    EMPTY TEMP-TABLE TT_pdf_info.
    EMPTY TEMP-TABLE TT_pdf_image.
    EMPTY TEMP-TABLE TT_pdf_font.
    EMPTY TEMP-TABLE TT_pdf_link.
  &ELSE
    FOR EACH TT_pdf_stream:
      DELETE TT_pdf_stream .
    END.

    FOR EACH TT_pdf_page:
      DELETE TT_pdf_page.
    END.

    FOR EACH TT_pdf_bookmark:
      DELETE TT_pdf_bookmark.
    END.

    FOR EACH TT_pdf_ReplaceTxt:
      DELETE TT_pdf_ReplaceTxt.
    END.

    FOR EACH TT_pdf_param:
      DELETE TT_pdf_param.
    END.

    FOR EACH TT_pdf_error:
      DELETE TT_pdf_error.
    END.

    FOR EACH TT_pdf_object:
      DELETE TT_pdf_object.
    END.

    FOR EACH TT_pdf_content:
      DELETE TT_pdf_content.
    END.

    FOR EACH TT_pdf_info:
      DELETE TT_pdf_info.
    END.

    FOR EACH TT_pdf_image:
      DELETE TT_pdf_image.
    END.

    FOR EACH TT_pdf_font:
      DELETE TT_pdf_font.
    END.

    FOR EACH TT_pdf_link:
      DELETE TT_pdf_link.
    END.

  &ENDIF
END. /* pdf_reset_all */

PROCEDURE pdf_reset_stream .
  /* Clear out an individual stream - reset the variables */
  DEFINE INPUT PARAMETER pdfStream     AS CHARACTER NO-UNDO.

  /* These are the only two variables that don't apear to be reset anywhere */
  ASSIGN pdf_inc_ContentSequence = 0
         pdf_inc_ObjectSequence  = 0.

  /* As far as I know, you gotta do a for each regardless of version */
  FOR EACH TT_pdf_stream WHERE TT_pdf_stream.obj_stream = pdfStream:
    DELETE TT_pdf_stream .
  END.

  FOR EACH TT_pdf_page WHERE TT_pdf_page.obj_stream = pdfStream:
    DELETE TT_pdf_page.
  END.

  FOR EACH TT_pdf_bookmark WHERE TT_pdf_bookmark.obj_stream = pdfStream:
    DELETE TT_pdf_bookmark.
  END.

  FOR EACH TT_pdf_ReplaceTxt WHERE TT_pdf_param.obj_stream = pdfStream:
    DELETE TT_pdf_ReplaceTxt.
  END.

  FOR EACH TT_pdf_param WHERE TT_pdf_param.obj_stream = pdfStream:
    DELETE TT_pdf_param.
  END.

  FOR EACH TT_pdf_error WHERE TT_pdf_error.obj_stream = pdfStream:
    DELETE TT_pdf_error.
  END.

  FOR EACH TT_pdf_object WHERE TT_pdf_object.obj_stream = pdfStream:
    DELETE TT_pdf_object.
  END.

  FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream = pdfStream:
    DELETE TT_pdf_content.
  END.

  FOR EACH TT_pdf_info WHERE TT_pdf_info.obj_stream = pdfStream:
    DELETE TT_pdf_info.
  END.

  FOR EACH TT_pdf_image WHERE TT_pdf_image.obj_stream = pdfStream:
    DELETE TT_pdf_image.
  END.

  FOR EACH TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream:
    DELETE TT_pdf_font.
  END.

  FOR EACH TT_pdf_link WHERE TT_pdf_link.obj_stream = pdfStream:
    DELETE TT_pdf_link.
  END.

END. /* pdf_reset_stream */

PROCEDURE pdf_wrap_text :
/*------------------------------------------------------------------------------
  Purpose:
  Parameters:  <none>
  Notes:
------------------------------------------------------------------------------*/
  DEFINE INPUT PARAMETER pdfStream     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfText       AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromColumn AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToColumn   AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pdfAlignMent  AS CHARACTER NO-UNDO.
  DEFINE OUTPUT PARAMETER pdfMaxY      AS INTEGER NO-UNDO.

  DEFINE VAR v-thisline  AS CHAR NO-UNDO.
  DEFINE VAR v-lastline  AS CHAR NO-UNDO.
  DEFINE VAR v-fontsize  AS DECIMAL NO-UNDO.
  DEFINE VAR v-mywidth   AS INTEGER NO-UNDO.
  DEFINE VAR v-maxwidth  AS INTEGER NO-UNDO.
  DEFINE VAR v-originalY AS INTEGER NO-UNDO.
  DEFINE VAR i           AS INTEGER NO-UNDO.
  DEFINE VAR v-index     AS INTEGER NO-UNDO.
  DEFINE VAR v-start     AS INTEGER NO-UNDO.
  DEFINE VAR v-leave     AS LOGICAL NO-UNDO.

  ASSIGN v-fontsize  = PDF_text_width(pdfStream,"i") * .81
         v-mywidth   = pdfToColumn - pdfFromColumn
         v-originalY = PDF_TextY(pdfStream).

  /* Replace Pipes in text string with processor directive &pipe; */
  ASSIGN pdfText = REPLACE(pdfText,"|","&pipe;").
  /* Spaces */
  ASSIGN pdfText = REPLACE(pdfText," ","| ").
  /* Hyphens */
  ASSIGN pdfText = REPLACE(pdfText,"-","-|").
  /* Commas */
  ASSIGN pdfText = REPLACE(pdfText,",",",|").
  /* Commas */
  ASSIGN pdfText = REPLACE(pdfText,CHR(10),"|&skip;|").

  /* Determine the maximum possible column using the W (assuming it as the 
    largest possible character.  This is useful when using Proportional fonts -
    it won't matter when using Fixed since they are all the same width */
  v-maxwidth = pdf_getnumfittingchars(pdfStream,
                                       FILL("W",v-mywidth),
                                       0,
                                       pdf_text_width(pdfstream,FILL("W",v-mywidth))).

  /* Divide up the pdf text into lines of width less than the
     available width */
  THIS-LOOP:
  DO i = 1 TO NUM-ENTRIES(pdfText,"|"):
    ASSIGN v-lastline = v-thisline.
    IF ENTRY(i,pdfText,"|") = "&skip;" THEN DO:
      
      CASE pdfAlignment:
        WHEN "left" THEN
          RUN pdf_text_at(pdfStream,v-thisline,pdfFromColumn).
        WHEN "right" THEN
          RUN pdf_text_to(pdfStream,v-thisline,pdfToColumn).
      END CASE.

      v-thisline = "".
      RUN pdf_skip(pdfStream).
      NEXT THIS-LOOP.
    END.

    ASSIGN v-thisline = v-thisline
                      + REPLACE(ENTRY(i,pdfText,"|"),"&pipe;","|").

    IF LENGTH(v-thisline,"character":u) >= v-maxwidth THEN DO:

      ASSIGN v-index = 0
             v-start = 1
             v-leave = FALSE.

      ASSIGN v-lastline = TRIM(v-lastline).

      DO WHILE TRUE:
        v-index = INDEX(v-lastline,CHR(10), v-start + 1).

        IF v-index = 0 THEN
          ASSIGN v-leave = TRUE
                 v-index = LENGTH(v-lastline, "character":u).
        ELSE v-leave = FALSE.

        CASE pdfAlignment:
          WHEN "left" THEN
            RUN pdf_text_at(pdfStream,SUBSTR(v-lastline,v-start,v-index, "character":u),pdfFromColumn).
          WHEN "right" THEN
            RUN pdf_text_to(pdfStream,SUBSTR(v-lastline,v-start,v-index, "character":u),pdfToColumn).
        END CASE.

        IF v-leave THEN LEAVE.

        ELSE DO:
          v-start = INDEX(v-lastline,CHR(10), v-index + 1).
          RUN pdf_skip(pdfStream).
        END.

      END. /* while true */

      RUN pdf_skip(pdfStream).
      ASSIGN v-thisline = TRIM(REPLACE(ENTRY(i,pdfText,"|"),"&pipe;","|")).
    END.
  END. /* i = NUM-ENTRIES */

  IF v-thisline NE "" THEN DO:
    ASSIGN v-index = 0
           v-start = 1
           v-leave = FALSE.

    ASSIGN v-lastline = TRIM(v-thisline).
    DO WHILE TRUE:
      ASSIGN v-lastline = SUBSTR(v-lastline,v-start,-1, "character":u)
             v-index    = INDEX(v-lastline,CHR(10), v-start + 1).

      IF v-index = 0 THEN
        ASSIGN v-leave = TRUE
               v-index = LENGTH(v-lastline, "character":u).
      ELSE
        ASSIGN v-leave = FALSE.

      CASE pdfAlignment:
        WHEN "left" THEN
          RUN pdf_text_at(pdfStream,SUBSTR(v-lastline,v-start,v-index, "character":u),pdfFromColumn).
        WHEN "right" THEN
          RUN pdf_text_to(pdfStream,SUBSTR(v-lastline,v-start,v-index, "character":u),pdfToColumn).
      END CASE.

      IF v-leave THEN LEAVE.

      ELSE DO:
        v-lastline = SUBSTR(v-lastline,v-index + 1,-1, "character":u) .
        RUN pdf_skip(pdfStream).
      END.

    END. /* while true */
  END. /* ThisLine <> "" */

  ASSIGN pdfMaxY = PDF_TextY(pdfStream).

END PROCEDURE. /* pdf_wrap_text */

FUNCTION pdf_TotalPages RETURN CHARACTER
   (INPUT pdfStream AS CHARACTER).

  RETURN "@@TOTALPages-" + pdfStream.

END FUNCTION. /* pdf_TotalPages */

PROCEDURE pdf_link:
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromX    AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfFromY    AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfWidth    AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfHeight   AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfLink     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfRed      AS DECIMAL   NO-UNDO.
  DEFINE INPUT PARAMETER pdfGreen    AS DECIMAL   NO-UNDO.
  DEFINE INPUT PARAMETER pdfBlue     AS DECIMAL   NO-UNDO.
  DEFINE INPUT PARAMETER pdfBorder   AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfStyle    AS CHARACTER NO-UNDO.

  pdfRed   = IF pdfRed < 0 THEN 0
             ELSE IF pdfRed > 1 THEN 1
             ELSE pdfRed.
  pdfGreen = IF pdfGreen < 0 THEN 0
             ELSE IF pdfGreen > 1 THEN 1
             ELSE pdfGreen.
  pdfBlue  = IF pdfBlue < 0 THEN 0
             ELSE IF pdfBlue > 1 THEN 1
             ELSE pdfBlue.

  IF LOOKUP(pdfStyle,"N,I,O,P") = 0 THEN
    pdfStyle = "N".

  CREATE TT_pdf_link.
  ASSIGN TT_pdf_link.obj_stream   = pdfStream
         TT_pdf_link.link_name    = pdfLink.
         TT_pdf_link.link_page    = pdf_Page(pdfStream).
  ASSIGN TT_pdf_link.link_x       = pdfFromX
         TT_pdf_link.link_y       = pdfFromY
         TT_pdf_link.link_h       = pdfHeight
         TT_pdf_link.link_w       = pdfWidth
         TT_pdf_link.link_color   = dec2string(pdfRed) + " "
                                  + dec2string(pdfGreen) + " "
                                  + dec2string(pdfBlue)
         TT_pdf_link.link_border  = IF pdfBorder < 0 THEN 0 ELSE pdfBorder
         TT_pdf_link.link_style   = pdfStyle.

END. /* pdf_link */

PROCEDURE pdf_new_page :
  DEFINE INPUT PARAMETER pdfStream   AS CHARACTER NO-UNDO.

  DEFINE BUFFER B_TT_pdf_Stream FOR TT_pdf_Stream.

  FIND FIRST B_TT_pdf_stream
       WHERE B_TT_pdf_stream.obj_stream = pdfStream NO-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_Stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_new_page","Cannot find Stream!").
    RETURN .
  END.

  /* If the Tool Suite is being used then re-run the appropriate tools.
     Only runs the Tools for pages greater than one since the pdf_tool_create
     should handle the first page */
  IF pdf_Page(pdfStream) >= 1 AND VALID-HANDLE(h_PDF-Tool) THEN DO:
    FOR EACH TT_pdf_tool NO-LOCK:
      IF TT_pdf_tool.tool_type = "TABLE" THEN
        RUN DisplayTableOutline IN h_PDF-tool
            (pdfStream, TT_pdf_tool.tool_name).
    END.
  END.


  IF B_TT_pdf_stream.obj_Footer <> "" THEN DO:
    IF pdf_Page(pdfStream) >= 1 THEN DO:
      pdf_ForFooter = TRUE.

      RUN pdf_set_TextY(pdfStream,pdf_BottomMargin(pdfStream)).

      CREATE TT_pdf_content.
      ASSIGN TT_pdf_content.obj_stream  = pdfStream.
             TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
      ASSIGN TT_pdf_content.obj_type    = "TEXT".
             TT_pdf_content.obj_content = "1 0 0 1 " + STRING(pdf_LeftMargin(pdfStream))
                                        + " " + STRING(pdf_BottomMargin(pdfStream)) + " Tm".
      ASSIGN TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
             TT_pdf_content.obj_page    = pdf_Page(pdfStream).

      RUN VALUE(B_TT_pdf_stream.obj_Footer) IN B_TT_pdf_stream.obj_CallProc NO-ERROR.

      RUN pdf_set_TextY(pdfStream,pdf_PageHeight(pdfStream)).

      pdf_ForFooter = FALSE.
    END.
  END.

  RUN pdf_set_Page(pdfStream,pdf_Page(pdfStream) + 1).

  CREATE TT_pdf_page.
  TT_pdf_page.obj_stream  = pdfStream.
  TT_pdf_page.page_nbr    = pdf_Page(pdfStream).
  TT_pdf_page.page_rotate = pdf_PageRotate(pdfStream).

  /* Note: the placement of the following commands is important due to the
      setting of the X and Y attributes.  DO NOT CHANGE unless tested
      thoroughly */
  RUN pdf_set_LeftMargin(pdfStream, pdf_LeftMargin(pdfStream)).
  RUN pdf_set_TopMargin(pdfStream, pdf_TopMargin(pdfStream)).
  RUN pdf_set_BottomMargin(pdfStream, pdf_BottomMargin(pdfStream)).

  CASE pdf_PageRotate(pdfStream):
    WHEN 0 OR WHEN 180 THEN DO:
      RUN pdf_set_TextX(pdfStream, pdf_LeftMargin(pdfStream)).
      RUN pdf_set_Angle(pdfStream,0).
      RUN pdf_set_TextY(pdfStream, pdf_PageHeight(pdfStream) - pdf_TopMargin(pdfStream) ).
    END.

    WHEN 90 OR WHEN 270 THEN DO:
      RUN pdf_set_TextX(pdfStream, pdf_LeftMargin(pdfStream)).
      RUN pdf_set_Angle(pdfStream,0).
      RUN pdf_set_TextY(pdfStream, pdf_PageWidth(pdfStream) - pdf_TopMargin(pdfStream) ).
    END.

  END CASE.

  RUN pdf_set_GraphicX(pdfStream, 0).
  RUN pdf_set_GraphicY(pdfStream, 0).

  IF B_TT_pdf_stream.obj_Header <> "" THEN DO:
    RUN VALUE(B_TT_pdf_stream.obj_Header) IN B_TT_pdf_Stream.obj_CallProc NO-ERROR.
  END.

  /* If the Tool Suite is being used then re-run the appropriate tools.
     Only runs the Tools for pages greater than one since the pdf_tool_create
     should handle the first page */
  IF pdf_Page(pdfStream) > 1 AND VALID-HANDLE(h_PDF-Tool) THEN DO:
    FOR EACH TT_pdf_tool NO-LOCK:
      IF TT_pdf_tool.tool_type = "TABLE" THEN
        RUN DisplayTableHeader IN h_PDF-tool
            (pdfStream, TT_pdf_tool.tool_name).
    END.
  END.
END. /* pdf_new_page */

PROCEDURE pdf_close :
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Content AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_Encrypt AS LOGICAL NO-UNDO INIT FALSE.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_close","Cannot find Stream!").
    RETURN .
  END.

  DEFINE VARIABLE L_EncryptStatus   AS CHARACTER NO-UNDO.

  FIND FIRST TT_pdf_stream WHERE TT_pdf_stream.obj_stream = pdfStream
       NO-ERROR.

  /* If the Tool Suite is being used then re-run the appropriate tools.
     Only runs the Tools for pages greater than one since the pdf_tool_create
     should handle the first page */
  IF pdf_Page(pdfStream) >= 1 AND VALID-HANDLE(h_PDF-Tool) THEN DO:
    FOR EACH TT_pdf_tool NO-LOCK:
      IF TT_pdf_tool.tool_type = "TABLE" THEN
        RUN DisplayTableOutline IN h_PDF-tool
            (pdfStream, TT_pdf_tool.tool_name).
    END.
  END.

  /* Ensure that Footer is run for the Last Page */
  IF TT_pdf_stream.obj_Footer <> "" THEN DO:
    pdf_ForFooter = TRUE.

    RUN pdf_set_TextY(pdfStream,pdf_BottomMargin(pdfStream)).

    CREATE TT_pdf_content.
    ASSIGN TT_pdf_content.obj_stream  = pdfStream.
           TT_pdf_content.obj_seq     = pdf_inc_ContentSequence().
    ASSIGN TT_pdf_content.obj_type    = "TEXT".
           TT_pdf_content.obj_content = "1 0 0 1 " + STRING(pdf_LeftMargin(pdfStream))
                                      + " " + STRING(pdf_BottomMargin(pdfStream)) + " Tm".
    ASSIGN TT_pdf_content.obj_line    = pdf_TextY(pdfStream).
           TT_pdf_content.obj_page    = pdf_Page(pdfStream).

    RUN VALUE(TT_pdf_Stream.obj_Footer) IN TT_pdf_stream.obj_CallProc NO-ERROR.

    RUN pdf_set_TextY(pdfStream,pdf_PageHeight(pdfStream)).

    pdf_ForFooter = FALSE.
  END. /* Footer */

  /* Determine Unique Stream ID -- this string is used to output temporary files
     that are used for building content or generating Encryption keys etc */
  TT_pdf_stream.obj_UniqueID = ENCODE(STRING(TODAY) + STRING(TIME) + TT_pdf_stream.obj_file).


  /* Now Build PDF File */
  OUTPUT STREAM S_pdf_inc TO VALUE( TT_pdf_stream.obj_file ) UNBUFFERED NO-CONVERT.

    IF pdf_get_parameter(pdfStream,"Encrypt") = "TRUE" THEN DO:

      RUN pdfencrypt.p PERSISTENT SET h_PDF-Encrypt.

      RUN BuildDocID IN h_PDF-encrypt (INPUT  TT_pdf_stream.obj_UniqueID,
                                       OUTPUT TT_pdf_stream.obj_id).

      ASSIGN TT_pdf_stream.obj_master = pdf_get_parameter(pdfStream,"MasterPassword")
             TT_pdf_stream.obj_user   = pdf_get_parameter(pdfStream,"UserPassword").

      RUN DetermineOwnerKey IN h_PDF-Encrypt (INPUT TT_pdf_stream.obj_UniqueID,
                                              INPUT TT_pdf_stream.obj_user,
                                              INPUT-OUTPUT TT_pdf_stream.obj_master) .

      RUN DetermineUserKey  IN h_PDF-Encrypt (INPUT TT_pdf_stream.obj_UniqueID,
                                              INPUT TT_pdf_stream.obj_ID,
                                              INPUT INT(pdf_get_parameter(pdfStream,"EncryptKey")),
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowPrint") = "FALSE" THEN 0 ELSE 1,
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowCopy") = "FALSE" THEN 0 ELSE 1,
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowModify") = "FALSE" THEN 0 ELSE 1,
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowAnnots") = "FALSE" THEN 0 ELSE 1,
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowForms") = "FALSE" THEN 0 ELSE 1,
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowExtract") = "FALSE" THEN 0 ELSE 1,
                                              INPUT IF pdf_get_parameter(pdfStream,"AllowAssembly") = "FALSE" THEN 0 ELSE 1,
                                              OUTPUT pdf-EncryptKeyMemPtr,
                                              INPUT-OUTPUT TT_pdf_stream.obj_user,
                                              OUTPUT TT_pdf_stream.obj_P).

      L_Encrypt = TRUE.
    END. /* Encryption has been set */


    /* Output PDF Header Requirements */
    RUN pdf_Header (INPUT TT_pdf_stream.obj_stream, l_Encrypt).

    /* Load 12 Base Fonts - exclude wingdings etc */
    RUN pdf_Encoding (pdfStream).

    RUN pdf_Fonts (pdfStream).

    IF CAN-FIND (FIRST TT_pdf_error
                 WHERE TT_pdf_error.obj_stream = pdfStream NO-LOCK) THEN DO:
      FOR EACH TT_pdf_error.
        DISPLAY TT_pdf_error.
      END.
    END.
    ELSE DO:
      /* Load Embedded Fonts */
      IF CAN-FIND( FIRST TT_pdf_font WHERE TT_pdf_font.obj_stream = pdfStream
                                       AND TT_pdf_font.font_file <> "PDFBASE14" NO-LOCK)
      THEN RUN pdf_Load_Fonts (pdfStream, L_Encrypt).

      /* Load Embedded Images */
      IF CAN-FIND( FIRST TT_pdf_image NO-LOCK)
      THEN RUN pdf_Load_Images (pdfStream,L_Encrypt).

      /* Load Annotations */
      IF CAN-FIND( FIRST TT_pdf_link NO-LOCK)
      THEN RUN pdf_Load_Links (pdfStream, L_Encrypt).

      IF pdf_get_parameter(pdfStream,"Encrypt") = "TRUE" THEN DO:
        ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "Encrypt", 0).
        TT_pdf_stream.obj_encryptdict = pdf_inc_ObjectSequence.

        RUN pdf_EncryptDict(pdfStream, pdf_inc_ObjectSequence).
      END.

      RUN pdf_Resources (pdfStream).
      RUN pdf_Content (pdfStream, L_Encrypt).

      RUN pdf_Xref (pdfStream, L_Encrypt).

      IF TT_pdf_stream.obj_Last <> "" THEN
        RUN VALUE(TT_pdf_Stream.obj_Last) IN TT_pdf_stream.obj_CallProc NO-ERROR.

    END.

  OUTPUT STREAM S_pdf_inc CLOSE.

  IF TT_pdf_stream.obj_encrypt THEN DO:
    RUN PDFpsp.w
        (INPUT TT_pdf_stream.obj_file,
         INPUT TT_pdf_stream.obj_master,
         INPUT TT_pdf_stream.obj_user,
         INPUT TT_pdf_stream.obj_access,
         INPUT TT_pdf_stream.obj_key,
         INPUT TT_pdf_stream.obj_mode,
         INPUT TT_pdf_stream.obj_silent,
         OUTPUT l_EncryptStatus).
  END.

  /* Ensure we clean up some memory */
  SET-SIZE(mContent) = 0.
  SET-SIZE(mHolder) = 0.

  IF VALID-HANDLE(h_PDF-encrypt) THEN
    DELETE PROCEDURE h_PDF-encrypt.

  RUN pdf_reset_stream(pdfStream).

END PROCEDURE. /* pdf_Close */

FUNCTION pdf_PageFooter RETURN LOGICAL (INPUT pdfStream     AS CHARACTER,
                                        INPUT pdfProcHandle AS HANDLE,
                                        INPUT pdfFooterProc AS CHARACTER):

  DEFINE BUFFER B_TT_pdf_stream FOR TT_pdf_Stream.

  FIND FIRST B_TT_pdf_stream
       WHERE B_TT_pdf_stream.obj_stream = pdfStream EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_PageFooter","Cannot find Stream = " + pdfStream).
    RETURN FALSE.
  END.

  ASSIGN B_TT_pdf_stream.obj_CallProc = pdfProCHandle
         B_TT_pdf_stream.obj_Footer   = pdfFooterProc.

  RETURN TRUE.
END FUNCTION. /* pdf_PageFooter */

FUNCTION pdf_PageHeader RETURN LOGICAL (INPUT pdfStream     AS CHARACTER,
                                        INPUT pdfProcHandle AS HANDLE,
                                        INPUT pdfHeaderProc AS CHARACTER):

  DEFINE BUFFER B_TT_pdf_stream FOR TT_pdf_Stream.

  FIND FIRST B_TT_pdf_stream
       WHERE B_TT_pdf_stream.obj_stream = pdfStream EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_PageHeader","Cannot find Stream = " + pdfStream).
    RETURN FALSE.
  END.

  ASSIGN B_TT_pdf_stream.obj_CallProc = pdfProCHandle
         B_TT_pdf_stream.obj_Header   = pdfHeaderProc.

  RETURN TRUE.
END FUNCTION. /* pdf_PageHeader */

FUNCTION pdf_LastProcedure RETURN LOGICAL (INPUT pdfStream     AS CHARACTER,
                                           INPUT pdfProcHandle AS HANDLE,
                                           INPUT pdfLastProc   AS CHARACTER):

  DEFINE BUFFER B_TT_pdf_stream FOR TT_pdf_Stream.

  FIND FIRST B_TT_pdf_stream
       WHERE B_TT_pdf_stream.obj_stream = pdfStream EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_LastProcedure","Cannot find Stream = " + pdfStream).
    RETURN FALSE.
  END.

  ASSIGN B_TT_pdf_stream.obj_CallProc = pdfProCHandle
         B_TT_pdf_stream.obj_Last     = pdfLastProc.

  RETURN TRUE.

END FUNCTION. /* pdf_LastProcedure */

PROCEDURE pdf_set_parameter:
  DEFINE INPUT PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfParameter AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfValue     AS CHARACTER NO-UNDO.

  DEFINE BUFFER B_TT_pdf_param FOR TT_pdf_param.
  DEFINE BUFFER B_TT_pdf_Stream FOR TT_pdf_Stream.

  DEFINE VARIABLE L_Integer       AS INTEGER NO-UNDO.

  DEFINE VARIABLE L_Error         AS LOGICAL INIT FALSE.
  DEFINE VARIABLE L_ErrorMsg      AS CHARACTER NO-UNDO.
  DEFINE VARIABLE L_Valid_params  AS CHARACTER NO-UNDO.


  L_Valid_params = "Compress,Encrypt,UserPassword,MasterPassword,EncryptKey,"
                 + "AllowPrint,AllowCopy,AllowModify,AllowAnnots,AllowForms,"
                 + "AllowExtract,AllowAssembly,"
                 + "LineSpacer".
  FIND FIRST B_TT_pdf_stream
       WHERE B_TT_pdf_stream.obj_stream = pdfStream NO-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_parameter","Cannot find Stream!").
    RETURN.
  END.

  IF LOOKUP(pdfParameter,L_valid_params) = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_parameter","Invalid parameter (" + pdfParameter + ") trying to be set!").
    RETURN.
  END.

  CASE pdfParameter:
    WHEN "Compress" OR WHEN "Encrypt" OR WHEN "AllowPrint"
      OR WHEN "AllowCopy" OR WHEN "AllowModify" OR WHEN "AllowAnnots"
      OR WHEN "AllowForms" OR WHEN "AllowExtract" OR WHEN "AllowAssembly" THEN DO:
        IF pdfValue <> "TRUE" AND pdfValue <> "FALSE" AND pdfValue <> "" THEN
          ASSIGN L_Error    = TRUE
                 L_ErrorMsg = "Only TRUE, FALSE or blank allowed for '" + pdfParameter + "' Parameter!".

      /* Default the Encryption Key to 40 Bits - Rev 2 */
      IF pdfParameter = "Encrypt" THEN
        RUN pdf_set_parameter(pdfStream,"EncryptKey","40").
    END.

    WHEN "EncryptKey" THEN DO:
      /* Currently only allow for 40-Bit Encryption */
      IF pdfValue <> "40" /* AND pdfValue <> "128" */ THEN
        ASSIGN L_Error    = TRUE
               L_ErrorMsg = "Only a value of 40 is allowed for '" + pdfParameter + "' Parameter!".
    END.

    WHEN "UserPassword" OR WHEN "MasterPassword" THEN DO:
      IF LENGTH(pdfValue, "character":u) > 32 THEN
        ASSIGN L_Error    = TRUE
               L_ErrorMsg = "Password string cannot be greater than 32 characters for '" + pdfParameter + "' Parameter!".
    END.

    WHEN "LineSpacer" THEN DO:
      L_Integer = INT(pdfValue) NO-ERROR.
      IF ERROR-STATUS:ERROR THEN
        ASSIGN L_Error    = TRUE
               L_ErrorMsg = "'LineSpacer' Parameter must be an integer value!".
    END.
  END CASE.

  IF L_Error THEN DO:
    RUN pdf_error(pdfStream,"pdf_set_parameter",L_Error).
    RETURN.
  END. /* L_error */

  FIND B_TT_pdf_param WHERE B_TT_pdf_param.obj_stream    = pdfStream
                        AND B_TT_pdf_param.obj_parameter = pdfParameter
                        EXCLUSIVE NO-ERROR.
  IF NOT AVAIL B_TT_pdf_param THEN DO:
    CREATE B_TT_pdf_param.
    ASSIGN B_TT_pdf_param.obj_stream    = pdfStream
           B_TT_pdf_param.obj_parameter = pdfParameter
           B_TT_pdf_param.obj_value     = pdfValue.
  END.
  ELSE
    ASSIGN B_TT_pdf_param.obj_value = pdfValue.

END. /* pdf_set_parameter */

/* igc - added Novermber 17, 2003 */
PROCEDURE pdf_Encrypt:
  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfMaster  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfUser    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfAccess  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfKey     AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfMode    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfSilent  AS LOGICAL   NO-UNDO.

  DEFINE BUFFER B_TT_pdf_Stream FOR TT_pdf_Stream.

  DEFINE VARIABLE l_Loop        AS INTEGER NO-UNDO.

  FIND FIRST B_TT_pdf_stream
       WHERE B_TT_pdf_stream.obj_stream = pdfStream EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL B_TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_Encrypt","Cannot find Stream!").
    RETURN.
  END.

  /* if either of the passwords have been supplied ensure that it doesn't match
     the other -- they must be different */
  IF pdfMaster <> "" OR pdfUser <> "" THEN DO:
    IF ENCODE(pdfMaster) = ENCODE(pdfUser) THEN DO:
      RUN pdf_error(pdfStream,"pdf_Encrypt","Master and User Passwords must be different!").
      RETURN .
    END.
  END.

  /* Ensure that the correct Access permissions have been specified -- if not
     error out and do not perform encryption */
  DO l_Loop = 1 TO NUM-ENTRIES(pdfAccess,";"):
    IF LOOKUP(ENTRY(l_Loop,pdfAccess,";"),
              "noprint,nohiresprint,nomodify,nocopy,noannots,noforms,noaccessible,noassemble")
       = 0 THEN DO:
      RUN pdf_error(pdfStream,"pdf_Encrypt","Invalid Access Type passed - " + ENTRY(l_Loop,pdfAccess)).
      RETURN.
    END.
  END.

  IF LOOKUP(STRING(pdfKey),"40,128") = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_Encrypt","Invalid Key Length passed - only 40 or 128 allowed!").
    RETURN .
  END.

  IF LOOKUP(pdfMode,"COM,SHARED,OS") = 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_Encrypt","Invalid MODE passed - only COM, SHARED or OS allowed!").
    RETURN.
  END.

  ASSIGN B_TT_pdf_stream.obj_encrypt = TRUE
         B_TT_pdf_stream.obj_master  = pdfMaster
         B_TT_pdf_stream.obj_user    = pdfUser
         B_TT_pdf_stream.obj_access  = pdfAccess
         B_TT_pdf_stream.obj_key     = pdfKey
         B_TT_pdf_stream.obj_mode    = pdfMode
         B_TT_pdf_stream.obj_silent  = pdfSilent.

END. /* pdf_Encrypt */

/* igc - added November 25, 2003 */
PROCEDURE pdf_Bookmark:
  DEFINE INPUT  PARAMETER pdfStream    AS CHARACTER NO-UNDO.
  DEFINE INPUT  PARAMETER pdfTitle     AS CHARACTER NO-UNDO.
  DEFINE INPUT  PARAMETER pdfParent    AS INTEGER   NO-UNDO.
  DEFINE INPUT  PARAMETER pdfExpand    AS LOGICAL   NO-UNDO.
  DEFINE OUTPUT PARAMETER pdfBookmark  AS INTEGER   NO-UNDO.

  DEFINE BUFFER B_TT_pdf_bookmark FOR TT_pdf_bookmark.
  DEFINE BUFFER BB_TT_pdf_bookmark FOR TT_pdf_bookmark.

  FIND FIRST TT_pdf_stream
       WHERE TT_pdf_stream.obj_stream = pdfStream EXCLUSIVE-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_stream THEN DO:
    RUN pdf_error(pdfStream,"pdf_Bookmark","Cannot find Stream!").
    RETURN.
  END.

  IF pdfParent <> 0
  AND NOT CAN-FIND(FIRST TT_pdf_bookmark
                   WHERE TT_pdf_bookmark.obj_stream  = pdfStream
                     AND TT_pdf_bookmark.book_parent = 0 NO-LOCK) THEN DO:

    RUN pdf_error(pdfStream,"pdf_Bookmark","Trying to add a Bookmark without having a Main Level defined!").
    RETURN.
  END.

  /* Find last Bookmark --- used to increment the unique Bookmark Number */
  FIND LAST TT_pdf_bookmark WHERE TT_pdf_bookmark.obj_stream = pdfStream NO-LOCK NO-ERROR.

  CREATE B_TT_pdf_bookmark.
  ASSIGN B_TT_pdf_bookmark.obj_stream  = pdfStream
         B_TT_pdf_bookmark.book_nbr    = IF NOT AVAIL TT_pdf_bookmark THEN
                                           1
                                         ELSE
                                           TT_pdf_bookmark.book_nbr + 1
         B_TT_pdf_bookmark.book_title  = pdfTitle
         B_TT_pdf_bookmark.book_parent = pdfParent
         B_TT_pdf_bookmark.book_expand = pdfExpand.
         B_TT_pdf_bookmark.book_page   = pdf_page(pdfstream).
         B_TT_pdf_bookmark.book_Y      = pdf_TextY(pdfstream)
                                       + pdf_PointSize(pdfStream).

  pdfBookmark = B_TT_pdf_bookmark.book_nbr.

  /* Now that we've added the Bookmark, find the parent and increment the
     parents child counter.  But only when the Parent is not 0 (Main Level) */
  IF pdfParent <> 0 THEN DO:
    FIND FIRST BB_TT_pdf_bookmark
         WHERE BB_TT_pdf_bookmark.obj_stream = pdfStream
           AND BB_TT_pdf_bookmark.book_nbr   = pdfParent NO-ERROR.
    IF AVAIL BB_TT_pdf_bookmark THEN DO:
      IF BB_TT_pdf_bookmark.book_expand THEN
        BB_TT_pdf_bookmark.book_child = BB_TT_pdf_bookmark.book_child + 1.
      ELSE
        BB_TT_pdf_bookmark.book_child = BB_TT_pdf_bookmark.book_child - 1.
    END.
  END. /* Increment Parent Children */

END. /* pdf_Bookmark */

PROCEDURE pdf_process_bookmarks:
  DEFINE INPUT PARAMETER pStream     AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pParentNbr  AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pParentObj  AS INTEGER NO-UNDO.
  DEFINE INPUT PARAMETER pEncrypt    AS LOGICAL NO-UNDO.

  DEFINE BUFFER B_TT_pdf_Bookmark    FOR TT_pdf_bookmark.
  DEFINE BUFFER LB_TT_pdf_Bookmark   FOR TT_pdf_bookmark.
  DEFINE BUFFER LBB_TT_pdf_Bookmark  FOR TT_pdf_bookmark.
  DEFINE BUFFER LBBB_TT_pdf_Bookmark FOR TT_pdf_bookmark.

  DEFINE VARIABLE L_Object           AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Count            AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Last             AS INTEGER NO-UNDO.

  DEFINE VARIABLE m_EncryptKey       AS MEMPTR NO-UNDO.

  FIND FIRST LBB_TT_pdf_Bookmark
       WHERE LBB_TT_pdf_Bookmark.obj_stream = pStream
         AND LBB_TT_pdf_bookmark.book_nbr   = pParentNbr NO-LOCK NO-ERROR.

  FOR EACH LB_TT_pdf_bookmark
     WHERE LB_TT_pdf_bookmark.obj_stream  = pStream
       AND LB_TT_pdf_bookmark.book_parent = pParentNbr
       BREAK BY LB_TT_pdf_bookmark.book_nbr:

    ObjectSequence(pStream, pdf_inc_ObjectSequence + 1, "BookMark", 0).
    LB_TT_pdf_bookmark.book_obj = pdf_inc_ObjectSequence.

    /* Find the associated Page Dictionary Object Number */
    FIND FIRST TT_pdf_object
         WHERE TT_pdf_object.obj_stream = pStream
           AND TT_pdf_object.obj_page   = LB_TT_pdf_bookmark.book_page
           NO-ERROR.
    IF NOT AVAIL TT_pdf_object THEN NEXT.
    L_object = TT_pdf_object.obj_nbr.

    PUT STREAM S_pdf_inc UNFORMATTED
        LB_TT_pdf_Bookmark.book_obj " 0 obj" CHR(13)
        "<<" CHR(13).

    IF pEncrypt THEN DO:
      PUT STREAM S_pdf_inc UNFORMATTED
        "/Title <".

      SET-SIZE(m_EncryptKey) = 10.
      RUN GetEncryptKey IN h_PDF-Encrypt
          (INPUT  TT_pdf_stream.obj_UniqueID,
           INPUT  pdf_inc_ObjectSequence,
           INPUT  0,
           INPUT  pdf-EncryptKeyMemPtr,
           OUTPUT m_EncryptKey).

      RUN ConvChar2MemPtr(LB_TT_pdf_bookmark.book_title).

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  mContent,
                          OUTPUT mContent).

      RUN OutputMemPtrAsHex (mContent).
      SET-SIZE(m_EncryptKey) = 0.

      PUT STREAM S_pdf_inc UNFORMATTED
        ">" CHR(13).
    END.

    ELSE DO:
      RUN pdf_replace_text(INPUT-OUTPUT LB_TT_pdf_bookmark.book_title).

      PUT STREAM S_pdf_inc UNFORMATTED
        "/Title ("
        LB_TT_pdf_bookmark.book_title ")" CHR(13).
    END.

    PUT STREAM S_pdf_inc UNFORMATTED
        "/Parent " LBB_TT_pdf_Bookmark.book_obj " 0 R" CHR(13).

   /* Determine if a Previous Bookmark exists on the same level */
    FIND LAST LBBB_TT_pdf_Bookmark
         WHERE LBBB_TT_pdf_Bookmark.obj_stream  = pStream
           AND LBBB_TT_pdf_Bookmark.book_parent = pParentNbr
           AND LBBB_TT_pdf_Bookmark.book_nbr    < LB_TT_pdf_Bookmark.book_nbr
           NO-LOCK NO-ERROR.
    IF AVAIL LBBB_TT_pdf_bookmark THEN
      PUT STREAM S_pdf_inc UNFORMATTED
          "/Prev " LBBB_TT_pdf_Bookmark.book_obj " 0 R" CHR(13).

   /* Determine if a Following (Next) Bookmark exists on the same level */
    FIND FIRST LBBB_TT_pdf_Bookmark
         WHERE LBBB_TT_pdf_Bookmark.obj_stream  = pStream
           AND LBBB_TT_pdf_Bookmark.book_parent = pParentNbr
           AND LBBB_TT_pdf_Bookmark.book_nbr    > LB_TT_pdf_Bookmark.book_nbr
           NO-LOCK NO-ERROR.
    IF AVAIL LBBB_TT_pdf_bookmark THEN
      PUT STREAM S_pdf_inc UNFORMATTED
          "/Next " pdf_OutlinesDict + LBBB_TT_pdf_Bookmark.book_nbr " 0 R" CHR(13).

    /* If Children are associated with this Bookmark then add some processing */
    IF LB_TT_pdf_bookmark.book_child <> 0 THEN DO:
      FIND LAST B_TT_pdf_Bookmark
          WHERE B_TT_pdf_Bookmark.obj_stream  = pStream
            AND B_TT_pdf_Bookmark.book_parent = LB_TT_pdf_bookmark.book_nbr
            NO-LOCK NO-ERROR.
      IF AVAIL B_TT_pdf_bookmark THEN DO:
        L_Last = pdf_OutlinesDict + B_TT_pdf_Bookmark.book_nbr.
      END.
      ELSE
        L_Last = pdf_OutlinesLast.

      PUT STREAM S_pdf_inc UNFORMATTED
          "~/First " LB_TT_pdf_Bookmark.book_obj + 1 " 0 R" CHR(13)
          "~/Last "  L_Last " 0 R" CHR(13)
          "~/Count " LB_TT_pdf_Bookmark.book_child CHR(13).

    END. /* Child <> 0 */

    PUT STREAM S_pdf_inc UNFORMATTED
        "/Dest [ " l_Object " 0 R /XYZ 0 " LB_TT_pdf_bookmark.book_Y " 0 ]"
        CHR(13)
         ">>" CHR(13)
        "endobj" CHR(13).

    IF LB_TT_pdf_Bookmark.book_child <> 0 THEN
      RUN pdf_process_bookmarks(pStream, LB_TT_pdf_bookmark.book_nbr, LBB_TT_pdf_bookmark.book_obj, pEncrypt).
  END.

END. /* pdf_process_bookmarks */

PROCEDURE pdf_load_bookmarks:

  DEFINE INPUT PARAMETER pdfStream  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Last    AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Object  AS INTEGER NO-UNDO.

  DEFINE VARIABLE L_Encrypt AS LOGICAL NO-UNDO INIT FALSE.

  DEFINE VARIABLE m_EncryptKey  AS MEMPTR NO-UNDO.

  DEFINE BUFFER B_TT_pdf_bookmark FOR TT_pdf_bookmark.

  IF pdf_get_parameter(pdfStream,"Encrypt") = "TRUE" THEN
    L_Encrypt = TRUE.

  FOR EACH TT_pdf_bookmark WHERE TT_pdf_bookmark.obj_stream = pdfStream
                             AND TT_pdf_bookmark.book_parent = 0
     BREAK BY TT_pdf_bookmark.book_nbr:

    IF FIRST(TT_pdf_bookmark.book_nbr) THEN DO:
      ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "BookMark", 0).
      TT_pdf_bookmark.book_obj = pdf_inc_ObjectSequence.
      pdf_OutlinesDict = pdf_inc_ObjectSequence.

      /* Determine the total number of Bookmarks in the document */
      FIND LAST B_TT_pdf_bookmark WHERE B_TT_pdf_bookmark.obj_stream = pdfStream NO-LOCK NO-ERROR.
      pdf_OutlinesLast = pdf_OutlinesDict + B_TT_pdf_bookmark.book_nbr.

      /* Output the Bookmark Dictionary */
      PUT STREAM S_pdf_inc UNFORMATTED
         pdf_OutlinesDict " 0 obj" CHR(13)
          "<< " CHR(13)
          "/Type /Outlines" CHR(13)
          "/First " pdf_OutlinesDict + 1 " 0 R" CHR(13)
          "/Last " pdf_OutlinesLast " 0 R" CHR(13)
          "/Count " B_TT_pdf_Bookmark.book_nbr CHR(13)
          ">>" CHR(13)
          "endobj" CHR(13).

    END. /* FIRST */

    ObjectSequence(pdfStream, pdf_inc_ObjectSequence + 1, "BookMark", 0).
    TT_pdf_bookmark.book_obj = pdf_inc_ObjectSequence.

    /* Find the associated Page Dictionary Object Number */
    FIND FIRST TT_pdf_object
         WHERE TT_pdf_object.obj_stream = pdfStream
           AND TT_pdf_object.obj_page   = TT_pdf_bookmark.book_page
           NO-ERROR.
    IF NOT AVAIL TT_pdf_object THEN NEXT.
    L_object = TT_pdf_object.obj_nbr.


    PUT STREAM S_pdf_inc UNFORMATTED
        TT_pdf_Bookmark.book_obj " 0 obj" CHR(13)
        "<<" CHR(13).

    IF l_Encrypt THEN DO:
      PUT STREAM S_pdf_inc UNFORMATTED
        "/Title <".

      SET-SIZE(m_EncryptKey) = 10.
      RUN GetEncryptKey IN h_PDF-Encrypt
          (INPUT  TT_pdf_stream.obj_UniqueID,
           INPUT  pdf_inc_ObjectSequence,
           INPUT  0,
           INPUT  pdf-EncryptKeyMemPtr,
           OUTPUT m_EncryptKey).

      RUN ConvChar2MemPtr(TT_pdf_bookmark.book_title).

      RUN EncryptContent IN h_PDF-Encrypt
                         (INPUT  m_EncryptKey,
                          INPUT  mContent,
                          OUTPUT mContent).

      RUN OutputMemPtrAsHex (mContent).

      PUT STREAM S_pdf_inc UNFORMATTED
        ">" CHR(13).

    END.

    ELSE DO:
      RUN pdf_replace_text(INPUT-OUTPUT TT_pdf_bookmark.book_title).

      PUT STREAM S_pdf_inc UNFORMATTED
          "/Title ("
          TT_pdf_bookmark.book_title
          ")" CHR(13).
    END.

    PUT STREAM S_pdf_inc UNFORMATTED
         "/Parent " pdf_OutlinesDict " 0 R" CHR(13).

   /* Determine if a Previous Bookmark exists on the same level */
    FIND LAST B_TT_pdf_Bookmark
         WHERE B_TT_pdf_Bookmark.obj_stream  = pdfStream
           AND B_TT_pdf_Bookmark.book_parent = 0
           AND B_TT_pdf_Bookmark.book_nbr    < TT_pdf_Bookmark.book_nbr
           NO-LOCK NO-ERROR.
    IF AVAIL B_TT_pdf_bookmark THEN
      PUT STREAM S_pdf_inc UNFORMATTED
         "/Prev " B_TT_pdf_Bookmark.book_obj " 0 R" CHR(13).

   /* Determine if a Following (Next) Bookmark exists on the same level */
    FIND FIRST B_TT_pdf_Bookmark
         WHERE B_TT_pdf_Bookmark.obj_stream  = pdfStream
           AND B_TT_pdf_Bookmark.book_parent = 0
           AND B_TT_pdf_Bookmark.book_nbr    > TT_pdf_Bookmark.book_nbr
           NO-LOCK NO-ERROR.
    IF AVAIL B_TT_pdf_bookmark THEN
      PUT STREAM S_pdf_inc UNFORMATTED
          "~/Next " pdf_OutlinesDict + B_TT_pdf_Bookmark.book_nbr " 0 R" CHR(13).

    /* If Children are associated with this Bookmark then add some processing */
    IF TT_pdf_bookmark.book_child <> 0 THEN DO:
      FIND LAST B_TT_pdf_Bookmark
          WHERE B_TT_pdf_Bookmark.obj_stream  = pdfStream
            AND B_TT_pdf_Bookmark.book_parent = TT_pdf_bookmark.book_nbr
            NO-LOCK NO-ERROR.
     IF AVAIL B_TT_pdf_bookmark THEN DO:
       L_Last = pdf_OutlinesDict + B_TT_pdf_Bookmark.book_nbr.
     END.
     ELSE
       L_Last = pdf_OutlinesLast.

     PUT STREAM S_pdf_inc UNFORMATTED
          "~/First " TT_pdf_Bookmark.book_obj + 1 " 0 R" CHR(13)
          "~/Last "  L_Last " 0 R" CHR(13)
          "~/Count " TT_pdf_Bookmark.book_child CHR(13).
    END.

    SET-SIZE(m_EncryptKey) = 0.

    PUT STREAM S_pdf_inc UNFORMATTED
        "/Dest [ " L_Object " 0 R /XYZ 0 " TT_pdf_bookmark.book_Y " 0 ]" CHR(13)
        ">>" CHR(13)
        "endobj" CHR(13).

    RUN pdf_process_bookmarks(pdfStream, TT_pdf_bookmark.book_nbr, TT_pdf_bookmark.book_obj, L_Encrypt).

  END. /* each TT_pdf_link */

  SET-SIZE(m_EncryptKey) = 0.
  SET-SIZE(mContent) = 0.

END. /* pdf_load_bookmarks */

/* ************************************************* */
/*                  PEKI's procedures                */
/* ************************************************* */
PROCEDURE pdf_merge_stream :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfNbrCopies          AS INT          NO-UNDO.

    /* Empty Temp-Table */
    FOR EACH TT-Merge-Pages:
        DELETE TT-Merge-Pages.
    END.

    /* No Merge done when error is found */
    IF CAN-FIND (FIRST TT_pdf_error
                 WHERE TT_pdf_error.obj_stream = pdfStreamFrom NO-LOCK)
    OR TRIM(pdfStreamFrom) = ""
    OR TRIM(pdfStreamTo  ) = ""
       THEN RETURN.

    IF pdfNbrCopies <= 0
       THEN ASSIGN pdfNbrCopies = 1.

    RUN pdf_merge_stream_content(INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ,
                                 INPUT pdfNbrCopies ).
    RUN pdf_merge_stream_page   (INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ,
                                 INPUT pdfNbrCopies ).
    RUN pdf_merge_stream_param  (INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ).
    RUN pdf_merge_stream_image  (INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ).
    RUN pdf_merge_stream_diff   (INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ).
    RUN pdf_merge_stream_link   (INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ,
                                 INPUT pdfNbrCopies ).
    RUN pdf_merge_stream_book   (INPUT pdfStreamFrom,
                                 INPUT pdfStreamTo  ,
                                 INPUT pdfNbrCopies ).
END PROCEDURE. /* pdf_merge_stream */

PROCEDURE pdf_merge_stream_content :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfNbrCopies          AS INT          NO-UNDO.


    DEF VAR vLastPage         AS INT             NO-UNDO.
    DEF VAR vCount            AS INT             NO-UNDO.

    DEFINE BUFFER B_TT_pdf_content FOR TT_pdf_content.
    DEFINE BUFFER B_TT_pdf_stream  FOR TT_pdf_stream.

    FIND B_TT_pdf_stream
         WHERE B_TT_pdf_stream.obj_stream = pdfStreamTo NO-LOCK NO-ERROR.
    IF NOT AVAIL B_TT_pdf_stream THEN DO:
      RUN pdf_error(pdfStreamTo,"pdf_merge_stream_content","Cannot find Stream = " + pdfStreamTo).
      RETURN.
    END.

    /* Run the Page Footer for the parent stream before merging the child 
       streams */
    IF B_TT_pdf_Stream.obj_footer <> "" THEN
      RUN VALUE(B_TT_pdf_Stream.obj_footer) IN B_TT_pdf_stream.obj_CallProc NO-ERROR.

    DO vCount = 1 TO pdfNbrCopies:

        /* Find Current Page */
        ASSIGN vLastPage = pdf_Page(pdfStreamTo).
        IF NOT ( vLastPage >= 1 )
           THEN ASSIGN vLastPage = 0.

        FOR EACH TT_pdf_content WHERE TT_pdf_content.obj_stream = pdfStreamFrom
                                  AND TT_pdf_content.obj_page > 0
                                NO-LOCK TRANSACTION
                                BREAK BY TT_pdf_content.obj_seq
                                      BY TT_pdf_content.obj_page:

            CREATE B_TT_pdf_content.
            BUFFER-COPY TT_pdf_Content   EXCEPT TT_pdf_Content.obj_stream
                                                TT_pdf_Content.obj_seq
                                                TT_pdf_Content.obj_page
                     TO B_TT_pdf_content.

            ASSIGN B_TT_pdf_Content.obj_stream = pdfStreamTo.
            ASSIGN B_TT_pdf_Content.obj_seq    = pdf_inc_ContentSequence().
            ASSIGN B_TT_pdf_Content.obj_page   = vLastPage + TT_pdf_Content.obj_page.

            /* Save the NEW page indications for other copies */
            IF NOT CAN-FIND (TT-Merge-Pages WHERE TT-Merge-Pages.PageFrom = TT_pdf_Content.obj_page
                                              AND TT-Merge-Pages.PageTo   = B_TT_pdf_Content.obj_page
                                              AND TT-Merge-Pages.MergeNr  = vCount)
            THEN DO:
                 CREATE TT-Merge-Pages.
                 ASSIGN TT-Merge-Pages.PageFrom = TT_pdf_Content.obj_page
                        TT-Merge-Pages.PageTo   = B_TT_pdf_Content.obj_page
                        TT-Merge-Pages.MergeNr  = vCount.
            END.


            /* Replace a piece of text, for example: 'Title' to 'Copy Title' */
            FOR EACH TT_pdf_ReplaceTxt WHERE TT_pdf_ReplaceTxt.obj_Stream = pdfStreamFrom
                                         AND TT_pdf_ReplaceTxt.MergeNr = vCount
                                       NO-LOCK:

                   IF INDEX(B_TT_pdf_content.obj_content,TT_pdf_ReplaceTxt.txt_from) <> 0
                      THEN ASSIGN B_TT_pdf_content.obj_content =
                                  REPLACE(B_TT_pdf_content.obj_content,
                                          TT_pdf_ReplaceTxt.txt_from  ,
                                          TT_pdf_ReplaceTxt.txt_to    ).
            END.

            IF LAST(TT_pdf_content.obj_seq)
               THEN ASSIGN vLastPage = B_TT_pdf_Content.obj_page.

            RELEASE B_TT_pdf_Content.

        END.

        /* Set the last page */
        RUN pdf_set_Page(pdfStreamTo,vLastPage).

    END. /* END pdfCount = 1 TO pdfNbrCopies: */

    /* Reset the pointer to current Stream, is disturbed by calling pdf_Page function */
    FIND FIRST TT_pdf_stream
         WHERE TT_pdf_stream.obj_stream = pdfStreamFrom NO-LOCK NO-ERROR.
    
END PROCEDURE. /* pdf_merge_stream_content */


PROCEDURE pdf_merge_stream_page:

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfNbrCopies          AS INT          NO-UNDO.

    DEFINE BUFFER B_TT_pdf_page FOR TT_pdf_page.

    DEF VAR vCount         AS INT             NO-UNDO.

    DO vCount = 1 TO pdfNbrCopies:

        FOR EACH TT-Merge-Pages WHERE TT-Merge-Pages.MergeNr = vCount
                                NO-LOCK
             ,FIRST TT_pdf_page WHERE TT_pdf_page.obj_stream = pdfStreamFrom
                                  AND TT_pdf_page.page_nbr   = TT-Merge-Pages.PageFrom
                                NO-LOCK :

                CREATE B_TT_pdf_page.
                BUFFER-COPY TT_pdf_page   EXCEPT TT_pdf_page.obj_stream
                                                 TT_pdf_page.page_nbr
                         TO B_TT_pdf_page ASSIGN B_TT_pdf_page.page_nbr   = TT-Merge-Pages.PageTo
                                                 B_TT_pdf_page.obj_stream = pdfStreamTo.

                RELEASE B_TT_pdf_page.

        END.
    END.

END. /* pdf_merge_stream_page */


PROCEDURE pdf_merge_stream_link :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfNbrCopies          AS INT          NO-UNDO.

    DEFINE BUFFER B_TT_pdf_link FOR TT_pdf_link.

    DEF VAR vCount         AS INT             NO-UNDO.

    DO vCount = 1 TO pdfNbrCopies:

        FOR EACH TT-Merge-Pages WHERE TT-Merge-Pages.MergeNr = vCount
                                NO-LOCK
           ,EACH TT_pdf_link    WHERE TT_pdf_link.obj_stream = pdfStreamFrom
                                  AND TT_pdf_link.link_page  = TT-Merge-Pages.PageFrom
                                NO-LOCK:
    
                CREATE B_TT_pdf_link.
                BUFFER-COPY TT_pdf_link   EXCEPT TT_pdf_link.obj_stream
                                                 TT_pdf_link.link_page
                         TO B_TT_pdf_link.

                ASSIGN B_TT_pdf_link.obj_stream = pdfStreamTo.
                ASSIGN B_TT_pdf_link.link_page  = TT-Merge-Pages.PageTo.

                RELEASE B_TT_pdf_link.
        END.
    END.

END PROCEDURE. /* pdf_merge_stream_link */


PROCEDURE pdf_merge_stream_book :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfNbrCopies          AS INT          NO-UNDO.

    DEFINE BUFFER B_TT_pdf_bookmark FOR TT_pdf_bookmark.

    DEF VAR vBookNr        AS INT             NO-UNDO.
    DEF VAR vCount         AS INT             NO-UNDO.

    DO vCount = 1 TO pdfNbrCopies:

        FOR EACH TT-Merge-Pages WHERE TT-Merge-Pages.MergeNr = vCount NO-LOCK :

            /* Set the Qty to add to TT_pdf_bookmark.book_nbr */
            FIND LAST TT_pdf_bookmark WHERE TT_pdf_bookmark.obj_stream  = pdfStreamTo NO-LOCK NO-ERROR.
            IF AVAILABLE TT_pdf_bookmark
               THEN ASSIGN vBookNr = TT_pdf_bookmark.book_nbr.

            FOR EACH TT_pdf_bookmark WHERE TT_pdf_bookmark.obj_stream = pdfStreamFrom
                                       AND TT_pdf_bookmark.book_page  = TT-Merge-Pages.PageFrom
                                     NO-LOCK
                                     BREAK BY TT_pdf_bookmark.book_nbr:

                CREATE B_TT_pdf_bookmark.
                BUFFER-COPY TT_pdf_bookmark EXCEPT TT_pdf_bookmark.obj_stream
                                                   TT_pdf_bookmark.book_nbr
                                                   TT_pdf_bookmark.book_obj
                                                   TT_pdf_bookmark.book_first
                                                   TT_pdf_bookmark.book_last
                         TO B_TT_pdf_bookmark.

                ASSIGN B_TT_pdf_bookmark.book_nbr    = TT_pdf_bookmark.book_nbr + vBookNr.
                ASSIGN B_TT_pdf_bookmark.obj_stream  = pdfStreamTo.
                ASSIGN B_TT_pdf_bookmark.book_page   = TT-Merge-Pages.PageTo.

                IF TT_pdf_bookmark.book_parent <> 0
                   THEN ASSIGN B_TT_pdf_bookmark.book_parent = TT_pdf_bookmark.book_parent + vBookNr.

                RELEASE B_TT_pdf_bookmark.
           END.
        END.
    END.

END PROCEDURE. /* pdf_merge_stream_book */


PROCEDURE pdf_merge_stream_param :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.

    DEFINE BUFFER B_TT_pdf_param FOR TT_pdf_param.

    FOR EACH TT_pdf_param WHERE TT_pdf_param.obj_stream     = pdfStreamFrom
                            AND TT_pdf_param.obj_parameter <> "Page"
                          NO-LOCK TRANSACTION:
    
        FIND FIRST B_TT_pdf_param WHERE B_TT_pdf_param.obj_stream    = pdfStreamTo
                                    AND B_TT_pdf_param.obj_parameter = TT_pdf_param.obj_parameter
                                  EXCLUSIVE-LOCK NO-WAIT NO-ERROR.
        IF NOT AVAILABLE B_TT_pdf_param
        THEN DO:
             CREATE B_TT_pdf_param.
             ASSIGN B_TT_pdf_param.obj_stream    = pdfStreamTo
                    B_TT_pdf_param.obj_parameter = TT_pdf_param.obj_parameter.
        END.

        ASSIGN B_TT_pdf_param.obj_valid = TT_pdf_param.obj_valid
               B_TT_pdf_param.obj_value = TT_pdf_param.obj_value.

        RELEASE B_TT_pdf_param.

    END.

END PROCEDURE.  /* pdf_merger_stream_param */


PROCEDURE pdf_merge_stream_image :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.

    DEFINE BUFFER B_TT_pdf_image FOR TT_pdf_image.

    FOR EACH TT_pdf_image WHERE TT_pdf_image.obj_stream = pdfStreamFrom
                          NO-LOCK TRANSACTION:
    
        IF CAN-FIND(FIRST B_TT_pdf_image WHERE B_TT_pdf_image.obj_stream = pdfStreamTo
                                           AND B_TT_pdf_image.image_name = TT_pdf_image.image_name
                                         NO-LOCK)
        THEN NEXT.

        BUFFER-COPY TT_pdf_image  EXCEPT TT_pdf_image.obj_stream
                 TO B_TT_pdf_image.

        ASSIGN B_TT_pdf_image.obj_stream = pdfStreamTo.

        RELEASE B_TT_pdf_image.

    END.

END PROCEDURE. /* pdf_merge_stream_image */


PROCEDURE pdf_merge_stream_diff :

    DEF INPUT        PARAM pdfStreamFrom         AS CHAR         NO-UNDO.
    DEF INPUT        PARAM pdfStreamTo           AS CHAR         NO-UNDO.

    DEFINE BUFFER B_TT_pdf_diff FOR TT_pdf_diff.

    FOR EACH TT_pdf_diff WHERE TT_pdf_diff.obj_stream = pdfStreamFrom
                         NO-LOCK TRANSACTION:
    
        IF CAN-FIND(FIRST B_TT_pdf_diff WHERE B_TT_pdf_diff.obj_stream = pdfStreamTo
                                          AND B_TT_pdf_diff.font_name  = TT_pdf_diff.font_name
                                        NO-LOCK)
        THEN NEXT.

        BUFFER-COPY TT_pdf_diff  EXCEPT TT_pdf_diff.obj_stream
                 TO B_TT_pdf_diff.

        ASSIGN B_TT_pdf_diff.obj_stream = pdfStreamTo.

        RELEASE B_TT_pdf_diff.

    END.

END PROCEDURE.  /* pdf_merge_stream_diff */


PROCEDURE pdf_ReplaceText:

   DEFINE INPUT PARAMETER pdfStream     AS CHARACTER NO-UNDO.
   DEFINE INPUT PARAMETER pdfMergeNbr   AS INTEGER   NO-UNDO.
   DEFINE INPUT PARAMETER pdfTextFrom   AS CHARACTER NO-UNDO.
   DEFINE INPUT PARAMETER pdfTextTo     AS CHARACTER NO-UNDO.

   CREATE TT_pdf_ReplaceTxt.
   ASSIGN TT_pdf_ReplaceTxt.obj_stream  = pdfStream
          TT_pdf_ReplaceTxt.mergenr     = pdfMergeNbr
          TT_pdf_ReplaceTxt.txt_from    = pdfTextFrom
          TT_pdf_ReplaceTxt.txt_to      = pdfTextTo.

END PROCEDURE. /* pdf_ReplaceText */

/* END PEKI's procedures ******************************* */

PROCEDURE pdf_EncryptDict:
  DEFINE INPUT PARAMETER pdfStream AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfObject AS INTEGER   NO-UNDO.

  FIND TT_pdf_stream WHERE TT_pdf_stream.obj_stream = pdfStream NO-ERROR.

  PUT STREAM S_pdf_inc UNFORMATTED
      pdfObject " 0 obj" {&pdfSKIP}
      "<<" {&pdfSKIP}
      "/Filter /Standard" {&pdfSKIP}
      "/V 1" {&pdfSKIP}
      "/R 2" {&pdfSKIP}
      /* "/Length 40" {&pdfSKIP} */
      "/O<" TT_pdf_stream.obj_master ">" {&pdfSKIP}
      "/U<" TT_pdf_stream.obj_user  ">" {&pdfSKIP}
      "/P " TT_pdf_stream.obj_P {&pdfSKIP}
      ">>" {&pdfSKIP}
      "endobj" {&pdfSKIP}.

END. /* pdf_EncryptDict */

/** For Testing Purposes Only
PROCEDURE PutMemPtr:
  DEFINE INPUT PARAMETER pMemPtr AS MEMPTR NO-UNDO.
  DEFINE INPUT PARAMETER pFile   AS CHARACTER NO-UNDO.

  DEFINE VARIABLE l_Loop  AS INTEGER NO-UNDO.

  OUTPUT TO VALUE(pFile) BINARY NO-MAP NO-CONVERT.
    EXPORT pMemPtr.
  OUTPUT CLOSE.
END. /* PutMemPtr */
**/

PROCEDURE OutputMemPtr:
  DEFINE INPUT PARAMETER pMemPtr AS MEMPTR NO-UNDO.

  DEFINE VARIABLE l_Loop  AS INTEGER NO-UNDO.

  EXPORT STREAM S_PDF_inc pMemPtr.
END. /* OutputMemPtr */

PROCEDURE OutputMemPtrAsHex:
  DEFINE INPUT PARAMETER pMemPtr AS MEMPTR NO-UNDO.

  DEFINE VARIABLE l_Loop  AS INTEGER NO-UNDO.

  DO l_Loop = 1 TO GET-SIZE(pMemPtr):
    IF GET-BYTE(pMemPtr,l_Loop) = -1  THEN
      PUT STREAM S_PDF_inc "00".
    ELSE
      PUT STREAM S_PDF_inc UNFORMATTED
          SUBSTR(Hex( ASC(CHR(GET-BYTE(pMemPtr,l_Loop)))),3,2, "character":u).
  END.
END. /* OutputMemPtr */

PROCEDURE AddCharContent :

  DEFINE INPUT PARAMETER pText  AS CHARACTER NO-UNDO.

  PUT UNFORMATTED ptext SKIP.

END. /* AddCharContent */

PROCEDURE ConvChar2MemPtr :

  DEFINE INPUT PARAMETER pText  AS CHARACTER NO-UNDO.

  DEFINE VARIABLE mWork     AS MEMPTR NO-UNDO.

  /* Set the Working Memptr and store the new text */
  SET-SIZE(mWork) = 0.
  SET-SIZE(mWork) = LENGTH(pText, "character":u).

  /* Now Add the Original plus the Working to the Content MemPtr */
  SET-SIZE(mContent) = 0.
  SET-SIZE(mContent) = GET-SIZE(mWork) + 1.
  PUT-STRING(mContent,1) = pText.

  SET-SIZE(mWork) = 0.

END. /* ConvChar2MemPtr */

PROCEDURE pdf_tool_add :

  DEFINE INPUT PARAMETER pdfStream      AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolName    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolType    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolData    AS HANDLE    NO-UNDO.

  IF NOT CAN-FIND(FIRST TT_pdf_stream
                  WHERE TT_pdf_stream.obj_stream = pdfStream NO-LOCK)
  THEN DO:
    RUN pdf_error(pdfStream,"pdf_tool_add","Cannot find Stream!").
    RETURN .
  END.

  IF INDEX(pdfToolName," ") > 0 THEN DO:
    RUN pdf_error(pdfStream,"pdf_tool_add","Tool Name cannot contain spaces!").
    RETURN .
  END.

  CREATE TT_pdf_tool.
  ASSIGN TT_pdf_tool.obj_stream   = pdfStream
         TT_pdf_tool.tool_name    = pdfToolName
         TT_pdf_tool.tool_Type    = pdfToolType
         TT_pdf_tool.tool_handle  = pdfToolData.

  /* Now configure a bunch of defaults for the Tool */
  CASE TT_pdf_tool.tool_type:
    WHEN "TABLE" THEN DO:
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"Outline",0,"0").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"HeaderFont",0,"Courier-Bold").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"HeaderFontSize",0,"10").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"HeaderBGColor",0,"255,255,255").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"HeaderTextColor",0,"0,0,0").

      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"DetailFont",0,"Courier").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"DetailFontSize",0,"10").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"DetailBGColor",0,"255,255,255").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"DetailTextColor",0,"0,0,0").
      RUN pdf_set_tool_parameter(pdfStream,pdfToolName,"ColumnPadding",0,"5").
    END.
  END CASE.
END. /* pdf_tool_add */

PROCEDURE pdf_tool_create :
  DEFINE INPUT PARAMETER pdfStream      AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolName    AS CHARACTER NO-UNDO.

  FIND TT_pdf_tool
       WHERE TT_pdf_tool.obj_stream = pdfStream
       AND TT_pdf_tool.tool_name  = pdfToolName NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_Tool THEN DO:
    RUN pdf_error(pdfStream,"pdf_tool_create","Cannot find Tool ("
                                              + pdfToolName + "for Stream ("
                                              + pdfStream + "!").
    RETURN .
  END.

  IF NOT VALID-HANDLE(h_PDF-Tool) THEN
    RUN VALUE(SEARCH("pdftool.p")) PERSISTENT SET h_PDF-Tool
        (INPUT THIS-PROCEDURE:HANDLE).

  /* If a call the pdf_new_page hasn't been done then ensure that we do it */
  IF pdf_Page(pdfStream) = 0 THEN
    RUN pdf_new_page(pdfStream).

  CASE TT_pdf_tool.tool_type:
    WHEN "TABLE" THEN
      RUN BuildTable IN h_PDF-Tool (INPUT pdfStream,
                                    INPUT pdfToolName,
                                    INPUT TT_pdf_tool.Tool_handle).
  END CASE.

END. /* pdf_tool_create */

PROCEDURE pdf_set_tool_parameter :

  DEFINE INPUT PARAMETER pdfStream      AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolName    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolParam   AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolCol     AS INTEGER   NO-UNDO.
  DEFINE INPUT PARAMETER pdfToolValue   AS CHARACTER NO-UNDO.

  DEFINE VARIABLE L_Integer      AS INTEGER NO-UNDO.
  DEFINE VARIABLE L_Decimal      AS DECIMAL NO-UNDO.

  DEFINE VARIABLE L_TableParams  AS CHARACTER NO-UNDO.
  L_TableParams = "Outline,"
                + "ColumnHeader,ColumnWidth,ColumnX,ColumnPadding,MaxX,MaxY,"
                + "HeaderFont,HeaderFontSize,HeaderBGColor,HeaderTextColor,"
                + "DetailFont,DetailFontSize,DetailBGColor,DetailTextColor".

  FIND TT_pdf_tool
       WHERE TT_pdf_tool.obj_stream = pdfStream
         AND TT_pdf_tool.tool_name  = pdfToolName NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_Tool THEN DO:
    RUN pdf_error(pdfStream,"pdf_tool_parameter","Cannot find Tool ("
                                              + pdfToolName + "for Stream ("
                                              + pdfStream + "!").
    RETURN .
  END.

  CASE TT_pdf_tool.tool_type:
    WHEN "TABLE" THEN DO:
      IF LOOKUP(pdfToolParam,L_TableParams) = 0 THEN DO:
        RUN pdf_error(pdfStream,"pdf_tool_parameter","Invalid Table Parameter entered(" + pdfToolName + ")!").
        RETURN .
      END.

      /* Verify for Integer Content */
      IF LOOKUP(pdfToolParam,"ColumnWidth,ColumnX,MaxX,MaxY,Outline") > 0 THEN DO:
        L_Integer = INT(pdfToolValue) NO-ERROR.
        IF ERROR-STATUS:ERROR THEN DO:
          RUN pdf_error(pdfStream,"pdf_tool_parameter",
                                  "Parameter (" + pdfToolParam + ") requires"
                                  + " an Integer value!").
          RETURN .
        END.

        IF LOOKUP(pdfToolParam,"Outline") = 0 AND L_Integer <= 0 THEN DO:
          RUN pdf_error(pdfStream,"pdf_tool_parameter",
                                  "Parameter (" + pdfToolParam + ") requires"
                                  + " a positive (non-zero) value!").
          RETURN .
        END.
      END. /* Integer Verification */

      ELSE IF LOOKUP(pdfToolParam,"HeaderFontSize,DetailFontSize") > 0 THEN DO:
        L_Decimal = DEC(pdfToolValue) NO-ERROR.
        IF ERROR-STATUS:ERROR THEN DO:
          RUN pdf_error(pdfStream,"pdf_tool_parameter",
                                  "Parameter (" + pdfToolParam + ") requires"
                                  + " a Decimal value!").
          RETURN .
        END.
      END. /* Decimal Verification */

      ELSE IF LOOKUP(pdfToolParam,"HeaderBGColor,DetailBGColor,HeaderTextColor,DetailTextColor") > 0 THEN DO:
        IF NUM-ENTRIES(pdfToolValue) <> 3 THEN DO:
          RUN pdf_error(pdfStream,"pdf_tool_parameter",
                                  "Parameter (" + pdfToolParam + ") requires"
                                  + " 3 Entries in RGB sequence comma-delimited!").
          RETURN .
        END.
      END. /* Color Verification */

      ELSE IF LOOKUP(pdfToolParam,"HeaderFont,DetailFont") > 0 THEN DO:
        IF NOT CAN-FIND(FIRST TT_pdf_font
                        WHERE TT_pdf_font.obj_stream = pdfStream
                          AND TT_pdf_font.font_name  = pdfToolValue) THEN DO:
          RUN pdf_error(pdfStream,"pdf_tool_parameter",
                                  "Parameter (" + pdfToolParam + ") requires"
                                  + " an Valid Font Name!").
          RETURN .
        END.
      END. /* Font Name Verification */

    END. /* Table Parameter Check */
  END CASE.

  FIND TT_pdf_tool_param
       WHERE TT_pdf_tool_param.obj_stream = pdfStream
         AND TT_pdf_tool_param.tool_name  = pdfToolName
         AND TT_pdf_tool_param.tool_param = pdfToolParam
         AND TT_pdf_tool_param.tool_col   = pdfToolCol
         EXCLUSIVE NO-ERROR.
  IF NOT AVAIL TT_pdf_tool_param THEN DO:
    CREATE TT_pdf_tool_param.
    ASSIGN TT_pdf_tool_param.obj_stream   = pdfStream
           TT_pdf_tool_param.tool_name    = pdfToolName
           TT_pdf_tool_param.tool_param   = pdfToolParam
           TT_pdf_tool_param.tool_col     = pdfToolCol.
  END.

  TT_pdf_tool_param.tool_value   = pdfToolValue.

END. /* pdf_set_tool_parameter */

FUNCTION pdf_get_tool_parameter RETURNS CHARACTER
        (INPUT  pdfStream      AS CHARACTER,
         INPUT  pdfToolName    AS CHARACTER,
         INPUT  pdfToolParam   AS CHARACTER,
         INPUT  pdfToolCol     AS INTEGER):

  FIND TT_pdf_tool
       WHERE TT_pdf_tool.obj_stream = pdfStream
         AND TT_pdf_tool.tool_name  = pdfToolName NO-LOCK NO-ERROR.
  IF NOT AVAIL TT_pdf_Tool THEN DO:
    RUN pdf_error(pdfStream,"pdf_tool_parameter","Cannot find Tool ("
                                              + pdfToolName + "for Stream ("
                                              + pdfStream + "!").
    RETURN "".
  END.

  FIND TT_pdf_tool_param
       WHERE TT_pdf_tool_param.obj_stream = pdfStream
         AND TT_pdf_tool_param.tool_name  = pdfToolName
         AND TT_pdf_tool_param.tool_param = pdfToolParam
         AND TT_pdf_tool_param.tool_col   = pdfToolCol NO-LOCK NO-ERROR.

  RETURN IF AVAIL TT_pdf_tool_param THEN TT_pdf_tool_param.tool_value
         ELSE "".

END FUNCTION. /* pdf_get_tool_parameter */

/* end of pdf_inc.p */
